# Copyright (c) 2006, Brian Frank, Andy Frank, and John Sublett.
# Licensed under BSD, see http://www.fandev.org/license.txt for terms.

#
# fake -- top level build tool
#
# @author   Brian Frank
# @creation 3 Mar 06
#

require 'adm/env.rb'

class Fake < Env

##########################################################################
# Main
##########################################################################

  # main
  def main()
    # check we have at least one args
    if ARGV.size < 2 then
      usage()
      return
    end
    
    targets = ARGV[1..-1]
    doc = true #targets.any? {|arg| arg == 'doc'}
    src = true #targets.any? {|arg| arg == 'src'}
    
    # init pods
    if (ARGV[0] == 'all')
      pods = @@all_pods
    else
      pods = [ ARGV[0] ]
    end
    
    # perform each target on pods
    targets.each do |target|
      case target
        when 'pod'
          do_pod(pods, doc)
        when 'native'
          do_native(pods)
        when 'clean'
          do_clean(pods)
        when 'full'
          do_clean(pods)
          do_pod(pods, doc, src)
          do_native(pods)
        when 'jstub'
          do_jstub(pods)
        when 'test'
          do_test(pods)
        when 'doc'
          # nothing - only applicable to 'pod'
        when 'src'
          # nothing - only applicable to 'pod'
        else
          raise "Unknown target: " + target
      end
      puts ""
      puts "### Success [#{target}] ###"
    end      
    
  end

##########################################################################
# Usage
##########################################################################

  def usage()
    puts "usage:"
    puts "  fake <podName|all> <target> [target]*"
    puts "targets:"
    puts "  pod         run fanc to build specified pod"
    puts "  native      build native jar"
    puts "  clean       clean any derived files"
    puts "  full        clean+pod+native"
    puts "  jstub       force jstub even if no native code"
    puts "  doc         build fandoc"
    puts "  src         include source" 
    puts "  test        run fant for specified pod"
  end

##########################################################################
# Pod
##########################################################################

  def do_pod(pods, doc, src)
    pods.each { |p| pod(p, doc, src) }    
  end

  def pod(pod, doc, src)
    puts "-- pod [" + pod + "] doc=#{doc} src=#{src}"
    
    # sys is special case that we compile jar before pod
    if pod == 'sys' then
      syscall("ruby #@fan_adm/makej.rb")
    end
    
    # call fanc
    exe = "fanc"
    if (doc)
      exe += " -doc"
    end
    if (src)
      exe += " -src"
    end
    exe += " #@fan_src/#{pod}"
    syscall(exe)
  end

##########################################################################
# Native
##########################################################################

  def do_native(pods)
    pods.each { |p| native(p) }
  end
  
  def native(pod)    
    # we don't jar sys here, it happens in pod
    return if pod == 'sys'
    
    # also don't jar if a java dir does not exist
    pod_home = File.join(@fan_src, pod)
    java_dir = File.join(pod_home, "java") 
    return unless File.exists?(java_dir)
    
    puts "-- jstub [" + pod + "]"
    syscall("ruby #@fan_adm/makenative.rb -j #{pod}")
  end

##########################################################################
# Jstub
##########################################################################

  def do_jstub(pods)
    pods.each { |p| jstub(p) }
  end
  
  def jstub(pod)    
    # we don't jstub sys here, it happens in pod
    return if pod == 'sys'
    
    # we only run jstub if a java dir does not exist (otherwise
    # native should have jstubbed, javac and rejarred)
    pod_home = File.join(@fan_src, pod)
    java_dir = File.join(pod_home, "java") 
    return if File.exists?(java_dir)
    
    puts "-- jstub [" + pod + "]"
    syscall("#@fan_bin/jstub.bat -d #@fan_lib_java #{pod}")
  end

##########################################################################
# Clean
##########################################################################

  def do_clean(pods)
    pods.each { |p| clean(p) }
  end
  
  def clean(pod)
    puts "-- clean [" + pod + "]"
    
    if (pod == 'sys') then
      syscall("ruby #@fan_adm/makej.rb clean")
    end
    
    nukeLog("#@fan_lib_fan/#{pod}.pod", true)
    nukeLog("#@fan_lib_java/#{pod}.jar", true)
    nukeLog("#@fan_src/#{pod}/temp-java", true)
  end

##########################################################################
# Test
##########################################################################

  def do_test(pods)
    puts "-- test [" + pods.join(',') + "]"
    puts ""
    puts ""
    syscall("fant #{pods.join(',')}")    
  end
  
end

# script
Fake.new.main