//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//

**
** Assembler assembles all the TypeDefs into their fcode representation.
**
class Assembler : CompilerSupport, FConst
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  new make(Compiler compiler)
    : super(compiler)
  {
  }

//////////////////////////////////////////////////////////////////////////
// Assemble
//////////////////////////////////////////////////////////////////////////

  FPod assemblePod()
  {
    fpod = FPod.make(compiler.pod.name, null)

    fpod.podName = compiler.input.podName
    fpod.version = compiler.input.version
    fpod.depends = compiler.input.depends
    fpod.facets  = assemblePodFacets(fpod)

    fpod.types  = FType[,]
    types.each |TypeDef t|
    {
      fpod.types.add(assembleType(t))
    }

    return fpod
  }

  private FFacet[] assemblePodFacets(FPod fpod)
  {
    input := compiler.input
    flags := FConst.Reflective
    asm := FacetsAsm.make(fpod)

    asm.utf("description", flags, input.description)
    asm.utf("build.host",  flags, Sys.hostName)
    asm.utf("build.user",  flags, Sys.userName)
    asm.utf("build.time",  flags, Time.now.toStr)

    return asm.facets
  }

  private FType assembleType(TypeDef def)
  {
    t := FType.make(fpod)

    t.hollow = false
    t.flags  = def.flags
    t.self   = typeRef(def)
    t.base   = (def.base == null) ? -1 : typeRef(def.base)
    def.mixins.map(t.mixins = Int[,]) |CType m->Obj| { return typeRef(m) }
    def.fieldDefs.map(t.fields = FField[,]) |FieldDef f->Obj| { return assembleField(f) }
    def.methodDefs.map(t.methods = FMethod[,]) |MethodDef m->Obj| { return assembleMethod(m) }

    facets := FacetsAsm.make(fpod)
    if (compiler.input.mode == CompilerInputMode.str)
      facets.sourceFile(def.location.fileUri)
    else
      facets.sourceFile(def.location.filename)
    facets.lineNumber(def.location.line)
    t.facets = facets.facets

    return t
  }

  FField assembleField(FieldDef def)
  {
    f := FField.make
    f.name    = name(def.name)
    f.flags   = def.flags
    f.typeRef = typeRef(def.fieldType)

    facets := FacetsAsm.make(fpod)
    facets.lineNumber(def.location.line)
    f.facets = facets.facets

    return f;
  }

  FMethod assembleMethod(MethodDef def)
  {
    facets := FacetsAsm.make(fpod)

    m := FMethod.make

    m.name         = name(def.name)
    m.flags        = def.flags
    m.ret          = typeRef(def.ret)
    m.inheritedRet = typeRef(def.inheritedReturnType)
    m.paramCount   = def.params.size
    m.localCount   = def.vars.size - def.params.size

    m.vars = FMethodVar[,]
    def.vars.map(m.vars) |MethodVar v->Obj|
    {
      f := FMethodVar.make
      f.name    = name(v.name)
      f.typeRef = typeRef(v.ctype)
      f.flags   = v.flags
      if (v.paramDef != null) f.def = assembleExpr(v.paramDef.def)
      return f
    }

    m.code = assembleCode(def, facets)

    facets.lineNumber(def.location.line)
    m.facets = facets.facets

    return m;
  }

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  Int typeRef(CType type)
  {
    return fpod.addTypeRef(type)
  }

  Int name(Str val)
  {
    return fpod.addName(val)
  }

  private Buf assembleCode(MethodDef def , FacetsAsm facets)
  {
    block := def.code
    if (block == null) return null

    asm := CodeAsm.make(compiler, def.location, fpod)
    if (def.ctorChain != null) asm.expr(def.ctorChain)
    asm.block(block)

    if (asm.errCount > 0) facets.add(ErrTable, 0, asm.finishErrTable)
    if (asm.lineCount > 0) facets.add(LineNumbers, 0, asm.finishLines)

    return asm.finishCode
  }

  private Buf assembleExpr(Expr expr)
  {
    if (expr == null) return null
    asm := CodeAsm.make(compiler, expr.location, fpod)
    asm.expr(expr)
    return asm.finishCode
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  FPod fpod
}