//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//   29 Aug 06  Brian Frank  Ported from Java to Fan
//

**
** Namespace is responsible for providing a unified view pods, types,
** and slots between the entities currently being compiled and the
** entities being imported from pre-compiled pods.
**
class Namespace
{

//////////////////////////////////////////////////////////////////////////
// Mapping
//////////////////////////////////////////////////////////////////////////

  **
  ** Map an imported Pod into a CPod
  **
  CPod importPod(Pod pod)
  {
    if (pods == null) pods = Str:CPod[:]
    cpod := pods[pod.name]
    if (cpod == null) pods[pod.name] = cpod = ImportedPod.make(this, pod)
    return cpod
  }

  **
  ** Map an imported Type into a CType
  **
  CType importType(Type t)
  {
    if (t == null) return null
    if (types == null) types = Str:CType[:]
    ctype := types[t.signature]
    if (ctype == null)
    {
      if (t.params.isEmpty)
        ctype = ImportedType.make(this, t)
      else
        ctype = GenericType.makeImport(this, t)
      types[t.signature] = ctype
    }
    return ctype
  }

  **
  ** Map one of the generic parameter types such as "sys::V" into a CType
  **
  CType genericParameter(Str id)
  {
    // force loading all the generic methods
    listType.methods
    mapType.methods
    methodType.methods

    // all the generic parameter types should be cached now
    ctype := types["sys::$id"]
    if (ctype == null) throw Err.make(id)
    return ctype
  }

  **
  ** Map a list of imported Types into a CTypes
  **
  CType[] importTypes(Type[] t)
  {
    return (CType[])t.map(CType[,]) |Type x->Obj| { return importType(x) }
  }

  **
  ** Map an imported Slot into a CSlot
  **
  CSlot importSlot(Slot slot)
  {
    if (slot is Method)
      return importMethod((Method)slot)
    else
      return importField((Field)slot)
  }

  **
  ** Map an imported Field into a CField
  **
  CField importField(Field f)
  {
    return ImportedField.make(importType(f.parent), f)
  }

  **
  ** Map an imported Method into a CMethod
  **
  CMethod importMethod(Method m)
  {
    return ImportedMethod.make(importType(m.parent), m)
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  Void dump()
  {
    echo("--- CPods ---")
    pods.values.sort.each |CPod p| { echo("  $p") }
    echo("--- CTypes ---")
    types.values.sort.each |CType t| { echo("  $t") }
  }

//////////////////////////////////////////////////////////////////////////
// Predefined
//////////////////////////////////////////////////////////////////////////

  // place holder type used for resolve errors
  readonly CType error := ImportedType.make(this, Void.type)

  readonly CPod sysPod := importPod(Obj.type.pod)

  readonly CType objType      := importType(Obj.type)
  readonly CType boolType     := importType(Bool.type)
  readonly CType enumType     := importType(Enum.type)
  readonly CType intType      := importType(Int.type)
  readonly CType floatType    := importType(Float.type)
  readonly CType strType      := importType(Str.type)
  readonly CType durationType := importType(Duration.type)
  readonly CType listType     := importType(List.type)
  readonly CType mapType      := importType(Map.type)
  readonly CType methodType   := importType(Method.type)
  readonly CType errType      := importType(Err.type)
  readonly CType typeType     := importType(Type.type)
  readonly CType rangeType    := importType(Range.type)
  readonly CType uriType      := importType(Uri.type)
  readonly CType voidType     := importType(Void.type)

  readonly CMethod objTrap      := importMethod(Slot.findMethod("sys::Obj.trap"))
  readonly CMethod boolNot      := importMethod(Slot.findMethod("sys::Bool.not"))
  readonly CMethod intIncrement := importMethod(Slot.findMethod("sys::Int.increment"))
  readonly CMethod intDecrement := importMethod(Slot.findMethod("sys::Int.decrement"))
  readonly CMethod intPlus      := importMethod(Slot.findMethod("sys::Int.plus"))
  readonly CMethod floatPlus    := importMethod(Slot.findMethod("sys::Float.plus"))
  readonly CMethod floatMinus   := importMethod(Slot.findMethod("sys::Float.minus"))
  readonly CMethod strPlus      := importMethod(Slot.findMethod("sys::Str.plus"))
  readonly CMethod strBufMake   := importMethod(Slot.findMethod("sys::StrBuf.make"))
  readonly CMethod strBufAdd    := importMethod(Slot.findMethod("sys::StrBuf.add"))
  readonly CMethod strBufToStr  := importMethod(Slot.findMethod("sys::StrBuf.toStr"))
  readonly CMethod listMake     := importMethod(Slot.findMethod("sys::List.make"))
  readonly CMethod listMakeObj  := importMethod(Slot.findMethod("sys::List.makeObj"))
  readonly CMethod listAdd      := importMethod(Slot.findMethod("sys::List.add"))
  readonly CMethod mapMake      := importMethod(Slot.findMethod("sys::Map.make"))
  readonly CMethod mapSet       := importMethod(Slot.findMethod("sys::Map.set"))
  readonly CMethod enumOrdinal  := importMethod(Slot.findMethod("sys::Enum.ordinal"))
  readonly CMethod rangeMakeInclusive := importMethod(Slot.findMethod("sys::Range.makeInclusive"))
  readonly CMethod rangeMakeExclusive := importMethod(Slot.findMethod("sys::Range.makeExclusive"))
  readonly CMethod slotFindMethod     := importMethod(Slot.findMethod("sys::Slot.findMethod"))

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  private Str:CPod pods       // keyed by pod name
  private Str:CType types     // keyed by type qname

}