//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   17 Sep 06  Brian Frank  Creation
//

**
** CheckErrorsTest
**
class CheckErrorsTest : CompilerTest
{

//////////////////////////////////////////////////////////////////////////
// Types
//////////////////////////////////////////////////////////////////////////

  Void testTypeFlags()
  {
    // parser stage
    verifyErrors("
      abstract mixin A {}
      final mixin B {}
      abstract enum C { none }
      const final enum D { none }
      public public class E {}
      abstract internal abstract class F {}
      const mixin G {}
      ",
       [
         2,  16, "The 'abstract' modifier is implied on mixin",
         3,  13, "Cannot use 'final' modifier on mixin",
         4,  16, "Cannot use 'abstract' modifier on enum",
         5,  19, "The 'const' modifier is implied on enum",
         5,  19, "The 'final' modifier is implied on enum",
         6,  14, "Repeated modifier",
         7,  25, "Repeated modifier",
         8,  13, "Cannot use 'const' modifier on mixin",
       ])

    // check errors stage
    verifyErrors("
      new class A {}
      private class B {}
      protected class C {}
      virtual static class D {}
      native class E {}
      readonly class F {}
      public internal class G {}
      abstract final class H {}
      ",
       [
         2,  11, "Cannot use 'new' modifier on type",
         3,  15, "Cannot use 'private' modifier on type",
         4,  17, "Cannot use 'protected' modifier on type",
         5,  22, "Cannot use 'static' modifier on type",
         5,  22, "Cannot use 'virtual' modifier on type",
         6,  14, "Cannot use 'native' modifier on type",
         7,  16, "Cannot use 'readonly' modifier on type",
         8,  23, "Invalid combination of 'public' and 'internal' modifiers",
         9,  22, "Invalid combination of 'abstract' and 'final' modifiers",
       ])
  }

  Void testTypeAbstractSlots()
  {
    // errors
    verifyErrors("
      class A { abstract Void x()  }
      class B { abstract Void x(); abstract Void y(); }
      class C : B {}
      class D : A { abstract Void y(); }
      class E : B, X { override Void a() {} override Void x() {} }
      mixin X { abstract Void a(); abstract Void b(); }
      ",
       [
         2,  7, "Class 'A' must be abstract since it contains abstract slots",
         3,  7, "Class 'B' must be abstract since it contains abstract slots",
         4,  7, "Class 'C' must be abstract since it inherits but doesn't override '$podName::B.x'",
         4,  7, "Class 'C' must be abstract since it inherits but doesn't override '$podName::B.y'",
         5,  7, "Class 'D' must be abstract since it inherits but doesn't override '$podName::A.x'",
         5,  7, "Class 'D' must be abstract since it contains abstract slots",
         6,  7, "Class 'E' must be abstract since it inherits but doesn't override '$podName::X.b'",
         6,  7, "Class 'E' must be abstract since it inherits but doesn't override '$podName::B.y'",
       ])
  }

  Void testTypeMisc()
  {
    // check inherit stage
    verifyErrors("
      class A { Type type }
      class B { Type type() { return null } }
      ",
       [
         2, 17, "Must specify override keyword to override 'sys::Obj.type'",
         3, 17, "Must specify override keyword to override 'sys::Obj.type'",
       ])

    // check errors stage
    verifyErrors("
      class A { override Type type }
      class B { override Type type() { return null } }
      ",
       [
         2, 17, "Cannot override Obj.type() knuckle head",
         3, 17, "Cannot override Obj.type() knuckle head",
       ])
  }

//////////////////////////////////////////////////////////////////////////
// Test protection scopes
//////////////////////////////////////////////////////////////////////////

  Void testProtectionScopes()
  {
    // first create a pod with internal types/slots
    compile("
      class Public
      {
        virtual public    Void mPublic()    {}
        virtual protected Void mProtected() {}
        virtual internal  Void mInternal()  {}
                private   Void mPrivate()   {} // can't mix virtual+private

        static public    Void msPublic()    {}
        static protected Void msProtected() {}
        static internal  Void msInternal()  {}
        static private   Void msPrivate()   {}

        virtual public    Int fPublic
        virtual protected Int fProtected
        virtual internal  Int fInternal
                private   Int fPrivate   // can't mix virtual+private

        public            Int fPublicProtected { protected set }
        readonly public   Int fPublicReadonly
        protected         Int fProtectedInternal { internal set }
      }

      internal class InternalClass
      {
        Void m() {}
      }

      internal mixin InternalMixin
      {
      }
      ")

    p := pod.types[0]
    ic := pod.types[1]
    im := pod.types[2]

    // CheckInherit step
    verifyErrors("
      using $p.pod.name

      class C00 : Public {}
      class C01 : InternalClass {}
      class C02 : InternalMixin {}
      mixin C03 : InternalMixin {}
    ",
    [
      5, 7, "Class 'C01' cannot access internal scoped class '$ic'",
      6, 7, "Type 'C02' cannot access internal scoped mixin '$im'",
      7, 7, "Type 'C03' cannot access internal scoped mixin '$im'",
    ])

    // Inherit step
    verifyErrors("
      using $p.pod.name

      class C01 : Public { override Void figgle() {} }
      class C02 : Public { Str toStr() { return null } }
      class C03 : Public { override Void mPublic() {} }    // ok
      class C04 : Public { override Void mProtected() {} } // ok

      class C05 : Public { override Void mInternal() {} }
      class C06 : Public { override Void mPrivate() {} }
    ",
    [
      4, 28, "Override of unknown virtual slot 'figgle'",
      5, 28, "Must specify override keyword to override 'sys::Obj.toStr'",

      // TODO: internal/privates never make it this far to tell you its a scope problem...
      9, 28, "Override of unknown virtual slot 'mInternal'",
     10, 28, "Override of unknown virtual slot 'mPrivate'",
    ])

    // CheckErrors step
    verifyErrors("
      using $p.pod.name

      class C04 : Public { Void f() { mPublic; x := fPublic } } // ok
      class C05 : Public { Void f() { mProtected; x := fProtected } } // ok
      class C06 { Void f(Public p) { p.mProtected  } }
      class C07 { Void f(Public p) { p.mInternal  } }
      class C08 { Void f(Public p) { p.mPrivate  } }
      class C09 { Void f() { Public.msProtected  } }
      class C10 { Void f() { Public.msInternal  } }
      class C11 { Void f() { Public.msPrivate  } }

      class C13 { Obj f(Public p) { return p.fPublic } } // ok
      class C14 : Public { Obj f(Public p) { return p.fProtected} } // ok
      class C15 { Obj f(Public p) { return p.fPublicProtected } } // ok
      class C16 { Obj f(Public p) { return p.fPublicReadonly } } // ok
      class C17 : Public { Obj f(Public p) { return p.fProtectedInternal } } // ok

      class C19 { Obj f(Public p) { return p.fProtected } }
      class C20 { Obj f(Public p) { return p.fProtectedInternal } }
      class C21 { Obj f(Public p) { return p.fInternal } }
      class C22 { Obj f(Public p) { return p.fPrivate } }

      class C24 { Void f(Public p) { p.fPublic = 7 } }  // ok
      class C25 : Public { Void f(Public p) { p.fProtected = 7 } } // ok
      class C26 : Public { Void f(Public p) { p.fPublicProtected = 7 } } // ok
      class C27 { Void f(Public p) { p.fProtected = 7 } }
      class C28 { Void f(Public p) { p.fInternal = 7 } }
      class C29 { Void f(Public p) { p.fPrivate = 7 } }
      class C30 { Void f(Public p) { p.fPublicProtected = 7; p.fPublicProtected++ } }
      class C31 { Void f(Public p) { p.fPublicReadonly = 7; p.fPublicReadonly++ } }
      class C32 : Public { Void f(Public p) { p.fProtectedInternal = 7; p.fProtectedInternal++ } }
    ",
    [
      6, 40, "Protected method '${p}.mProtected' not accessible",
      7, 40, "Internal method '${p}.mInternal' not accessible",
      8, 40, "Private method '${p}.mPrivate' not accessible",
      9, 37, "Protected method '${p}.msProtected' not accessible",
     10, 37, "Internal method '${p}.msInternal' not accessible",
     11, 37, "Private method '${p}.msPrivate' not accessible",

     19, 46, "Protected field '${p}.fProtected' not accessible",
     20, 46, "Protected field '${p}.fProtectedInternal' not accessible",
     21, 46, "Internal field '${p}.fInternal' not accessible",
     22, 46, "Private field '${p}.fPrivate' not accessible",

     27, 40, "Protected field '${p}.fProtected' not accessible",
     28, 40, "Internal field '${p}.fInternal' not accessible",
     29, 40, "Private field '${p}.fPrivate' not accessible",
     30, 40, "Protected setter of field '${p}.fPublicProtected' not accessible",
     30, 64, "Protected setter of field '${p}.fPublicProtected' not accessible",
     31, 40, "Private setter of field '${p}.fPublicReadonly' not accessible",
     31, 63, "Private setter of field '${p}.fPublicReadonly' not accessible",
     32, 49, "Internal setter of field '${p}.fProtectedInternal' not accessible",
     32, 75, "Internal setter of field '${p}.fProtectedInternal' not accessible",
    ])
  }

  Void testClosureProtectionScopes()
  {
    // verify closure get access to external class privates
    compile("
      class Foo : Goo
      {
        private static Int x() { return 'x' }
        static Int testX()
        {
          f := |->Int| { return x }
          return f.call0
        }

        protected static Int y() { return 'y' }
        static Int testY()
        {
          f := |->Int|
          {
            g := |->Int| { return y  }
            return g.call0
          }
          return f.call0
        }

        static Int testZ()
        {
          f := |->Int| { return z }
          return f.call0
        }
      }

      class Goo
      {
        protected static Int z() { return 'z' }
      }")

     t := pod.types[1]
     verifyEq(t.name, "Foo")
     verifyEq(t.method("testX").call0, 'x')
     verifyEq(t.method("testY").call0, 'y')
     verifyEq(t.method("testZ").call0, 'z')
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  Void testFieldFlags()
  {
    // parser stage
    verifyErrors("
      abstract class Foo
      {
        readonly readonly Str f01
        Str f02 { override get { return f02 } }
        Str f03 { internal override get { return f03 } }
        Str f04 { override set {} }
      }
      ",
       [
         4,  18, "Repeated modifier",
         5,  19, "Cannot use modifiers on field getter",
         6,  19, "Cannot use modifiers on field getter",
         7,  19, "Cannot use modifiers on field setter except to narrow protection",
       ])

    // check errors stage
    verifyErrors("
      abstract class Foo : Bar
      {
        // new Str f00 - parser actually catches this
        final Str f01
        native Str f02

        public protected Str f03
        public private Str f04
        public internal Str f05
        protected private Str f06
        protected internal Str f07
        internal private Str f08

        Str f10 { public protected set {} }
        Str f11 { public private  set {} }
        Str f12 { public internal  set {} }
        Str f13 { protected private  set {} }
        Str f14 { protected internal  set {} }
        Str f15 { internal private  set {} }

        private Str f20 { public set {} }
        private Str f21 { protected set {} }
        private Str f22 { internal set {} }
        internal Str f23 { public set {} }
        internal Str f24 { protected set {} }
        protected Str f25 { public set {} }
        protected Str f26 { internal set {} } // ok

        const abstract Str f30
        //const override Str f31 TODO
        const virtual  Str f32

        virtual private Str f33
      }

      class Bar
      {
        virtual Str f31
      }
      ",
       [
         5,  9, "Cannot use 'final' modifier on field",
         6,  9, "Cannot use 'native' modifier on field",

         8,  9, "Invalid combination of 'public' and 'protected' modifiers",
         9,  9, "Invalid combination of 'public' and 'private' modifiers",
        10,  9, "Invalid combination of 'public' and 'internal' modifiers",
        11,  9, "Invalid combination of 'protected' and 'private' modifiers",
        12,  9, "Invalid combination of 'protected' and 'internal' modifiers",
        13,  9, "Invalid combination of 'private' and 'internal' modifiers",

        15,  9, "Invalid combination of 'public' and 'protected' modifiers",
        16,  9, "Invalid combination of 'public' and 'private' modifiers",
        17,  9, "Invalid combination of 'public' and 'internal' modifiers",
        18,  9, "Invalid combination of 'protected' and 'private' modifiers",
        19,  9, "Invalid combination of 'protected' and 'internal' modifiers",
        20,  9, "Invalid combination of 'private' and 'internal' modifiers",

        22,  9, "Setter cannot have wider visibility than the field",
        23,  9, "Setter cannot have wider visibility than the field",
        24,  9, "Setter cannot have wider visibility than the field",
        25,  9, "Setter cannot have wider visibility than the field",
        26,  9, "Setter cannot have wider visibility than the field",
        27,  9, "Setter cannot have wider visibility than the field",

        30,  9, "Invalid combination of 'const' and 'abstract' modifiers",
        //31,  9, "Invalid combination of 'const' and 'override' modifiers", TODO
        32,  9, "Invalid combination of 'const' and 'virtual' modifiers",

        34,  9, "Invalid combination of 'private' and 'virtual' modifiers",
       ])
  }

  Void testFields()
  {
    verifyErrors("
      mixin MixIt
      {
        Str a
        virtual Int b
        abstract Int c { get { return @c } }
        abstract Int d { set { @d = val } }
        abstract Int e { get { return @e } set { @e = val } }
        const Int f := 3
        abstract Int g := 5
      }

      abstract class Foo
      {
        abstract Int c { get { return @c } }
        abstract Int d { set { @d = val } }
        abstract Int e { get { return @e } set { @e = val } }
        abstract Int f := 3
      }
      ",
       [
         4,  9, "Mixin field 'a' must be abstract",
         5,  9, "Mixin field 'b' must be abstract",
         6,  9, "Abstract field 'c' cannot have getter or setter",
         7,  9, "Abstract field 'd' cannot have getter or setter",
         8,  9, "Abstract field 'e' cannot have getter or setter",
         9,  9, "Mixin field 'f' must be abstract",
        10, 27, "Abstract field 'g' cannot have initializer",

        15,  9, "Abstract field 'c' cannot have getter or setter",
        16,  9, "Abstract field 'd' cannot have getter or setter",
        17,  9, "Abstract field 'e' cannot have getter or setter",
        18, 27, "Abstract field 'f' cannot have initializer",
       ])
  }

  Void testConst()
  {
    // Parser step
    verifyErrors("
      const class Foo
      {
        const static Int a { get { return 3 } }
        const static Int b { set {  } }
        const static Int c { get { return 3 } set { } }

        const Int d { get { return 3 } }
        const Int e { set {  } }
        const Int f { get { return 3 } set { } }
      }
      ",
       [
         4, 30, "Const field 'a' cannot have getter",
         5, 30, "Const field 'b' cannot have setter",
         6, 30, "Const field 'c' cannot have getter",
         6, 47, "Const field 'c' cannot have setter",

         8, 23, "Const field 'd' cannot have getter",
         9, 23, "Const field 'e' cannot have setter",
        10, 23, "Const field 'f' cannot have getter",
        10, 40, "Const field 'f' cannot have setter",
       ])

    // CheckErrors step
    verifyErrors("
      const class Foo : Bar
      {
        static Int a := 3

        const static Int b := 3
        static { b = 5 }
        static Void goop() { b = 7; b += 3; b++ }

        //const static Int c { get { return 3 } }
        //const static Int d { set {  } }
        //const static Int e { get { return 3 } set { } }

        const Int f := 3
        new make() { f = 5 }
        Void wow() { f = 7; f++; }
        static Void bow(Foo o) { o.f = 9; o.f += 2 }

        //const Int g { get { return 3 } }
        //const Int h { set {  } }
        //const Int i { get { return 3 } set { } }

        private Str j
        private const StrBuf k
        const Buf[] l
        const Str:Obj m
        const Obj:Int n
        const Num:Duration ok1
        const Num:Str[][] ok2
      }

      class Bar {}
      class Roo : Foo {}
      enum Boo { none;  private Int x }

      const class Outside : Foo
      {
        new make() { f = 99 }
        static { b++ }
      }
      ",
       [
         4,  9, "Static field 'a' must be const",

         8, 30, "Cannot set const static field 'b' outside of static initializer",
         8, 37, "Cannot set const static field 'b' outside of static initializer",
         8, 45, "Cannot set const static field 'b' outside of static initializer",

        16, 22, "Cannot set const field 'f' outside of constructor",
        16, 29, "Cannot set const field 'f' outside of constructor",
        17, 36, "Cannot set const field 'f' outside of constructor",
        17, 45, "Cannot set const field 'f' outside of constructor",

        24,  9, "Const field 'k' has non-const type 'sys::StrBuf'",
        25,  9, "Const field 'l' has non-const type 'sys::Buf[]'",
        26,  9, "Const field 'm' has non-const type '[sys::Str:sys::Obj]'",
        27,  9, "Const field 'n' has non-const type '[sys::Obj:sys::Int]'",

         2, 13, "Const class 'Foo' cannot subclass non-const class '$podName::Bar'",
        23,  9, "Const class 'Foo' cannot contain non-const field 'j'",

        33,  7, "Non-const class 'Roo' cannot subclass const class '$podName::Foo'",
        34, 25, "Const class 'Boo' cannot contain non-const field 'x'",

        38, 22, "Cannot set const field '$podName::Foo.f'",
        39, 18, "Cannot set const field '$podName::Foo.b'",
       ])
  }

  Void testConstCollections()
  {
    verifyErrors("
      const class Foo
      {
        new make()
        {
          a = [0, 1, 2].toConst          // ok
          a = [0, 1, 2]                  // missing toConst()
          a = toConst([0, 1, 2])         // doesn't count
          a = toConst([0, 1, 2]).toConst // ok
        }

        static Int[] toConst(Int[] x) { return x }

        const Int[] a
        const static Int:Str b := [99:\"yikes!\"]
        static { b = [99:\"yikes!\"].toConst } // ok
      }

      ",
       [
         7, 15, "Must call List.toConst() when setting const field 'a'",
         8, 15, "Must call List.toConst() when setting const field 'a'",
        15, 35, "Must call Map.toConst() when setting const field 'b'",
       ])
  }

  Void testFieldStorage()
  {
    verifyErrors("
      class Foo : Root
      {
        Int m00() { return @r00 }
        Int m01() { return this.@r00 }

        Int f00 { get { return f00 } }
        Int f01 { set { f01 = val } }
        Int f02 { get { return f02 } set { f02 = val } }

        override Int r01 { set { @r01 = val } }
      }

      class Root
      {
        Int r00
        virtual Int r01
      }
      ",
       [
         4, 28, "Field storage for '$podName::Root.r00' not accessible",
         5, 33, "Field storage for '$podName::Root.r00' not accessible",

         7, 32, "Cannot use field accessor inside accessor itself - use '@' operator",
         8, 25, "Cannot use field accessor inside accessor itself - use '@' operator",
         9, 32, "Cannot use field accessor inside accessor itself - use '@' operator",
         9, 44, "Cannot use field accessor inside accessor itself - use '@' operator",

        11, 34, "Field storage of inherited field '$podName::Root.r01' not accessible (might try super)",
       ])
  }

//////////////////////////////////////////////////////////////////////////
// Methods
//////////////////////////////////////////////////////////////////////////

  Void testMethodFlags()
  {
    // parser stage
    verifyErrors("
      abstract class Foo
      {
        abstract internal abstract Void m01()
      }
      ",
       [
         4,  27, "Repeated modifier",
       ])

    // check errors stage
    verifyErrors("
      abstract class Foo : Whatever
      {
        final Void m00() {}
        const Void m01() {}
        readonly Void m02() {}

        public protected Void m10() {}
        public private Void m11() {}
        public internal Void m12() {}
        protected private Void m13() {}
        protected internal Void m14() {}
        internal private Void m15() {}

        new override m22() {}
        new virtual m23() {}
        abstract native Void m24()
        static abstract Void m25()
        static override Void m26() {}
        static virtual Void m27() {}

        private virtual Void m28() {}
      }

      abstract class Bar
      {
        new abstract m20 ()
        new native m21()
      }

      abstract class Whatever
      {
        virtual Void m22() {}
        virtual Void m26() {}
      }

      ",
       [
         4,  9, "Cannot use 'final' modifier on method",
         5,  9, "Cannot use 'const' modifier on method",
         6,  9, "Cannot use 'readonly' modifier on method",

         8,  9, "Invalid combination of 'public' and 'protected' modifiers",
         9,  9, "Invalid combination of 'public' and 'private' modifiers",
        10,  9, "Invalid combination of 'public' and 'internal' modifiers",
        11,  9, "Invalid combination of 'protected' and 'private' modifiers",
        12,  9, "Invalid combination of 'protected' and 'internal' modifiers",
        13,  9, "Invalid combination of 'private' and 'internal' modifiers",

        15,  9, "Invalid combination of 'new' and 'override' modifiers",
        16,  9, "Invalid combination of 'new' and 'virtual' modifiers",
        17,  9, "Invalid combination of 'abstract' and 'native' modifiers",
        18,  9, "Invalid combination of 'static' and 'abstract' modifiers",
        19,  9, "Invalid combination of 'static' and 'override' modifiers",
        20,  9, "Invalid combination of 'static' and 'virtual' modifiers",

        22,  9, "Invalid combination of 'private' and 'virtual' modifiers",

        27,  9, "Invalid combination of 'new' and 'abstract' modifiers",
        28,  9, "Invalid combination of 'new' and 'native' modifiers",
       ])
  }

  Void testMethods()
  {
    // errors
    verifyErrors("
      class A { new make(Str n) {}  }
      class B { private new make() {} }
      class C : A { }
      class D : B { }
      class E : A { new makeIt() {} }
      class F : B { new makeIt() {} }
      mixin G { new make() {} }
      class H { Void f(Int a := null, Int b) {} }
      ",
       [
         4,  7, "Must call super class constructor in 'make'",
         5,  7, "Must call super class constructor in 'make'",
         6, 21, "Must call super class constructor in 'makeIt'",
         7, 21, "Must call super class constructor in 'makeIt'",
         8, 17, "Mixins cannot have constructors",
         9, 39, "Parameter 'b' must have default",
       ])
  }

//////////////////////////////////////////////////////////////////////////
// Statements
//////////////////////////////////////////////////////////////////////////

  Void testStmt()
  {
    // errors
    verifyErrors("class Foo
      {
        static Obj m03() { if (0) return 1; return 2; }
        static Obj m04() { throw 3 }
        static Str m05() { return 6 }
        static Obj m06() { for (;\"x\";) m03(); return null }
        static Obj m07() { while (\"x\") m03(); return null }
        static Void m08() { break; continue }
        static Void m09() { Str x := 4.0 }
        static Void m10() { try { m03 } catch (Str x) {} }
        static Void m11() { try { m03 } catch (IOErr x) {} catch (IOErr x) {} }
        static Void m12() { try { m03 } catch (Err x) {} catch (IOErr x) {} }
        static Void m13() { try { m03 } catch (Err x) {} catch {} }
        static Void m14() { try { m03 } catch {} catch {} }
        static Void m15() { switch (Weekday.sun) { case 4: return } }
        static Void m16() { switch (2) { case 0: case 0: return } }
        static Void m17() { switch (Weekday.sun) { case Weekday.sun: return; case Weekday.sun: return } }

        static Void m19() { try { return } finally { return } }
        static Int m20() { try { return 1 } finally { return 2 } }
        static Obj m21() { try { try { return m03 } finally { return 8 } } finally { return 9 } }
        static Obj m22() { try { try { return m03 } finally { return 8 } } finally {} }
        static Obj m23() { try { try { return m03 } finally { } } finally { return 9 } }
        static Void m24() { while (true) { try { echo(3) } finally { break } } }
        static Void m25() { while (true) { try { echo(3) } finally { continue } } }
        static Void m26() { for (;;) { try { try { m03 } finally { break } } finally { continue } } }
      }",
       [3, 32, "If condition must be Bool, not 'sys::Int'",
        4, 34, "Must throw Err, not 'sys::Int'",
        5, 35, "Cannot return 'sys::Int' as 'sys::Str'",
        6, 34, "For condition must be Bool, not 'sys::Str'",
        7, 35, "While condition must be Bool, not 'sys::Str'",
        8, 29, "Break outside of loop (break is implicit in switch)",
        8, 36, "Continue outside of loop",
        9, 38, "'sys::Float' is not assignable to 'sys::Str'",
        10, 48, "Must catch Err, not 'sys::Str'",
        11, 60, "Already caught 'sys::IOErr'",
        12, 58, "Already caught 'sys::IOErr'",
        13, 58, "Already caught 'sys::Err'",
        14, 50, "Already caught 'sys::Err'",
        15, 57, "Incomparable types 'sys::Int' and 'sys::Weekday'",
        16, 55, "Duplicate case label",
        17, 91, "Duplicate case label",

        19, 54, "Cannot leave finally block",
        20, 55, "Cannot leave finally block",
        21, 63, "Cannot leave finally block",
        21, 86, "Cannot leave finally block",
        22, 63, "Cannot leave finally block",
        23, 77, "Cannot leave finally block",
        24, 70, "Cannot leave finally block",
        25, 70, "Cannot leave finally block",
        26, 68, "Cannot leave finally block",
        26, 88, "Cannot leave finally block",
       ])
  }

//////////////////////////////////////////////////////////////////////////
// Expressions
//////////////////////////////////////////////////////////////////////////

  Void testExpr()
  {
    // errors
    verifyErrors("class Foo
      {
        new make() { return }
        static Obj m00() { return 1.0..2 }
        static Obj m01() { return 2..[,] }
        static Obj m02() { return !4 }
        static Obj m03() { return 4 && true }
        static Obj m04() { return 0ns || [,] }
        static Void m05(Str x) { x = true }
        Void m06() { this.make }
        Void m07() { this.m00 }
        static Void m08(Str x) { m07; Foo.m07() }
        Void m09(Str x) { this.sf.size }
        static Void m10(Str x) { f.size; Foo.f.size }
        static Void m11(Str x) { this.m06; super.hash() }
        static Obj m12(Str x) { return 1 ? 2 : 3 }
        static Bool m14(Str x, Int y) { return x === y }
        static Bool m15(Str x, Int y) { return x !== y }
        static Bool m16(Str x) { return x == m10(null) }
        static Bool m17(Str x) { return x != x.size }
        static Obj m18(Str x) { return (Num)x }
        static Obj m19(Str x) { return x is Num}
        static Obj m20(Str x) { return x as Num }
        static Obj m21() { return Bar.make }

        Str f
        const static Str sf
      }

      abstract class Bar
      {
      }",
       [4, 35, "Range must be Int..Int, not 'sys::Float..sys::Int'",
        5, 35, "Range must be Int..Int, not 'sys::Int..sys::Obj[]'",
        6, 35, "Cannot apply '!' operator to 'sys::Int'",
        7, 35, "Cannot apply '&&' operator to 'sys::Int'",
        8, 35, "Cannot apply '||' operator to 'sys::Duration'",
        8, 42, "Cannot apply '||' operator to 'sys::Obj[]'",
        9, 38, "'sys::Bool' is not assignable to 'sys::Str'",
       10, 27, "Cannot call constructor 'make' on instance",
       11, 27, "Cannot call static method 'm00' on instance",
       12, 34, "Cannot call instance method 'm07' in static context",
       12, 43, "Cannot call instance method 'm07' in static context",
       13, 32, "Cannot access static field 'sf' on instance",
       14, 34, "Cannot access instance field 'f' in static context",
       14, 46, "Cannot access instance field 'f' in static context",
       15, 34, "Cannot access 'this' in static context",
       15, 44, "Cannot access 'super' in static context",
       16, 40, "Ternary condition must be Bool, not 'sys::Int'",
       17, 48, "Incomparable types 'sys::Str' and 'sys::Int'",
       18, 48, "Incomparable types 'sys::Str' and 'sys::Int'",
       19, 41, "Incomparable types 'sys::Str' and 'sys::Void'",
       20, 41, "Incomparable types 'sys::Str' and 'sys::Int'",
       21, 40, "Inconvertible types 'sys::Str' and 'sys::Num'",
       22, 40, "Inconvertible types 'sys::Str' and 'sys::Num'",
       23, 40, "Inconvertible types 'sys::Str' and 'sys::Num'",
       24, 39, "Calling constructor on abstract class",
       ])
  }

  Void testNotAssignable()
  {
    // errors
    verifyErrors("class Foo
      {

        Void m00(Int a) { 3 = a }
        Void m01(Int a) { 3 += a }
        Void m02(Int a) { i = a }
        Void m03(Int a) { i += a }
        Void m04(Int a) { i++ }
        Void m05(Foo a) { this = a }
        Void m06(Foo a) { super = a }
        Void m07(Foo a) { this += a }
        Void m08(Foo a) { this++ }

        Int i() { return null }
        Foo plus(Foo a) { return this }
        Void increment() {}
      }",
       [
         4, 27, "Left hand side is not assignable",
         5, 27, "Target is not assignable",
         6, 27, "Left hand side is not assignable",
         7, 27, "Target is not assignable",
         8, 27, "Target is not assignable",
         9, 27, "Left hand side is not assignable",
        10, 27, "Left hand side is not assignable",
        11, 27, "Target is not assignable",
        12, 27, "Target is not assignable",
       ])
  }

  Void testInvalidArgs()
  {
    // errors
    verifyErrors("class Foo
      {
        static Obj m00() { return 3.increment(true) }
        static Obj m01() { return 3.plus }
        static Obj m02() { return 3.plus(3ns) }
        static Obj m03() { return 3.plus(4, 5) }
        static Obj m04() { return sys::Str.spaces }
        static Obj m05() { return sys::Str.spaces(true) }
        static Obj m06() { return sys::Str.spaces(1, 2) }
        static Obj m07() { return \"abcb\".index(\"b\", true) }
        static Void m08() { m := |Int a| {}; m(3ns) }
        static Void m09() { m := |Str a| {}; m() }
        static Void m10() { m := &Int.plus;  m() }
        static Void m11() { (&2.plus)(true) }
        static Void m12() { (&2.plus)(3, 5) }
      }",
       [3, 37, "Invalid args increment(), not (sys::Bool)",
        4, 37, "Invalid args plus(sys::Int), not ()",
        5, 37, "Invalid args plus(sys::Int), not (sys::Duration)",
        6, 37, "Invalid args plus(sys::Int), not (sys::Int, sys::Int)",
        7, 44, "Invalid args spaces(sys::Int), not ()",
        8, 44, "Invalid args spaces(sys::Int), not (sys::Bool)",
        9, 44, "Invalid args spaces(sys::Int), not (sys::Int, sys::Int)",
       10, 42, "Invalid args index(sys::Str, sys::Int), not (sys::Str, sys::Bool)",
       11, 46, "Invalid args |sys::Int|, not (sys::Duration)",
       12, 46, "Invalid args |sys::Str|, not ()",
       13, 46, "Invalid args |sys::Int, sys::Int|, not ()",
       14, 38, "Invalid args |sys::Int|, not (sys::Bool)",
       15, 38, "Invalid args |sys::Int|, not (sys::Int, sys::Int)",
       ])
  }

  Void testExprInClosure()
  {
    // errors
    verifyErrors("class Foo                                   // 1
      {                                                       // 2
        Void m00a() { |,| { this.make }.call0 }               // 3
        Void m00b() { |,| { |,| { this.make }.call0 }.call0 } // 4
        Void m01a() { |,| { this.m02a }.call0 }               // 5
        Void m01b() { |,| { |,| { this.m02a }.call0 }.call0 } // 6
        static Void m02a() { |,| { m00a; Foo.m00a() }.call0 } // 7
        static Void m02b() { |,| { |,| { m00a; Foo.m00a() }.call0 }.call0 } // 8
        Void m03a(Str x) { |,| { this.sf.size }.call0 }       // 9
        Void m03b(Str x) { |,| { |,| { this.sf.size }.call0 }.call0 } // 10
        static Void m04a(Str x) { |,| { f.size; Foo.f.size }.call0 }
        static Void m04b(Str x) { |,| { |,| { f.size; Foo.f.size }.call0 }.call0 }

        Str f
        const static Str sf
      }",

       [3, 34, "Cannot call constructor 'make' on instance",
        4, 40, "Cannot call constructor 'make' on instance",
        5, 34, "Cannot call static method 'm02a' on instance",
        6, 40, "Cannot call static method 'm02a' on instance",
        7, 36, "Cannot call instance method 'm00a' in static context",
        7, 46, "Cannot call instance method 'm00a' in static context",
        8, 42, "Cannot call instance method 'm00a' in static context",
        8, 52, "Cannot call instance method 'm00a' in static context",
        9, 39, "Cannot access static field 'sf' on instance",
       10, 45, "Cannot access static field 'sf' on instance",
       11, 41, "Cannot access instance field 'f' in static context",
       11, 53, "Cannot access instance field 'f' in static context",
       12, 47, "Cannot access instance field 'f' in static context",
       12, 59, "Cannot access instance field 'f' in static context",
       ])
  }

  Void testAbstractSupers()
  {
    verifyErrors("
      class Foo : Base, A
      {
        override Int x { get { return super.x } set { A.super.x = val } }
        override Void n() { super.n }
        override Void m() { A.super.m() }
      }

      abstract class Base
      {
        abstract Int x
        abstract Void n()
      }

      mixin A
      {
        abstract Int x
        abstract Void m()
      }
      ",
       [
         4, 39, "Cannot use super to access abstract field '$podName::Base.x'",
         4, 55, "Cannot use super to access abstract field '$podName::A.x'",
         5, 29, "Cannot use super to call abstract method '$podName::Base.n'",
         6, 29, "Cannot use super to call abstract method '$podName::A.m'",
       ])
  }

  Void testNotStmt()
  {
    // Parser level errors
    verifyErrors("class Foo
      {
        Void x(Int i, Str s, Obj o)
        {
          i + Int;
        }
      }",
       [5, 15, "Unexpected type literal",])

    // CheckErrors level errors
    verifyErrors("class Foo
      {
        Void x(Int i, Str s, Obj o)
        {
          true;
          3;
          i + 2;
          f;
          this.f;
          (Int)o;
          o is Int;
          o as Int;
          i == 4 ? 0ns : 1ns;
          |,| {};
          i == 2;
          i === 2;
        }

        Str f
      }",

       [
         5, 11, "Not a statement",
         6, 11, "Not a statement",
         7, 11, "Not a statement",
         8, 11, "Not a statement",
         9, 16, "Not a statement",
        10, 11, "Not a statement",
        11, 11, "Not a statement",
        12, 11, "Not a statement",
        13, 11, "Not a statement",
        14, 11, "Not a statement",
        15, 11, "Not a statement",
        16, 11, "Not a statement",
       ])
  }

}