//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   19 Sep 06  Brian Frank  Creation
//

**
** MiscTest for various steps: DefaultCtor, Normalize, CheckParamDefs
**
class MiscTest : CompilerTest
{

//////////////////////////////////////////////////////////////////////////
// DefaultCtor
//////////////////////////////////////////////////////////////////////////

  Void testDefaultCtor()
  {
    compile("
      class Foo
      {
        Int x() { return 7 }
      }")
     t := pod.types.first
     mk := t.method("make")
     verifyEq(mk.isCtor, true)
     verifyEq(mk.isPublic, true)
     verifyEq(mk.params.isEmpty, true)
     verifyEq(mk.call0->x, 7)

    verifyErrors("
       class A { Void make() {} }
       class B { Int make }
       class C : Buf { } // ok
       class D { static D make() { return null } private new privateMake() { return } }
       class E : D {}
       ",
       [2, 8, "Default constructor 'make' conflicts with slot at Script(2,18)",
        3, 8, "Default constructor 'make' conflicts with slot at Script(3,18)",
        6, 8, "Default constructor 'make' conflicts with inherited slot '$podName::D.make'",
       ])
  }

//////////////////////////////////////////////////////////////////////////
// Normalize
//////////////////////////////////////////////////////////////////////////

  Void testNormalize()
  {
    compile("
      class Foo
      {
        new makeItBaby() {}
        Void x() {}
        static Void y(Int v) {}

        Int i := 6
        Int j := 7

        static { y(33) }
        const static Int k := 8
        static { y(44) }
      }

      class Bar : Foo
      {
        new make() {}

        Int g
      }")
     t := compiler.types.first

     // implicit return
     x := t.slotDef("x")->code as Block
     verifyEq(x.size, 1)
     verifyEq(x.stmts[0].id, StmtId.returnStmt)

     // instance$init
     iInit := t.slotDef("instance\$init")->code as Block
     verifyEq(iInit.size, 3)
     verifyEq(iInit.stmts[0]->expr->id, ExprId.assign)
     verifyEq(iInit.stmts[0]->expr->lhs->name, "i")
     verifyEq(iInit.stmts[0]->expr->rhs->val, 6)
     verifyEq(iInit.stmts[1]->expr->id, ExprId.assign)
     verifyEq(iInit.stmts[1]->expr->lhs->name, "j")
     verifyEq(iInit.stmts[1]->expr->rhs->val, 7)
     verifyEq(iInit.stmts[2].id, StmtId.returnStmt)

     // static$init (each one broken up into if (true) stmt)
     sInit := t.slotDef("static\$init")->code as Block
     verifyEq(sInit.size, 4)
     verifyEq(sInit.stmts[0]->trueBlock->stmts->first->expr->id, ExprId.call)
     verifyEq(sInit.stmts[0]->trueBlock->stmts->first->expr->args->get(0)->val, 33)
     verifyEq(sInit.stmts[1]->expr->id, ExprId.assign)
     verifyEq(sInit.stmts[1]->expr->lhs->name, "k")
     verifyEq(sInit.stmts[1]->expr->rhs->val, 8)
     verifyEq(sInit.stmts[2]->trueBlock->stmts->first->expr->id, ExprId.call)
     verifyEq(sInit.stmts[2]->trueBlock->stmts->first->expr->args->get(0)->val, 44)
     verifyEq(sInit.stmts[3].id, StmtId.returnStmt)

     // super ctor
     bar := compiler.types[1]
     ctor := bar.slotDef("make") as MethodDef
     verifyEq(ctor.ctorChain.target.id, ExprId.superExpr)
     verifyEq(ctor.ctorChain.method.parent.name, "Foo")
     verifyEq(ctor.ctorChain.method.name, "makeItBaby")

     // g field getter
     g := bar.slot("g") as FieldDef
     verify(bar.slotDefs.find |SlotDef s->Bool| { return s === g.get } != null)
     verifyEq(g.get.name, "g")
     verifyEq(g.get.returnType.qname, "sys::Int")
     verifyEq(g.get.params.size, 0)
     verifyEq(g.get.code.stmts.size, 1)
     verifyEq(g.get.code.stmts[0].id, StmtId.returnStmt)
     verifyEq(g.get.code.stmts[0]->expr->id, ExprId.field)
     verifyEq(g.get.code.stmts[0]->expr->name, "g")

     // g field setter
     verify(bar.slotDefs.find |SlotDef s->Bool| { return s === g.set } != null)
     verifyEq(g.set.name, "g")
     verifyEq(g.set.returnType.qname, "sys::Void")
     verifyEq(g.set.params.size, 1)
     verifyEq(g.set.params[0].paramType.qname, "sys::Int")
     verifyEq(g.set.params[0].name, "val")
     verifyEq(g.set.code.stmts.size, 2)
     verifyEq(g.set.code.stmts[0].id, StmtId.expr)
     verifyEq(g.set.code.stmts[0]->expr->id, ExprId.assign)
     verifyEq(g.set.code.stmts[0]->expr->lhs->id, ExprId.field)
     verifyEq(g.set.code.stmts[0]->expr->lhs->name, "g")
     verifyEq(g.set.code.stmts[1].id, StmtId.returnStmt)
  }

//////////////////////////////////////////////////////////////////////////
// Static Init Scoping
//////////////////////////////////////////////////////////////////////////

  Void testStaticInitScoping()
  {
    // verify two static init blocks with same local
    // are given different scopes
    compile("
        class Foo
        {
          const static Int i
          static
          {
            x := 3
            i = x
          }

          const static Str s
          static
          {
            x := \"hello\"
            s = x
          }
        }
      ")

    t := pod.types[0]
    verifyEq(t.field("i").get, 3)
    verifyEq(t.field("s").get, "hello")
  }

//////////////////////////////////////////////////////////////////////////
// Field Type Inference
//////////////////////////////////////////////////////////////////////////

  Void testFieldTypeInference()
  {
   verifyErrors("
      class Foo
      {
        Void something() { a.ouch() }
        a := false
        b := 0
        c := \"hello\"
        d := d+1
        e := d { get { return d } }
      }",
      [
        5, 9, "Type inference not supported for fields",
        6, 9, "Type inference not supported for fields",
        7, 9, "Type inference not supported for fields",
        8, 9, "Type inference not supported for fields",
        9, 9, "Type inference not supported for fields",
      ])
  }

//////////////////////////////////////////////////////////////////////////
// CheckParamDefs
//////////////////////////////////////////////////////////////////////////

  Void testCheckParamDefs()
  {
    compile("
      class Foo
      {
        Void f(Int a := 0, Int b := a+1, Int c := a+2, Int d := -c) {}
      }")
     t := compiler.types.first
     f := t.slot("f") as MethodDef
     verifyEq(f.paramDefs[0].def.id, ExprId.assign)    // save to local, used by b, c
     verifyEq(f.paramDefs[1].def.id, ExprId.shortcut)  // not saved to local
     verifyEq(f.paramDefs[2].def.id, ExprId.assign)    // save to local, used by d
     verifyEq(f.paramDefs[3].def.id, ExprId.shortcut)  // not saved to local
  }

//////////////////////////////////////////////////////////////////////////
// Generic with Generic Params
//////////////////////////////////////////////////////////////////////////

  Void testGenericWithGenericParams()
  {
    compile("
      class Foo : Test
      {
        static Str it(Int[] a, Int:Str b, |Int x| c) { return a.toStr }
        Obj testIt() { return type.method(\"it\").call3([1, 2, 3], [4:4.toStr], |Int x| {}) }
      }")

     t := pod.types.first
     verifyEq(t.method("testIt").callOn(t.make, [,]), "[1, 2, 3]")
  }

//////////////////////////////////////////////////////////////////////////
// Currying
//////////////////////////////////////////////////////////////////////////

  Void testCurrying()
  {
    ["static", ""].each |Str mod|
    {
      compile("
        class Foo
        {
          $mod Method t00() { return &s0 }
          $mod Int[] t01() { m := &s0; return m() }

          $mod Method t02() { return &s1(2) }
          $mod Int[] t03() { m := &s1; return m(66) }

          $mod Method t04() { return &s2(99) }
          $mod Int[] t05() { return (&s2)(5, 6) }
          $mod Int[] t06() { m := &s2(99); return m.call1(5) }
          $mod Int[] t07(Int a, Int b) { return (&s2(a, b)).call([5, 6, 7]) }

          $mod Int[] t08(Int a, Int h) { return (&s8(a, 2, 3, 4, 5, 6, 7))(h) }
          $mod Int[] t09(Int a, Int b, Int c, Int d, Int e, Int f, Int g, Int h) { return (&s8)(a, b, c, d, e, f, g, h) }

          $mod Int[] t10(Int a, Int b, Int c, Int d, Int e, Int f, Int g, Int h, Int i, Int j) { return (&s10(a, b, c, d))(e, f, g, h, i, j) }
          $mod Int[] t11(Int a, Int b, Int c, Int d, Int e, Int f, Int g, Int h, Int i, Int j) { return (&s10(99)).call([b, c, d, e, f, g, h, i, j]) }

          $mod Int[] s0() { return Int[,] }
          $mod Int[] s1(Int a) { return [a] }
          $mod Int[] s2(Int a, Int b) { return [a, b] }
          $mod Int[] s8(Int a, Int b, Int c, Int d, Int e, Int f, Int g, Int h) { return [a, b, c, d, e, f, g, h] }
          $mod Int[] s10(Int a, Int b, Int c, Int d, Int e, Int f, Int g, Int h, Int i, Int j) { return [a, b, c, d, e, f, g, h, i, j] }
        }")

       // compiler.fpod.dump
       t := pod.types.first
       obj := t.make

       if (mod == "static")
         verifySame(obj->t00, t.method("s0"))
       verifyEq(obj->t01, Int[,])
       verify(obj->t01 != t.method("s1"))

       verifyEq(obj->t02->params->size, 0)
       verifyEq(obj->t02->returnType, Int[].type)
       verifyEq(obj->t03, [66])

       verifyEq(obj->t04->params->size, 1)
       verifyEq(obj->t04->params->first->paramType, Int.type)
       verifyEq(obj->t04->returnType, Int[].type)
       verifyEq(obj->t05, [5, 6])
       verifyEq(obj->t06, [99, 5])
       verifyEq(obj->t07(99, 88), [99, 88])

       verifyEq(obj->t08('a', 'h'), ['a', 2, 3, 4, 5, 6, 7, 'h'])
       verifyEq(obj->t09(1, 2, 3, 4, 5, 6, 7, 8), [1, 2, 3, 4, 5, 6, 7, 8])

       verifyEq(obj->s10(1, 2, 3, 4, 5, 6, 7, 8, 9, 10), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
       verifyEq(obj->t10(1, 2, 3, 4, 5, 6, 7, 8, 9, 10), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
       verifyEq(obj->t11(1, 2, 3, 4, 5, 6, 7, 8, 9, 10), [99, 2, 3, 4, 5, 6, 7, 8, 9, 10])
     }
  }

  Void testCurryingMore()
  {
    compile("
      class Foo
      {
        Obj[] t00() { m := &pass(this).x3; return m(0, 1, 2) }
        Obj[] t01(Int a) { return (&pass(this).x3(a))(1, 2) }
        static Obj[] t02(Int a, Int b, Int c, Foo foo) { m := &foo.x3(7, 8, 9); return m() }
        static Obj[] t03(Foo foo) { m := &x3; return m(foo, 4, 5, 6) }

        Obj[] t04(Int a)
        {
          return |->Obj[]|
          {
            m := &x3; return m(a, 77, 88)
          }()
        }

        Int[] t05(Int x)
        {
          a := &Int.plus
          b := &1.plus
          c := &x.plus(2)
          return [a(4, 7), b(8), c()]
        }

        static Foo pass(Foo f) { return f }

        Obj[] x3(Int a, Int b, Int c) { return [this, a, b, c] }
      }")

     // compiler.fpod.dump
     t := pod.types.first
     obj := t.make

     verifyEq(obj->t00, [obj, 0, 1, 2])
     verifyEq(obj->t01(7), [obj, 7, 1, 2])
     verifyEq(obj->t02(0, 1, 2, obj), [obj, 7, 8, 9])
     verifyEq(obj->t03(obj), [obj, 4, 5, 6])

     verifyEq(obj->t04(18), [obj, 18, 77, 88])

     verifyEq(obj->t05(3), [11, 9, 5])
  }

  Void testCurryingMixin()
  {
    compile("
      mixin Foo
      {
        static Obj[] t00() { m := &s2; return m(1, 2) }
        static Obj[] t01() { m := &s2(7); return m(8) }

        Obj[] t02() { m := &i2; return m('a', 'b') }
        Obj[] t03(Int b) { return (&pass(this).self.i2('A'))(b) }

        Foo self() { return this }
        static Foo pass(Foo foo) { return foo }

        static Obj[] s2(Int a, Int b) { return [a, b] }
        Obj[] i2(Int a, Int b) { return [a, b] }
      }

      class Bar : Foo {}
      ")

     // compiler.fpod.dump
     t := pod.types.first
     obj := pod.types[1].make

     verifyEq(t->method("t00")->call0, [1, 2])
     verifyEq(t->method("t01")->call0, [7, 8])

     verifyEq(obj->t02, ['a', 'b'])
     verifyEq(obj->t03(7), ['A', 7])
}


}