//
// Copyright (c) 2006, John Sublett
// Licensed under the Academic Free License version 3.0
//
// History:
//   29 Jun 06  John Sublett  Creation
//

**
** FantextParser parses fantext plain text source into a Node tree.
**
class FantextParser
{
  **
  ** Make a parser for the specified input stream.
  **
  new make(InStream in)
  {
    tokens = Tokenizer.make(in)
  }

  new makeForStr(Str in)
  {
    tokens = Tokenizer.make(InStream.makeForStr(in))
  }

  **
  ** Parse the input into a tree of Nodes.
  **
  Node parse()
  {
    return parseDoc()
  }

  static Void parseFile()
  {
    InStream in := File.make(Uri.parse(Sys.args[0])).in
    doc := FantextParser.make(in).parse
    doc.dump(Sys.out, 0)
    Sys.out.flush
    in.close
  }

  static Void parseFileToHtml()
  {
    in := File.make(Uri.parse(Sys.args[0])).in
    FantextParser.make(in).translate
    Sys.out.flush
    in.close
  }

  OutStream translate(OutStream out := Sys.out)
  {
    doc := parse
    doc.write(out)
    return out
  }

//////////////////////////////////////////////////////////////////////////
// Doc
//////////////////////////////////////////////////////////////////////////

  internal Node parseDoc()
  {
    doc := Doc.make

    t := tokens.next
    while (t.ttype != TokenType.eof)
    {
      if (t.ttype == TokenType.spaces)
      {
        // it might be an indented list item
        // TODO: this isn't the right check for
        // a list item
        if (tokens.peek.ttype == TokenType.dash)
          continue;
        // is it a code block?
        if ((t.lex.size === codeBlockIndent) &&
            (tokens.peek.ttype !== TokenType.eol) &&
            (tokens.peek.ttype !== TokenType.eof))
        {
          tokens.push(t)
          doc.addChild(parseCodeBlock)
        }
        // otherwise it must be something else
        else
        {
          tokens.push(t)
          doc.addChild(parseParagraph)
        }
      }
      else if (t.ttype == TokenType.dash)
      {
        if (tokens.peek.ttype == TokenType.spaces)
        {
          tokens.push(t)
          doc.addChild(parseList(false))
        }
        else
        {
          tokens.push(t)
          doc.addChild(parseParagraph)
        }
      }
      else if (t.ttype == TokenType.number)
      {
        if (tokens.peek.ttype == TokenType.dot)
        {
          tokens.push(t)
          doc.addChild(parseList(true))
        }
        else
        {
          tokens.push(t)
          doc.addChild(parseParagraph)
        }
      }
      else if (t.ttype == TokenType.bash)
      {
        tokens.push(t)
        doc.addChild(parseHeader)
      }
      else if (t.ttype == TokenType.eol)
      {
      }
      else
      {
        tokens.push(t)
        doc.addChild(parseParagraph)
      }
      t = tokens.next
    }

    return doc
  }

//////////////////////////////////////////////////////////////////////////
// Paragraph
//////////////////////////////////////////////////////////////////////////

  internal Node parseParagraph()
  {
    para := Paragraph.make
    t := tokens.next
    eolCount := 0
    while (t.ttype != TokenType.eof)
    {
      if (t.ttype === TokenType.eol)
      {
        // if the next token is indented, bail
        if (tokens.peek.ttype === TokenType.spaces)
          return para
        else
        {
          if ((++eolCount) == 2) return para
        }
      }
      else
      {
        if (eolCount != 0)
        {
          // At this point I have a multiline paragraph
          // but the previous line may have ended with
          // a hard line break.  If it didn't, I need
          // to add a space.
          if (para.children.isEmpty ||
             (para.children[-1].nodeType != NodeType.lineBreak))
            para.addChild(Text.make(" "));
        }
        eolCount = 0
        tokens.push(t)
        para.addChild(parseAny)
      }

      t = tokens.next
    }

    return para
  }

//////////////////////////////////////////////////////////////////////////
// Header
//////////////////////////////////////////////////////////////////////////

  internal Node parseHeader()
  {
    bcount := 0

    // count the bashes to determine the header level
    while (tokens.peek.ttype === TokenType.bash)
    {
      bcount++
      tokens.next
    }

    // skip any spaces before the first word
    while (tokens.peek.ttype === TokenType.spaces)
      tokens.next

    // create the header and add everything on the line to it
    // allowing for other markup
    header := Header.make(bcount)
    topType := tokens.peek.ttype
    while ((topType !== TokenType.bash) &&
           (topType !== TokenType.eol) &&
           (topType !== TokenType.eof))
    {
      header.addChild(parseAny)
      topType = tokens.peek.ttype
    }

    // skip any trailing bashes
    while (tokens.peek.ttype === TokenType.bash)
      tokens.next

    return header
  }

//////////////////////////////////////////////////////////////////////////
// CodeBlock
//////////////////////////////////////////////////////////////////////////

  internal Node parseCodeBlock()
  {
    matchSpaces(2, tokens)

    code := StrBuf.makeCapacity(256)
    while (true)
    {
      // read a line at a time verbatim
      while ((tokens.peek.ttype !== TokenType.eol) &&
             (tokens.peek.ttype !== TokenType.eof))
      {
        t := tokens.next
        if (t.ttype === TokenType.ampersand)
          code.add("&amp;")
        else if (t.ttype == TokenType.lt)
          code.add("&lt;")
        else if (t.ttype == TokenType.gt)
          code.add("&gt;")
        else
          code.add(t.lex)
      }

      eol := tokens.next
      nextLine := tokens.peek
      if ((nextLine.ttype === TokenType.spaces) &&
          (nextLine.lex.size >= 2))
      {
        code.addChar('\n')
        code.add(nextLine.lex[2..-1])
        match(TokenType.spaces)
      }
      else
        break
    }

    return CodeBlock.make(code.toStr)
  }

//////////////////////////////////////////////////////////////////////////
// InlineCode
//////////////////////////////////////////////////////////////////////////

  internal Node parseInlineCode()
  {
    match(TokenType.backtick)

    code := StrBuf.makeCapacity(256)

    // read until close verbatim
    while ((tokens.peek.ttype !== TokenType.backtick) &&
           (tokens.peek.ttype !== TokenType.eol) &&
           (tokens.peek.ttype !== TokenType.eof))
    {
      t := tokens.next
      if (t.ttype === TokenType.ampersand)
        code.add("&amp;")
      else if (t.ttype == TokenType.lt)
        code.add("&lt;")
      else if (t.ttype == TokenType.gt)
        code.add("&gt;")
      else
        code.add(t.lex)
    }

    match(TokenType.backtick)

    return InlineCode.make(code.toStr)
  }

//////////////////////////////////////////////////////////////////////////
// Any
//////////////////////////////////////////////////////////////////////////

  internal Node parseAny()
  {
    t := tokens.next
    // Word
    if (t.ttype == TokenType.word)
    {
      tokens.push(t)
      return parseText
    }
    // Maybe Break
    else if (t.ttype === TokenType.spaces)
    {
      tokens.push(t)
      if (t.lex.size < 2)
        return parseText
      else
        return lineBreak
    }
    else if (t.ttype === TokenType.lbracket)
    {
      tokens.push(t)
      return parseHyperlink
    }
    else if (t.ttype === TokenType.backtick)
    {
      tokens.push(t)
      return parseInlineCode
    }
    else
      return Text.make(t.lex)
  }

  internal Node lineBreak()
  {
    t := tokens.next
    if ((t.ttype === TokenType.spaces) &&
        (t.lex.size >= 2) &&
        (tokens.peek.ttype === TokenType.eol))
      return LineBreak.make
    else
    {
      tokens.push(t)
      return parseText
    }
  }

//////////////////////////////////////////////////////////////////////////
// Text
//////////////////////////////////////////////////////////////////////////

  **
  ** Parse normal text.
  **
  internal Node parseText()
  {
    text := StrBuf.makeCapacity(64)
    while (!notText.contains(tokens.peek.ttype))
    {
      t := tokens.next
      if (t.ttype === TokenType.spaces)
      {
        // check for line break
        if ((t.lex.size >= 2) && (tokens.peek.ttype === TokenType.eol))
        {
          tokens.push(t)
          break
        }
        else
          text.add(t.lex)
      }
      else if (t.ttype === TokenType.ampersand)
        text.add("&amp;")
      else if (t.ttype === TokenType.lt)
        text.add("&lt;")
      else if (t.ttype === TokenType.gt)
        text.add("&gt;")
      else
        text.add(t.lex)
    }
    return Text.make(text.toStr)
  }

//////////////////////////////////////////////////////////////////////////
// Hyperlink
//////////////////////////////////////////////////////////////////////////

  internal Node parseHyperlink()
  {
    Str text := null
    Str url := null
    Str title := null

    // Keep the tokens in case things don't work out as planned.
    // If the hyperlink is not formatted correctly, treat
    // it like plain text.
    save := Token[,]

    try
    {
      save.add(match(TokenType.lbracket))
      text = parseHyperlinkText(tokens, [TokenType.rbracket], save)
      save.add(match(TokenType.rbracket))
      save.add(match(TokenType.lparen))
      url = parseHyperlinkText(tokens, [TokenType.spaces, TokenType.rparen], save)
      if (tokens.peek.ttype === TokenType.spaces)
      {
        save.add(match(TokenType.spaces))
        save.add(match(TokenType.quotes))
        title = parseHyperlinkText(tokens, [TokenType.quotes], save)
        save.add(match(TokenType.quotes))
      }
      save.add(match(TokenType.rparen))

      return Hyperlink.make(text, url, title)
    }
    catch (Err e)
    {
      plain := StrBuf.makeCapacity(save.size*16)
      save.each |Token t| { plain.add(t.lex) }
      return Text.make(plain.toStr)
    }

    return null
  }

  internal Str parseHyperlinkText(Tokenizer tokens, TokenType[] terms, List save := null)
  {
    StrBuf text := StrBuf.makeCapacity(32)

    t := tokens.next
    save.add(t)
    while ((t.ttype != TokenType.eof) &&
           (t.ttype != TokenType.eol) &&
           !terms.contains(t.ttype))
    {
      text.add(t.lex)
      t = tokens.next
      save.add(t)
    }
    tokens.push(t)
    save.pop
    return text.toStr
  }

//////////////////////////////////////////////////////////////////////////
// List
//////////////////////////////////////////////////////////////////////////

  internal Node parseList(Bool ordered)
  {
    result := (Node)(ordered ? OrderedList.make : UnorderedList.make)
    while ((tokens.peek.ttype == TokenType.eol) || isListStart(tokens))
    {
      if (tokens.peek.ttype == TokenType.eol)
      {
        eol := tokens.next
        if (!isListStart(tokens))
        {
          tokens.push(eol)
          break;
        }
        else
          tokens.push(eol)
      }
      result.addChild(parseListItem)
    }

    return result
  }

//////////////////////////////////////////////////////////////////////////
// ListItem
//////////////////////////////////////////////////////////////////////////

  internal Node parseListItem()
  {
    // if preceded by blank lines then enclose
    // in paragraph, skip over the blank lines
    Bool enclose := false
    while (tokens.peek.ttype == TokenType.eol)
    {
      enclose = true
      tokens.next
    }

    matchListStart()
    result := ListItem.make
    para := Paragraph.make
    pcount := 1
    t := tokens.next
    while (t.ttype != TokenType.eof)
    {
      // add a line at a time
      while ((t.ttype != TokenType.eof) && (t.ttype != TokenType.eol))
      {
        tokens.push(t)
        para.addChild(parseAny)
        t = tokens.next
      }

      // at the end of each line, check for continuation or
      // termination of the item, termination is a new list
      // item or a blank line followed by a new paragraph
      peek := tokens.peek
      if (peek.ttype != TokenType.eol)
      {
        // if the next line is an item start
        if (isListStart(tokens))
          break
        else if (peek.ttype == TokenType.spaces)
        {
          spaces := tokens.next
          if (spaces.lex.size == 4)
          {
            // code block
          }
          else
            continue
        }
        else
        {
          t = tokens.next
          continue
        }
      }
      else
      {
        // skip empty lines
        Token eol := tokens.next
        while (tokens.peek.ttype == TokenType.eol)
          eol = tokens.next

        // blank lines between list items, enclose
        // in paragraph
        if (isListStart(tokens))
        {
          tokens.push(eol)
          enclose = true
          break
        }
        else if (tokens.peek.ttype == TokenType.spaces)
        {
          spaces := tokens.next
          if (spaces.lex.size == 4)
          {
            // code block
          }
          else
          {
            result.addChild(para)
            para = Paragraph.make
            continue
          }
        }
        else
          break
      }
    }

    if ((pcount == 1) && !enclose)
      result.takeChildren(para)
    else
      result.addChild(para)

    return result
  }

  static Bool isListStart(Tokenizer tokens)
  {
    peek := tokens.peek
    if ((peek.ttype != TokenType.dash) &&
        (peek.ttype != TokenType.number))
    {
      return false
    }
    else
    {
      maybe := tokens.next
      next := tokens.peek
      tokens.push(maybe)

      // unordered
      if ((maybe.ttype == TokenType.dash) &&
          (next.ttype == TokenType.spaces))
      {
        return true
      }
      // ordered
      else if ((maybe.ttype == TokenType.number) &&
               (next.ttype == TokenType.dot))
      {
        return true
      }
      else
        return false
    }
  }

  internal Void matchListStart()
  {
    t := tokens.next
    if ((t.ttype != TokenType.dash) && (t.ttype != TokenType.number))
      throw err("Unexpected token type: " + t.ttype)
    else
    {
      if (t.ttype == TokenType.dash)
        match(TokenType.spaces)
      else
        match(TokenType.dot)
    }
  }

//////////////////////////////////////////////////////////////////////////
// Util
//////////////////////////////////////////////////////////////////////////

  **
  ** Match the token of the specified type and return it.
  **
  Token match(TokenType ttype)
  {
    t := tokens.next
    if (t.ttype != ttype)
    {
      tokens.push(t)
      throw err("Unexpected token type: " + ttype)
    }
    return t
  }

  **
  ** Match the specified number of spaces and return the token.
  **
  Token matchSpaces(Int size, Tokenizer tokens)
  {
    t := tokens.next
    if (t.ttype !== TokenType.spaces)
    {
      tokens.push(t)
      throw err("Unexpected token type: " + t.ttype)
    }
    if (t.lex.size !== size)
    {
      tokens.push(t)
      throw err("Expecting " + size + " spaces, not " + t.lex.size + ".")
    }
    return t
  }

  Err err(Str msg)
  {
    return Err.make(msg)
  }

//////////////////////////////////////////////////////////////////////////
// Constants
//////////////////////////////////////////////////////////////////////////

  const static Int codeBlockIndent := 2

  const static TokenType[] notText :=
  [
    TokenType.asterisk,
    TokenType.backtick,
    TokenType.bash,
    TokenType.lbracket,
    TokenType.eol,
    TokenType.eof
  ].toConst

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  private Tokenizer tokens

}