//
// Copyright (c) 2006, John Sublett
// Licensed under the Academic Free License version 3.0
//
// History:
//   29 Jun 06  John Sublett  Creation
//

**
** Tokenizer - The fantext tokenizer parses fantext formatted input into
** a stream of tokens.
**
internal class Tokenizer
{
  **
  ** Simple test driver against a file.
  **
  static Void main()
  {
    in := File.make(Uri.parse(Sys.args[0])).in
    tokens := Tokenizer.make(in)

    Token t
    while ((t = tokens.next).ttype !== TokenType.eof) echo(t)
    echo(t)
    in.close
  }

  **
  ** Make a Tokenizer for the specified input.
  **
  new make(InStream in)
  {
    this.in = in
  }

  **
  ** Get the next token from the input.
  **
  Token next()
  {
    if (!pushback.isEmpty) return (Token)pushback.pop

    Int ch := in.readChar
    if (ch == null) return Token.make(TokenType.eof, null)

    if (ch.isAlpha)
    {
      in.unreadChar(ch)
      return word
    }
    else if (ch.isDigit)
    {
      in.unreadChar(ch)
      return number
    }
    else if (ch === ' ')
    {
      in.unreadChar(ch)
      return spaces
    }
    else if (ch === '*')
      return Token.make(TokenType.asterisk, "*")
    else if (ch === '&')
      return Token.make(TokenType.ampersand, "&")
    else if (ch === '`')
      return Token.make(TokenType.backtick, "`")
    else if ((ch === '\n') || (ch === '\r'))
      return endOfLine(ch)
    else if (ch === '\t')
      return Token.make(TokenType.tab, "\t")
    else if (ch === '#')
      return Token.make(TokenType.bash, "#")
    else if (ch === '-')
      return Token.make(TokenType.dash, "-")
    else if (ch === '<')
      return Token.make(TokenType.lt, "<")
    else if (ch === '>')
      return Token.make(TokenType.gt, ">")
    else if (ch === '(')
      return Token.make(TokenType.lparen, "(")
    else if (ch === ')')
      return Token.make(TokenType.rparen, ")")
    else if (ch === '[')
      return Token.make(TokenType.lbracket, "[")
    else if (ch === ']')
      return Token.make(TokenType.rbracket, "]")
    else if (ch === '\"')
      return Token.make(TokenType.quotes, "\"")
    else if (ch === '.')
      return Token.make(TokenType.dot, ".")
    else
      return Token.make(TokenType.other, ch.toChar)
  }

  **
  ** Look at the next token but leave it in the tokenizer.
  **
  Token peek()
  {
    if (pushback.isEmpty) pushback.push(next)
    return pushback.peek
  }

  **
  ** Push a token back onto the token stream.
  **
  Void push(Token t)
  {
    pushback.push(t)
  }

  **
  ** Return an end of line token but make sure
  ** to handle \n\r and \r\n pairs.
  **
  private Token endOfLine(Int ch)
  {
    x := in.readChar
    if (ch === '\n')
    {
      if (x !== '\r') in.unreadChar(x)
    }
    else if (ch === '\r')
    {
      if (x !== '\n') in.unreadChar(x)
    }

    return Token.make(TokenType.eol, "\n")
  }

  **
  ** Return a word token which is a string of letters and numbers.
  **
  private Token word()
  {
    Int ch

    while ((ch = in.readChar) != null)
    {
      if (ch.isAlphaNum)
        buf.addChar(ch)
      else
      {
        in.unreadChar(ch)
        Token t := Token.make(TokenType.word, buf.toStr)
        buf.clear
        return t
      }
    }

    Token t := Token.make(TokenType.word, buf.toStr)
    buf.clear
    return t
  }

  **
  ** Return a number token which is a string of numbers.
  **
  private Token number()
  {
    Int ch

    while ((ch = in.readChar) != null)
    {
      if (ch.isDigit)
        buf.addChar(ch)
      else
      {
        in.unreadChar(ch)
        Token t := Token.make(TokenType.number, buf.toStr)
        buf.clear
        return t
      }
    }

    Token t := Token.make(TokenType.number, buf.toStr)
    buf.clear
    return t
  }

  **
  ** Return a "spaces" token which indicates a string of
  ** consecutive spaces.  The lexeme for a "spaces" token
  ** maintains the number of spaces that were encountered.
  **
  private Token spaces()
  {
    Int ch

    while ((ch = in.readChar) != null)
    {
      if (ch === ' ')
        buf.addChar(ch)
      else
      {
        in.unreadChar(ch)
        Token t := Token.make(TokenType.spaces, buf.toStr)
        buf.clear
        return t
      }
    }

    Token t := Token.make(TokenType.spaces, buf.toStr)
    buf.clear
    return t
  }

  private InStream in
  StrBuf buf := StrBuf.make()
  Token[] pushback := Token[,]
}