//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   11 Dec 05  Brian Frank  Creation
//
package fan.sys;

/**
 * Duration
 */
public final class Duration
  extends FanObj
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public static Duration make(Int ticks) { return make(ticks.val); }
  public static Duration make(long ticks)
  {
    if (ticks == 0) return Zero;
    return new Duration(ticks);
  }

  public static Duration now()
  {
    return new Duration(System.nanoTime());
  }

  private Duration(long ticks)
  {
    this.ticks = ticks;
  }

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

  public Bool equals(Obj obj)
  {
    if (obj instanceof Duration)
      return ticks == ((Duration)obj).ticks ? Bool.True : Bool.False;
    else
      return Bool.False;
  }

  public Int compare(Obj obj)
  {
    long that = ((Duration)obj).ticks;
    if (ticks < that) return Int.LT; return ticks  == that ? Int.EQ : Int.GT;
  }

  public int hashCode()
  {
    return (int)(ticks ^ (ticks >>> 32));
  }

  public Int hash()
  {
    return Int.make(ticks);
  }

  public Int ticks()
  {
    return Int.make(ticks);
  }

  public Type type()
  {
    return Sys.DurationType;
  }

//////////////////////////////////////////////////////////////////////////
// Operators
//////////////////////////////////////////////////////////////////////////

  public Duration negate()
  {
    return make(-ticks);
  }

  public Duration plus(Duration x)
  {
    return make(ticks + x.ticks);
  }

  public Duration minus(Duration x)
  {
    return make(ticks - x.ticks);
  }

  public Duration star(Float x)
  {
    return make((long)(ticks * x.val));
  }

  public Duration slash(Float x)
  {
    return make((long)(ticks / x.val));
  }

  public Duration percent(Float x)
  {
    return make((long)(ticks % x.val));
  }

//////////////////////////////////////////////////////////////////////////
// Conversion
//////////////////////////////////////////////////////////////////////////

  public Str toStr()
  {
    if (ticks == 0) return ZeroStr;

    // if clean millisecond boundary
    long ns = ticks;
    if (ns % 1000000L == 0)
    {
      if (ns % 86400000000000L == 0)
      {
        return Str.make(ns/86400000000000L + "days");
      }
      if (ns % 3600000000000L == 0)
      {
        return Str.make(ns/3600000000000L + "hr");
      }
      if (ns % 60000000000L == 0)
      {
        return Str.make(ns/60000000000L + "min");
      }
      if (ns % 1000000000L == 0)
      {
        return Str.make(ns/1000000000L + "sec");
      }
      return Str.make(ns/1000000L + "ms");
    }

    // return in nanoseconds
    return Str.make(ns + "ns");
  }

  public Int toMillis()
  {
    return Int.make(ticks/1000000L);
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public static final Duration Zero = new Duration(0);
  public static final Str ZeroStr = Str.make("0ns");

  public final long ticks;
  private Str str;

}