//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   13 Jan 06 (Fri 13th)  Brian Frank  Creation
//
package fan.sys;

import java.io.*;
import java.lang.reflect.*;
import java.util.ArrayList;
import java.util.HashMap;
import fanx.fcode.*;
import fanx.emit.*;

/**
 * GenericType is the base class for ListType, MapType, and MethodType
 * which all support parameterization of the generic parameter types (such
 * as A-H, V, K).  Instances of GenericType are used to represent generic
 * instances (for example an instance of ListType is used to represent Str[]).
 */
public abstract class GenericType
  extends Type
{

//////////////////////////////////////////////////////////////////////////
// Variegate
//////////////////////////////////////////////////////////////////////////

  /**
   * This method is done during fcode reflect to variegate a variant type (List,
   * Map, Method).  At this point we parameterize the variance types from the
   * specified tuple binary fcode representation.
   */
   static Type variegate(Pod pod, Type type, FTuple ftuple)
   {
     Variances var = new Variances().load(pod, ftuple.val);
     if (type == Sys.ListType) return var.v.toListOf();
     throw Err.make("Not variant type: " + type).val;
   }

   static class Variances
   {
     public Variances load(Pod pod, int[] val)
     {
       int varCount = (val.length-2)/2;
       for (int i=0; i<varCount; ++i)
         load(pod, val[i*2+2], val[i*2+3]);
       return this;
     }

     public void load(Pod pod, int id, int typeRef)
     {
       Type type = pod.findType(typeRef);
       switch (id)
       {
         case 'A': a = type; break;
         case 'B': b = type; break;
         case 'C': c = type; break;
         case 'D': d = type; break;
         case 'E': e = type; break;
         case 'F': f = type; break;
         case 'G': g = type; break;
         case 'H': h = type; break;
         case 'K': k = type; break;
         case 'R': r = type; break;
         case 'V': v = type; break;
         default: throw Err.make("Unknown variance " + (char)id + " [" + id + "]").val;
       }
     }

     Type a, b, c, d, e, f, g, h, k, r, v;
   }

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  GenericType(Type base)
  {
    super(base.pod, base.name.val, base.flags, base.facets);
  }

//////////////////////////////////////////////////////////////////////////
// Type
//////////////////////////////////////////////////////////////////////////

  public abstract Type base();

  public List mixins()
  {
    return base().mixins();
  }

  public abstract Str signature();

  public final boolean isGenericInstance() { return true; }

  public boolean is(Type type)
  {
    if (type == this || type == base()) return true;
    return base().is(type);
  }

  public Map params()
  {
    if (params == null) params = makeParams();
    return params;
  }

  abstract Map makeParams();

//////////////////////////////////////////////////////////////////////////
// Reflect
//////////////////////////////////////////////////////////////////////////

  /**
   * On reflection, we parameterize the master's methods.
   */
  protected void doReflect()
  {
    // ensure master type is reflected
    Type master = base();
    master.reflect();
    List masterSlots = master.slots;

    // allocate slot data structures
    fields = new List(Sys.FieldType, master.fields.sz());
    methods = new List(Sys.MethodType, master.methods.sz());
    slots = new List(Sys.SlotType, masterSlots.sz());
    slotsByName = new HashMap(masterSlots.sz()*3);

    // parameterize master's slots
    for (int i=0; i<masterSlots.sz(); ++i)
    {
      Slot slot = (Slot)masterSlots.get(i);
      if (slot instanceof Method)
      {
        slot = parameterize((Method)slot);
        methods.add(slot);
      }
      else
      {
        fields.add(slot);
      }
      slots.add(slot);
      slotsByName.put(slot.name.val, slot);
    }
  }

  /**
   * Parameterize the specified method (if reuse if generic
   * parameterization isn't necessary).
   */
  Method parameterize(Method m)
  {
    // if not generic, short circuit and reuse original
    if (!m.isGenericMethod()) return m;

    // new signature
    Type ret;
    List params = new List(Sys.MethodParamType, m.params.sz());

    // parameterize return type
    if (m.returnType.isGenericParameter())
      ret = parameterize(m.returnType);
    else
      ret = m.returnType;

    // narrow params (or just reuse if not parameterized)
    for (int i=0; i<m.params.sz(); ++i)
    {
      MethodParam p = (MethodParam)m.params.get(i);
      if (p.paramType.isGenericParameter())
      {
        params.add(new MethodParam(p.name, parameterize(p.paramType), p.mask));
      }
      else
      {
        params.add(p);
      }
    }

    return new Method(this, m.name, m.flags, m.facets, ret, m.inheritedReturnType, params, m);
  }

  /**
   * Parameterize t, where t is a generic parameter type such as V.
   */
  final Type parameterize(Type t)
  {
    if (t instanceof ListType)
      return parameterizeListType((ListType)t);
    else if (t instanceof MethodType)
      return parameterizeMethodType((MethodType)t);
    else
      return doParameterize(t);
  }

  /**
   * Recursively parameterize the a generic list type.
   */
  final Type parameterizeListType(ListType t)
  {
    return doParameterize(t.v).toListOf();
  }

  /**
   * Recursively parameterize the params of a method type.
   */
  final MethodType parameterizeMethodType(MethodType t)
  {
    Type[] params = new Type[t.params.length];
    for (int i=0; i<params.length; ++i)
    {
      Type param = t.params[i];
      if (param.isGenericParameter()) param = doParameterize(param);
      params[i] = param;
    }

    Type ret = t.ret;
    if (ret.isGenericParameter()) ret = doParameterize(ret);

    return new MethodType(params, ret);
  }

  /**
   * Parameterize t, where t is a generic parameter type such as V.
   */
  protected abstract Type doParameterize(Type t);

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  private Map params;
}