//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   2 Dec 05  Brian Frank  Creation
//
package fan.sys;

import java.lang.ref.*;
import java.io.File;
import java.util.HashMap;
import java.util.Iterator;
import java.util.zip.*;
import fanx.fcode.*;
import fanx.emit.*;
import fanx.util.*;

/**
 * Pod is a module containing Types.  A Pod is always backed by a FPod
 * instance which defines all the definition tables.  Usually the FPod
 * is in turn backed by a FStore for the pod's zip file.  However in the
 * case of memory-only pods defined by the compiler, the fpod.store field
 * will be null.
 *
 * Pods is loaded as soon as it is constructed:
 *  1) All the types defined by the fpod are mapped into hollow Types.
 *  2) It is emitted as a Java class called "fan.{podName}.$Pod".  The
 *     emitted class is basically a manifestation of the literal tables,
 *     after which we can clear the fpod data structures.
 */
public class Pod
  extends FanObj
{

//////////////////////////////////////////////////////////////////////////
// Management
//////////////////////////////////////////////////////////////////////////

  public static Pod find(Str name) { return find(name.val, true, null); }
  public static Pod find(Str name, Bool checked) { return find(name.val, checked.val, null); }
  public static Pod find(String name, boolean checked, FPod fpod)
  {
    try
    {
      synchronized(podsByName)
      {
        SoftReference ref = (SoftReference)podsByName.get(name);
        if (ref == null || ref.get() == null)
        {
          // if fpod is non-null, then we are "creating" this pod in
          // memory direct from the compiler, otherwise we need to
          // find the pod zip file and load it's meta-data
          if (fpod == null) fpod = readFPod(name);

          // sanity check
          if (!fpod.podName.equals(name))
            throw new Exception("Mismatched pod name b/w pod.def and pod zip filename: " + fpod.podName + " != " + name);

          // create the pod and register it
          ref = new SoftReference(new Pod(fpod));
          podsByName.put(name, ref);
        }
        return (Pod)ref.get();
      }
    }
    catch (UnknownPodErr.Val e)
    {
      if (!checked) return null;
      throw e;
    }
    catch (Exception e)
    {
      e.printStackTrace();
      if (!checked) return null;
      throw UnknownPodErr.make(name, e).val;
    }
  }

  public static Pod load(InStream in)
  {
    FPod fpod = null;
    try
    {
      fpod = new FPod(null, null);
      fpod.readFully(new ZipInputStream(SysInStream.java(in)));
    }
    catch (Exception e)
    {
      throw Err.make(e).val;
    }

    String name = fpod.podName;
    synchronized(podsByName)
    {
      // check for duplicate pod name
      if (podsByName.get(name) != null)
        throw Err.make("Duplicate pod name: " + name).val;

      // create Pod and add to master table
      Pod pod = new Pod(fpod);
      podsByName.put(name, new SoftReference(pod));
      return pod;
    }
  }

  public static FPod readFPod(String name)
    throws Exception
  {
    // check that pod file even exists
    File file = new File(Sys.PodsDir, name + ".pod");
    if (!file.exists()) throw UnknownPodErr.make(name).val;

    // verify case since Windoze is case insensitive
    String actualName = file.getCanonicalFile().getName();
    actualName = actualName.substring(0, actualName.length()-4);
    if (!actualName.equals(name)) throw UnknownPodErr.make("Mismatch case: " + name + " != " + actualName).val;

    // read in the FPod tables
    FPod fpod = new FPod(name, FStore.openToRead(new java.util.zip.ZipFile(file)));
    fpod.read();
    return fpod;
  }

  public static List list()
  {
    synchronized(podsByName)
    {
      // TODO - eventually we need a faster way to load
      //  pod meta-data into memory without actually loading
      //  every pod into memory
      if (allPodsList == null)
      {
        List pods = new List(Sys.PodType);
        File[] list = Sys.PodsDir.listFiles();
        for (int i=0; i<list.length; ++i)
        {
          File f = list[i];
          String n = f.getName();
          if (!f.isDirectory() && n.endsWith(".pod"))
          {
            n = n.substring(0, n.length()-".pod".length());
            try
            {
              pods.add(find(n, true, null));
            }
            catch (Throwable e)
            {
              System.out.println("ERROR: Invalid pod file: " + f);
              e.printStackTrace();
            }
          }
        }
        allPodsList = pods.ro();
      }
      return allPodsList;
    }
  }

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  Pod(FPod fpod)
  {
    this.name = Str.make(fpod.podName);
    load(fpod);
  }

  // TODO: used by ShimPod and Sys.stubSysPod
  public Pod(String name)
  {
    this.name = Str.make(name);
  }

//////////////////////////////////////////////////////////////////////////
// Methods
//////////////////////////////////////////////////////////////////////////

  public Type type() { return Sys.PodType; }

  public Str name()  { return name; }

  public Version version()
  {
    if (version == null)
      version = Version.parse(Str.make(fpod.version));
    return version;
  }

  public List depends()
  {
    if (depends == null)
      depends = new List(Sys.DependType, fpod.depends).ro();
    return depends;
  }

  public Facets facets()
  {
    if (facets == null)
      facets = Facets.make(fpod, name.val, null, fpod.facets);
    return facets;
  }

//////////////////////////////////////////////////////////////////////////
// Types
//////////////////////////////////////////////////////////////////////////

  public List types() { return new List(Sys.TypeType, types); }

  public Type findType(Str name) { return findType(name.val, true); }
  public Type findType(Str name, Bool checked) { return findType(name.val, checked.val); }
  public Type findType(String name, boolean checked)
  {
    Type type = (Type)typesByName.get(name);
    if (type != null) return type;
    if (checked) throw UnknownTypeErr.make(this.name + "::" + name).val;
    return null;
  }

//////////////////////////////////////////////////////////////////////////
// Files
//////////////////////////////////////////////////////////////////////////

  public Map files()
  {
    if (files == null)
      files = fpod.store.podFiles();
    return files;
  }

//////////////////////////////////////////////////////////////////////////
// Conversion
//////////////////////////////////////////////////////////////////////////

  public Str toStr() { return name; }

//////////////////////////////////////////////////////////////////////////
// Load
//////////////////////////////////////////////////////////////////////////

  void load(FPod fpod)
  {
    this.fpod = fpod;
    this.typesByName = new HashMap();

    // create a hollow Type for each FType (this requires two steps,
    // because we don't necessary have all the Types created for
    // superclasses until this loop completes)
    types = new Type[fpod.types.length];
    for (int i=0; i<fpod.types.length; ++i)
    {
      // create type instance
      Type type = new Type(this, fpod.types[i]);

      // add to my data structures
      types[i] = type;
      if (typesByName.put(type.name.val, type) != null)
        throw Err.make("Invalid pod: " + name + " type already defined: " + type.name).val;
    }

    // get TypeType to use for mixin List (we need to handle case
    // when loading sys itself - and lookup within my own pod)
    Type typeType = Sys.TypeType;
    if (typeType == null)
      typeType = (Type)typesByName.get("Type");

    // now that everthing is mapped, we can fill in the super
    // class fields (unless something is wacked, this will only
    // use Types in my pod or in pods already loaded)
    for (int i=0; i<fpod.types.length; ++i)
    {
      FType ftype = fpod.types[i];
      Type type   = types[i];
      type.base   = findType(ftype.base);

      List mixins = new List(typeType, ftype.mixins.length);
      for (int j=0; j<ftype.mixins.length; ++j)
        mixins.add(findType(ftype.mixins[j]));
      type.mixins = mixins.ro();
    }

    // emit the pod class itself (which declares all constants)
    if (!name.val.equals("sys"))
      cls = FPodEmit.emitAndLoad(fpod);

    // now it is safe to clear the literal tables
    fpod.freeLiteralTables();
  }

  Type findType(int qname)
  {
    if (qname == 0xFFFF || qname == -1) return null;

    // lookup type with typeRef index
    FTypeRef ref = fpod.typeRef(qname);

    // if generic instance, then this type must be used in a method
    // signature, not type meta-data (b/c I can't subclass generic types),
    // so it's safe that my pod has been loaded and is now registered (in
    // case the generic type is parameterized via types in my pod)
    if (ref.isGenericInstance())
      return TypeParser.load(ref.sig, true, this);

    // otherwise I need to handle if I am loading my own pod, because
    // I might not yet be added to the system namespace if I'm just
    // loading my own hollow types
    String podName  = fpod.name(ref.podName);
    String typeName = fpod.name(ref.typeName);
    Pod pod = podName.equals(name.val) ? this : Pod.find(podName, true, null);
    Type type = pod.findType(typeName, false);
    if (type != null) return type;

    // handle variance types (for sys pod only)
    if (this.name.val.equals("sys"))
    {
      type = Sys.genericParameterType(typeName);
      if (type != null) return type;
    }

    // lost cause
    throw UnknownTypeErr.make(podName + "::" + typeName).val;
  }

  /*
  void shutdown()
  {
    try
    {
      if (fpod != null && fpod.store != null)
        fpod.store.close();
    }
    catch (Exception e)
    {
    }
  }
  */

  /*
  public static void dumpMem()
  {
    int numPods  = 0;
    int numFreed = 0;
    int numTypes = 0;
    Iterator it = podsByName.values().iterator();
    while (it.hasNext())
    {
      SoftReference ref = (SoftReference)it.next();
      if (ref.get() != null)
      {
        numPods++;
        numTypes += ((Pod)ref.get()).types.length;
      }
      else
      {
        numFreed++;
      }
    }
    long mem = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
    System.out.println(">>> Pods=" + numPods + "  Freed=" + numFreed + "  Types=" + numTypes + " Mem=" + mem/(1024L*1024L) + "MB");
  }
  */

//////////////////////////////////////////////////////////////////////////
// Compiler Support
//////////////////////////////////////////////////////////////////////////

  // stub is used when sys.pod is not found, so that we can stub
  // enough to do self tests and compile sys itself
  Type stub(Type type)
  {
    if (typesByName == null) typesByName = new HashMap();
    typesByName.put(type.name().val, type);
    return type;
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  static HashMap podsByName = new HashMap();
  static List allPodsList = null;

  final Str name;
  FPod fpod;
  Version version;
  List depends;
  Facets facets;
  Type[] types;
  HashMap typesByName;
  Class cls;
  Map files;

}