//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   4 Dec 05  Brian Frank  Creation
//
package fan.sys;

import java.util.HashMap;
import fanx.util.StrUtil;

/**
 * Str is a string of Unicode characters.
 */
public final class Str
  extends FanObj
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public static Str make(String val)
  {
    if (val == null) return null;
    if (val == "")   return Empty;
    return new Str(val);
  }

  public static Str makeTrim(StringBuilder s)
  {
    int start = 0;
    int end = s.length();
    while (start < end) if (Int.isSpace(s.charAt(start))) start++; else break;
    while (end > start) if (Int.isSpace(s.charAt(end-1))) end--; else break;
    return make(s.substring(start, end));
  }

  private Str(String val)
  {
    this.val = val;
  }

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

  public final Bool equals(Obj obj)
  {
    if (obj instanceof Str)
      return val.equals(((Str)obj).val) ? Bool.True : Bool.False;
    else
      return Bool.False;
  }

  public final Int compare(Obj obj)
  {
    int cmp = val.compareTo(((Str)obj).val);
    if (cmp < 0) return Int.LT;
    return cmp == 0 ? Int.EQ : Int.GT;
  }

  public final int hashCode()
  {
    return val.hashCode();
  }

  public final Int hash()
  {
    return Int.make(val.hashCode());
  }

  public final Str toStr()
  {
    return this;
  }

  public final Type type()
  {
    return Sys.StrType;
  }

//////////////////////////////////////////////////////////////////////////
// Operators
//////////////////////////////////////////////////////////////////////////

  public final Int get(Int index)
  {
    int i = (int)index.val;
    if (i < 0) i = val.length()+i;
    return Int.pos[val.charAt(i)];
  }

  public final Str slice(Range r)
  {
    int size = val.length();

    int s = r.start(size);
    int e = r.end(size);
    if (e+1 < s) throw IndexErr.make(r).val;

    return make(val.substring(s, e+1));
  }

  public final Str plus(Obj obj)
  {
    if (obj == null) return make(val.concat("null"));
    Str x = obj.toStr();
    if (x.val == "") return this;
    return make(val.concat(x.val));
  }

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

  public final Str intern()
  {
    synchronized (interns)
    {
      Str s = (Str)interns.get(val);
      if (s == null) interns.put(val, s = this);
      return s;
    }
  }

  public final Bool isEmpty()
  {
    return val.length() == 0 ? Bool.True : Bool.False;
  }

  public final Int size()
  {
    return Int.pos(val.length());
  }

  public final Bool startsWith(Str s)
  {
    return Bool.make(val.startsWith(s.val));
  }

  public final Bool endsWith(Str s)
  {
    return Bool.make(val.endsWith(s.val));
  }

  public final Bool contains(Str s)
  {
    return index(s, Int.Zero) != null ? Bool.True : Bool.False;
  }

  public final Int index(Str s) { return index(s, Int.Zero); }
  public final Int index(Str s, Int off)
  {
    int i = (int)off.val;
    if (i < 0) i = val.length()+i;

    int r;
    String sval = s.val;
    if (sval.length() == 1)
      r = val.indexOf(sval.charAt(0), i);
    else
      r = val.indexOf(sval, i);

    if (r < 0) return null;
    return Int.make(r);
  }

  public final Int indexr(Str s) { return indexr(s, Int.NegOne); }
  public final Int indexr(Str s, Int off)
  {
    int i = (int)off.val;
    if (i < 0) i = val.length()+i;

    int r;
    String sval = s.val;
    if (sval.length() == 1)
      r = val.lastIndexOf(sval.charAt(0), i);
    else
      r = val.lastIndexOf(sval, i);

    if (r < 0) return null;
    return Int.make(r);
  }

//////////////////////////////////////////////////////////////////////////
// Iterators
//////////////////////////////////////////////////////////////////////////

  public final void each(Method m)
  {
    for (int i=0; i<val.length(); ++i)
      m.call2(Int.pos(val.charAt(i)), Int.pos(i));
  }

  public final void eachr(Method m)
  {
    for (int i=val.length()-1; i>=0; --i)
      m.call2(Int.pos(val.charAt(i)), Int.pos(i));
  }

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  public static Str spaces(Int n)
  {
    // do an array lookup for reasonable length
    // strings since that is the common case
    int count = (int)n.val;
    try { return spaces[count]; } catch (ArrayIndexOutOfBoundsException e) {}

    // otherwise we build a new one
    StringBuilder s = new StringBuilder(spaces[spaces.length-1].val);
    for (int i=spaces.length-1; i<count; ++i)
      s.append(' ');
    return Str.make(s.toString());
  }
  static Str[] spaces = new Str[20];
  static
  {
    StringBuilder s = new StringBuilder();
    for (int i=0; i<spaces.length; ++i)
    {
      spaces[i] = Str.make(s.toString());
      s.append(' ');
    }
  }

  public final Str lower() { return Str.make(lower(val)); }
  public static String lower(String val)
  {
    StringBuilder s = new StringBuilder(val.length());
    for (int i=0; i<val.length(); ++i)
    {
      int ch = val.charAt(i);
      if ('A' <= ch && ch <= 'Z') ch |= 0x20;
      s.append((char)ch);
    }
    return s.toString();
  }

  public final Str upper()
  {
    String val = this.val;
    StringBuilder s = new StringBuilder(val.length());
    for (int i=0; i<val.length(); ++i)
    {
      int ch = val.charAt(i);
      if ('a' <= ch && ch <= 'z') ch &= ~0x20;
      s.append((char)ch);
    }
    return make(s.toString());
  }

  public final Str capitalize()
  {
    String val = this.val;
    if (val.length() > 0)
    {
      int ch = val.charAt(0);
      if ('a' <= ch && ch <= 'z')
      {
        StringBuilder s = new StringBuilder(val.length());
        s.append((char)(ch & ~0x20));
        s.append(val, 1, val.length());
        return make(s.toString());
      }
    }
    return this;
  }

  public final Str decapitalize()
  {
    String val = this.val;
    if (val.length() > 0)
    {
      int ch = val.charAt(0);
      if ('A' <= ch && ch <= 'Z')
      {
        StringBuilder s = new StringBuilder(val.length());
        s.append((char)(ch | 0x20));
        s.append(val, 1, val.length());
        return make(s.toString());
      }
    }
    return this;
  }

  public final Str justl(Int width)
  {
    String val = this.val;
    int w = (int)width.val;
    if (val.length() >= w) return this;
    StringBuilder s = new StringBuilder(w);
    s.append(val);
    for (int i=val.length(); i<w; ++i)
      s.append(' ');
    return make(s.toString());
  }

  public final Str justr(Int width)
  {
    String val = this.val;
    int w = (int)width.val;
    if (val.length() >= w) return this;
    StringBuilder s = new StringBuilder(w);
    for (int i=val.length(); i<w; ++i)
      s.append(' ');
    s.append(val);
    return make(s.toString());
  }

  public final Str reverse()
  {
    String val = this.val;
    if (val.length() < 2) return this;
    StringBuilder s = new StringBuilder(val.length());
    for (int i=val.length()-1; i>=0; --i)
      s.append(val.charAt(i));
    return make(s.toString());
  }

  public final Str trim()
  {
    String val = this.val;
    int len = val.length();
    if (len == 0) return this;
    if (val.charAt(0) > ' ' && val.charAt(len-1) > ' ') return this;
    return make(val.trim());
  }

  public final List split(Str seps)
  {
    // TODO - just temp hack
    String[] split = val.split(seps.val);
    Str[] str = new Str[split.length];
    for (int i=0; i<str.length; ++i) str[i] = Str.make(split[i]);
    return new List(Sys.StrType, str);
  }

  public final Str replace(Str from, Str to)
  {
    return make(StrUtil.replace(val, from.val, to.val));
  }

//////////////////////////////////////////////////////////////////////////
// Conversion
//////////////////////////////////////////////////////////////////////////

  public final Bool toBool() { return Bool.parse(this, Bool.True); }
  public final Bool toBool(Bool checked) { return Bool.parse(this, checked); }

  public final Int toInt() { return Int.parse(this, Int.Ten, Bool.True); }
  public final Int toInt(Int radix) { return Int.parse(this, radix, Bool.True); }
  public final Int toInt(Int radix, Bool checked) { return Int.parse(this, radix, checked); }

  public final Float toFloat() { return Float.parse(this, Bool.True); }
  public final Float toFloat(Bool checked) { return Float.parse(this, checked); }

  public final Uri toUri() { return Uri.parse(this); }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public static final Str Empty = new Str("");
  public static final Str Null = new Str("null");

  private static final HashMap interns = new HashMap();

  static final Str[] ascii = new Str[128];
  static
  {
    for (int i=0; i<ascii.length; ++i)
      ascii[i] = make(String.valueOf((char)i)).intern();
  }

  public final String val;

}