//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   2 Dec 05  Brian Frank  Creation
//
package fan.sys;

import java.lang.ref.*;
import java.io.File;
import java.util.Iterator;
import fanx.fcode.*;
import fanx.util.*;

/**
 * Sys provides static access to the system's environment.
 */
public class Sys
  extends FanObj
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  private Sys() {}

//////////////////////////////////////////////////////////////////////////
// Environment
//////////////////////////////////////////////////////////////////////////

  public static List args() { return args.ro(); }

  public static Map env() { return env; }

  public static fan.sys.File homeDir() { return homeDir; }

  public static Str hostName() { return hostName; }

  public static Str userName() { return userName; }

  public static void exit() { exit(Int.Zero); }
  public static void exit(Int status) { System.exit((int)status.val); }

  public static InStream  in()  { return StdIn; }
  public static OutStream out() { return StdOut; }
  public static OutStream err() { return StdErr; }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public static final File HomeDir;
  public static final File PodsDir;
  public static final Pod  SysPod;
  public static final InStream  StdIn  = new SysInStream(System.in);
  public static final OutStream StdOut = new SysOutStream(System.out);
  public static final OutStream StdErr = new SysOutStream(System.err);

  // are we using the bootstrap stub versions of sys types
  public static boolean isStub;

  // if true then we disable FanClassLoader and all classes
  // must be available precompiled into bytecode from the
  // system's classloader (or webapp, etc); default is false
  public static boolean usePrecompiledOnly;

//////////////////////////////////////////////////////////////////////////
// Boot
//////////////////////////////////////////////////////////////////////////

  static
  {
    try
    {
      // get fan home directory
      String fanHome = System.getProperty("fan.home");
      if (fanHome == null)
      {
        fanHome = System.getenv("fan_home");
        if (fanHome == null)
          throw new RuntimeException("Missing fan.home system property or fan_home env var");
        if (!fanHome.endsWith("/"))
          fanHome += "/";
      }
      HomeDir = new File(fanHome);
      if (!HomeDir.exists() || !HomeDir.isDirectory())
        throw new RuntimeException("Invalid fan.home dir: " + HomeDir);
      PodsDir = new File(HomeDir, "lib" + File.separator + "fan");

      // check fan.usePrecompiledOnly
      usePrecompiledOnly = System.getProperty("fan.usePrecompiledOnly", "false").equals("true");

      // add shutdown hook - this is the only way to
      // cleanup pod files marked deleteOnExit
      /*
      Runtime.getRuntime().addShutdownHook(new Thread("fan.Sys Shutdown Hook") {
        public void run()
        {
          Iterator it = podsByName.values().iterator();
          while (it.hasNext()) ((Pod)it.next()).shutdown();
        }
      });
      */

      // load sys pod
      SysPod = new Pod("sys");
      Pod.podsByName.put("sys", new SoftReference(SysPod));
      try
      {
        SysPod.load(Pod.readFPod("sys"));
        isStub = false;
      }
      catch (Exception e)
      {
        System.out.println("WARNING: Using stub for 'sys' pod");
        Sys.isStub = true;
      }
    }
    catch (RuntimeException e)
    {
      System.out.println("ERROR: Sys.static");
      e.printStackTrace();
      throw e;
    }
  }

//////////////////////////////////////////////////////////////////////////
// Built-in Types
//////////////////////////////////////////////////////////////////////////

  // the Eve of all types
  public static final Type ObjType       = builtin("Obj",      null);

  // basic primitives
  public static final Type NumType       = builtin("Num",      ObjType);
  public static final Type EnumType      = builtin("Enum",     ObjType);
  public static final Type BoolType      = builtin("Bool",     ObjType);
  public static final Type DurationType  = builtin("Duration", ObjType);
  public static final Type IntType       = builtin("Int",      NumType);
  public static final Type ListType      = builtin("List",     ObjType);
  public static final Type MapType       = builtin("Map",      ObjType);
  public static final Type MonthType     = builtin("Month",    EnumType);
  public static final Type PodType       = builtin("Pod",      ObjType);
  public static final Type RangeType     = builtin("Range",    ObjType);
  public static final Type FloatType     = builtin("Float",     NumType);
  public static final Type StrType       = builtin("Str",      ObjType);
  public static final Type StrBufType    = builtin("StrBuf",   ObjType);
  public static final Type SysType       = builtin("Sys",      ObjType);
  public static final Type TestType      = builtin("Test",     ObjType);
  public static final Type ThreadType    = builtin("Thread",   ObjType);
  public static final Type TimeType      = builtin("Time",     ObjType);
  public static final Type TypeType      = builtin("Type",     ObjType);
  public static final Type WeekdayType   = builtin("Weekday",  EnumType);
  public static final Type VoidType      = builtin("Void",     ObjType);

  // reflection
  public static final Type SlotType        = builtin("Slot",        ObjType);
  public static final Type FieldType       = builtin("Field",       SlotType);
  public static final Type MethodType      = builtin("Method",      SlotType);
  public static final Type MethodParamType = builtin("MethodParam", ObjType);
  public static final Type FacetsType      = builtin("Facets",   ObjType);

  // IO
  public static final Type CharsetType      = builtin("Charset",      ObjType);
  public static final Type InStreamType     = builtin("InStream",     ObjType);
  public static final Type SysInStreamType  = builtin("SysInStream",  ObjType);
  public static final Type OutStreamType    = builtin("OutStream",    ObjType);
  public static final Type SysOutStreamType = builtin("SysOutStream", ObjType);
  public static final Type FileType         = builtin("File",         ObjType);
  public static final Type LocalFileType    = builtin("LocalFile",    FileType);
  public static final Type ZipEntryFileType = builtin("ZipEntryFile", FileType);
  public static final Type BufType          = builtin("Buf",          ObjType);
  public static final Type UriType          = builtin("Uri",          ObjType);
  public static final Type ZipType          = builtin("Zip",          ObjType);

  // utils
  public static final Type ProcessType      = builtin("Process",  ObjType);
  public static final Type DependType       = builtin("Depend",   ObjType);
  public static final Type VersionType      = builtin("Version",  ObjType);

  // exceptions
  public static final Type ErrType            = builtin("Err",            ObjType);
  public static final Type ArgErrType         = builtin("ArgErr",         ErrType);
  public static final Type CastErrType        = builtin("CastErr",        ErrType);
  public static final Type IOErrType          = builtin("IOErr",          ErrType);
  public static final Type IndexErrType       = builtin("IndexErr",       ErrType);
  public static final Type NullErrType        = builtin("NullErr",        ErrType);
  public static final Type ReadonlyErrType    = builtin("ReadonlyErr",    ErrType);
  public static final Type UnknownPodErrType  = builtin("UnknownPodErr",  ErrType);
  public static final Type UnknownSlotErrType = builtin("UnknownSlotErr", ErrType);
  public static final Type UnknownTypeErrType = builtin("UnknownTypeErr", ErrType);
  public static final Type UnsupportedErrType = builtin("UnsupportedErr", ErrType);

  // generic parameter types used with generic types List, Map, and Method
  public static final Type[] GenericParameterTypes = new Type[256];
  public static final Type AType, BType, CType, DType, EType, FType, GType,
                           HType, KType, LType, MType, RType, VType;
  static
  {
    GenericParameterTypes['A'] = AType = new Type(SysPod, "A", 0, null);  // A-H Params
    GenericParameterTypes['B'] = BType = new Type(SysPod, "B", 0, null);
    GenericParameterTypes['C'] = CType = new Type(SysPod, "C", 0, null);
    GenericParameterTypes['D'] = DType = new Type(SysPod, "D", 0, null);
    GenericParameterTypes['E'] = EType = new Type(SysPod, "E", 0, null);
    GenericParameterTypes['F'] = FType = new Type(SysPod, "F", 0, null);
    GenericParameterTypes['G'] = GType = new Type(SysPod, "G", 0, null);
    GenericParameterTypes['H'] = HType = new Type(SysPod, "H", 0, null);
    GenericParameterTypes['K'] = KType = new Type(SysPod, "K", 0, null);  // Key
    GenericParameterTypes['L'] = LType = new Type(SysPod, "L", 0, null);  // Variegated List
    GenericParameterTypes['M'] = MType = new Type(SysPod, "M", 0, null);  // Variegated Map
    GenericParameterTypes['R'] = RType = new Type(SysPod, "R", 0, null);  // Return
    GenericParameterTypes['V'] = VType = new Type(SysPod, "V", 0, null);  // Value

    List noMixins = new List(TypeType, 0).ro();
    for (int i=0; i<GenericParameterTypes.length; ++i)
    {
      Type gp = GenericParameterTypes[i];
      if (gp == null) continue;
      gp.base = ObjType;
      gp.mixins = noMixins;
    }
  }

  public static Type genericParameterType(String name)
  {
    if (name.length() == 1 && name.charAt(0) < GenericParameterTypes.length)
      return GenericParameterTypes[name.charAt(0)];
    else
      return null;
  }

  static Type builtin(String name, Type base)
  {
    if (SysPod.fpod != null) return SysPod.findType(name, true);
    return SysPod.stub(Type.stub(SysPod, name, base));
  }

  public static final List args = new List(StrType);
  public static final LocalFile homeDir = new LocalFile(HomeDir);

  public static final Str hostName;
  static
  {
    String name;
    try
    {
      name = java.net.InetAddress.getLocalHost().getHostName();
    }
    catch (Throwable e)
    {
      name = "unknown";
    }
    hostName = Str.make(name);
  }

  public static final Str userName = Str.make(System.getProperty("user.name", "unknown"));

//////////////////////////////////////////////////////////////////////////
// Env
//////////////////////////////////////////////////////////////////////////

  private static Map env = new Map(StrType, StrType);
  static
  {
    try
    {
      // environment variables
      java.util.Map getenv = System.getenv();
      Iterator it = getenv.keySet().iterator();
      while (it.hasNext())
      {
        String key = (String)it.next();
        String val = (String)getenv.get(key);
        env.set(Str.make(key), Str.make(val));
      }

      // Java system properties
      it = System.getProperties().keySet().iterator();
      while (it.hasNext())
      {
        String key = (String)it.next();
        String val = System.getProperty(key);
        env.set(Str.make(key), Str.make(val));
      }

      // sys.properties
      LocalFile f = new LocalFile(new File(HomeDir, "lib" + File.separator + "sys.props"));
      if (f.exists().val)
      {
        try
        {
        Map props = f.readProps();
        env.setAll(props);
        }
        catch (Exception e)
        {
          System.out.println("ERROR: Invalid props file: " + f);
          System.out.println("  " + e);
        }
      }
    }
    catch (Throwable e)
    {
      e.printStackTrace();
    }
    env = env.ro();
  }

}