//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   2 Dec 05  Brian Frank  Creation
//
package fan.sys;

import java.io.*;
import java.lang.reflect.*;
import java.util.ArrayList;
import java.util.HashMap;
import fanx.fcode.*;
import fanx.emit.*;
import fanx.util.*;

/**
 * Type models a static type definition for an Obj class.  A Type lifecycle:
 *
 *  1) Hollow: in this state we know basic identity of the type, and
 *     it's inheritance hierarchy.  A type is setup to be hollow during
 *     Pod.load().
 *  2) Reflected: in this state we read all the slot definitions from the
 *     fcode to populate the slot tables used to for reflection.  At this
 *     point clients can discover the signatures of the Type.
 *  3) Emitted: the final state of loading a Type is to emit to a Java
 *     class called "fan.{pod}.{type}".  Once emitted we can instantiate
 *     the type or call it's methods.
 *  4) Finished: once we have reflected the slots into memory and emitted
 *     the Java class, the last stage is to bind the all the java.lang.reflect
 *     representations to the Slots for dynamic dispatch.  We delay this
 *     until needed by Method or Field for a reflection invocation
 *
 *  A special stub state is also used to bootstrap the compiler self tests
 *  and to compile sys itself.  Stub types use java reflection to map their
 *  Java API to their Fan API (rather crudely).
 */
public class Type
  extends FanObj
{

//////////////////////////////////////////////////////////////////////////
// Management
//////////////////////////////////////////////////////////////////////////

  public static Type find(Str sig) { return TypeParser.load(sig.val, true, null); }
  public static Type find(String sig) { return TypeParser.load(sig, true, null); }
  public static Type find(Str sig, Bool checked) { return TypeParser.load(sig.val, checked.val, null); }
  public static Type find(String sig, boolean checked) { return TypeParser.load(sig, checked, null); }
  public static Type find(String podName, String typeName, boolean checked)
  {
    Pod pod = Pod.find(podName, checked, null);
    if (pod == null) return null;
    return pod.findType(typeName, checked);
  }

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  Type(Pod pod, FType ftype)
  {
    this.pod    = pod;
    this.ftype  = ftype;
    this.name   = Str.make(pod.fpod.name(pod.fpod.typeRef(ftype.self).typeName));
    this.qname  = Str.make(pod.name.val + "::" + name.val);
    this.flags  = ftype.flags;
    if (Debug) System.out.println("-- init:   " + qname);
  }

  // ShimType and parameterized type constructor
  public Type(Pod pod, String name, int flags, Facets facets)
  {
    this.pod    = pod;
    this.name   = Str.make(name);
    this.qname  = Str.make(pod.name.val + "::" + name);
    this.flags  = flags;
    this.facets = facets;
  }

//////////////////////////////////////////////////////////////////////////
// Naming
//////////////////////////////////////////////////////////////////////////

  public final Type type() { return Sys.TypeType; }

  public final Pod pod()   { return pod; }
  public final Str name()  { return name; }
  public final Str qname() { return qname; }
  public Str signature()   { return qname; }

//////////////////////////////////////////////////////////////////////////
// Flags
//////////////////////////////////////////////////////////////////////////

  public final Bool isAbstract() { return Bool.make(flags & FConst.Abstract); }
  public final Bool isClass() { return Bool.make((flags & (FConst.Enum|FConst.Mixin)) == 0); }
  public final Bool isConst() { return Bool.make(flags & FConst.Const); }
  public final Bool isEnum() { return Bool.make(flags & FConst.Enum); }
  public final Bool isFinal() { return Bool.make(flags & FConst.Final); }
  public final Bool isInternal() { return Bool.make(flags & FConst.Internal); }
  public final Bool isMixin() { return Bool.make(flags & FConst.Mixin); }
  public final Bool isPublic() { return Bool.make(flags & FConst.Public); }
  public final Bool isSynthetic() { return Bool.make(flags & FConst.Synthetic); }

//////////////////////////////////////////////////////////////////////////
// Generics
//////////////////////////////////////////////////////////////////////////

  /**
   * A generic type means that one or more of my slots contain signatures
   * using a generic parameter (such as V or K).  Fan supports three built-in
   * generic types: List, Map, and Method.  A generic instance (such as Str[])
   * is NOT a generic type (all of its generic parameters have been filled in).
   * User defined generic types are not supported in Fan.
   */
  public boolean isGenericType()
  {
    return this == Sys.ListType || this == Sys.MapType || this == Sys.MethodType;
  }

  /**
   * A generic instance is a type which has "instantiated" a generic type
   * and replaced all the generic parameter types with generic argument
   * types.  The type Str[] is a generic instance of the generic type
   * List (V is replaced with Str).  A generic instance always has a signature
   * which different from the qname.
   */
  public boolean isGenericInstance()
  {
    return false;
  }

  /**
   * Return if this type is a generic parameter (such as V or K) in a
   * generic type (List, Map, or Method).  Generic parameters serve
   * as place holders for the parameterization of the generic type.
   * Fan has a predefined set of generic parameters which are always
   * defined in the sys pod with a one character name.
   */
  public boolean isGenericParameter()
  {
    return pod == Sys.SysPod && name.val.length() == 1;
  }

  /*
   * If this type is a generic parameter (V, L, etc), then return
   * the actual type used in the Java method.  For example V is Obj,
   * and L is List.  This is the type we actually use when constructing
   * a signature for the invoke opcode.
   */
  public Type getRawType()
  {
    if (!isGenericParameter()) return this;
    if (this == Sys.LType) return Sys.ListType;
    if (this == Sys.MType) return Sys.MapType;
    if (this instanceof ListType)   return Sys.ListType;
    if (this instanceof MapType)    return Sys.MapType;
    if (this instanceof MethodType) return Sys.MethodType;
    return Sys.ObjType;
  }

  public final Bool isGeneric()
  {
    return isGenericType() ? Bool.True : Bool.False;
  }

  public Map params()
  {
    if (noParams == null)
      noParams = new Map(Sys.StrType, Sys.TypeType).ro();
    return (Map)noParams;
  }

  public final Type parameterize(Map params)
  {
    if (this == Sys.ListType)
    {
      Type v = (Type)params.get(Str.ascii['V']);
      if (v == null) throw ArgErr.make("List.parameterize - V undefined").val;
      return v.toListOf();
    }

    if (this == Sys.MapType)
    {
      Type v = (Type)params.get(Str.ascii['V']);
      Type k = (Type)params.get(Str.ascii['K']);
      if (v == null) throw ArgErr.make("Map.parameterize - V undefined").val;
      if (k == null) throw ArgErr.make("Map.parameterize - K undefined").val;
      return new MapType(k, v);
    }

    if (this == Sys.MethodType)
    {
      Type r = (Type)params.get(Str.ascii['R']);
      if (r == null) throw ArgErr.make("Map.parameterize - R undefined").val;
      ArrayList p = new ArrayList();
      for (int i='A'; i<='H'; ++i)
      {
        Type x = (Type)params.get(Str.ascii[i]);
        if (x == null) break;
        p.add(x);
      }
      return new MethodType((Type[])p.toArray(new Type[p.size()]), r);
    }

    throw UnsupportedErr.make("not generic: " + this).val;
  }

  public final synchronized Type toListOf()
  {
    if (listOf == null) listOf = new ListType(this);
    return listOf;
  }

//////////////////////////////////////////////////////////////////////////
// Slots
//////////////////////////////////////////////////////////////////////////

  public final List fields()  { return reflect().fields; }
  public final List methods() { return reflect().methods; }
  public final List slots()   { return reflect().slots; }

  public final Field field(Str name) { return (Field)slot(name.val, true); }
  public final Field field(Str name, Bool checked) { return (Field)slot(name.val, checked.val); }
  public final Field field(String name, boolean checked) { return (Field)slot(name, checked); }

  public final Method method(Str name) { return (Method)slot(name.val, true); }
  public final Method method(Str name, Bool checked) { return (Method)slot(name.val, checked.val); }
  public final Method method(String name, boolean checked) { return (Method)slot(name, checked); }

  public final Slot slot(Str name) { return slot(name.val, true); }
  public final Slot slot(Str name, Bool checked) { return slot(name.val, checked.val); }
  public final Slot slot(String name, boolean checked)
  {
    Slot slot = (Slot)reflect().slotsByName.get(name);
    if (slot != null) return slot;
    if (checked) throw UnknownSlotErr.make(this.qname + "." + name).val;
    return null;
  }

  public final Obj make() { return make(null); }
  public final Obj make(List args)
  {
    return method("make", true).call(args);
  }

//////////////////////////////////////////////////////////////////////////
// Inheritance
//////////////////////////////////////////////////////////////////////////

  public Type base() { return base; }

  public List mixins() { return mixins; }

  public List inheritance()
  {
    if (inheritance == null)
    {
      HashMap map = new HashMap();
      List acc = new List(Sys.TypeType);

      // always add Obj first
      if (this != Sys.ObjType)
      {
        map.put("sys::Obj", this);
        acc.add(Sys.ObjType);
      }

      // add myself
      map.put(qname.val, this);
      acc.add(this);

      // add my direct inheritance inheritance
      addInheritance(base(), acc, map);
      List mixins = mixins();
      for (int i=0; i<mixins.sz(); ++i)
        addInheritance((Type)mixins.get(i), acc, map);

      inheritance = acc.trim();
    }
    return inheritance;
  }

  private void addInheritance(Type t, List acc, HashMap map)
  {
    if (t == null) return;
    List ti = t.inheritance();
    for (int i=0; i<ti.sz(); ++i)
    {
      Type x = (Type)ti.get(i);
      if (map.get(x.qname.val) == null)
      {
        map.put(x.qname.val, x);
        acc.add(x);
      }
    }
  }

  public final Bool fits(Type type) { return is(type) ? Bool.True : Bool.False; }
  public boolean is(Type type)
  {
    if (this.inheritance == null) inheritance();
    List inheritance = this.inheritance;
    for (int i=0; i<inheritance.sz(); ++i)
      if (inheritance.get(i) == type) return true;
    return false;
  }

  /**
   * Given a list of types, compute the most specific type which they all share,
   * or at worst return sys::Obj.  This method does not take into account
   * interfaces, only extends class inheritance.
   */
  public static Type common(Type[] types)
  {
    if (types.length == 0) return Sys.ObjType;
    Type best = types[0];
    for (int i=1; i<types.length; ++i)
    {
      Type t = types[i];
      while (!t.is(best))
      {
        best = best.base;
        if (best == null) return Sys.ObjType;
      }
    }
    return best;
  }

//////////////////////////////////////////////////////////////////////////
// Facets
//////////////////////////////////////////////////////////////////////////

  public Facets facets()
  {
    return reflect().facets;
  }

  public Str sourceFile()
  {
    return facets().get(Str.make(FConst.SourceFile));
  }

  public Int lineNumber()
  {
    return facets().getInt(Str.make(FConst.LineNumber));
  }

//////////////////////////////////////////////////////////////////////////
// Documentation
//////////////////////////////////////////////////////////////////////////

  public Str doc()
  {
    if (!docLoaded)
    {
      try
      {
        InputStream in = pod.fpod.store.read("doc/" + name + ".fandoc");
        if (in != null)
        {
          try { FDoc.read(in); } finally { in.close(); }
        }
      }
      catch (Exception e)
      {
        e.printStackTrace();
      }
      docLoaded = true;
    }
    return doc;
  }

//////////////////////////////////////////////////////////////////////////
// Conversion
//////////////////////////////////////////////////////////////////////////

  public Str toStr() { return signature(); }

//////////////////////////////////////////////////////////////////////////
// Reflection
//////////////////////////////////////////////////////////////////////////

  protected final synchronized Type reflect()
  {
    // short circuit if already reflected
    if (slotsByName != null) return this;

    // if this is a sys stub, use java.lang.reflection
    if (isStub) { stubReflect(); return this; }

    if (Debug) System.out.println("-- reflect: " + qname + " " + slotsByName);

    // do it
    doReflect();

    // return this
    return this;
  }

  protected void doReflect()
  {
    // if the ftype is non-null, that means it was passed in non-hollow
    // ftype (in-memory compile), otherwise we need to read it from the pod
    FPod fpod   = this.pod.fpod;
    FType ftype = this.ftype;
    if (ftype.hollow)
    {
      try
      {
        ftype.read();
      }
      catch (IOException e)
      {
        e.printStackTrace();
        throw IOErr.make("Cannot read " + qname + " from pod", e).val;
      }
    }

    // these are working accumulators used to build the
    // data structures of my defined and inherited slots
    List slots  = new List(Sys.SlotType, 64);
    HashMap nameToSlot  = new HashMap();   // String -> Slot
    HashMap nameToIndex = new HashMap();   // Strnig -> Int

    // merge in base class and mixin classes
    merge(base, slots, nameToSlot, nameToIndex);
    for (int i=0; i<mixins().sz(); ++i) merge((Type)mixins.get(i), slots, nameToSlot, nameToIndex);

    // merge in all my slots
    for (int i=0; i<ftype.fields.length; ++i)
    {
      Field f = map(fpod, ftype.fields[i]);
      merge(f, slots, nameToSlot, nameToIndex);
    }
    for (int i=0; i<ftype.methods.length; ++i)
    {
      Method m = map(fpod, ftype.methods[i]);
      merge(m, slots, nameToSlot, nameToIndex);
    }

    // break out into fields and methods
    List fields  = new List(Sys.FieldType,  slots.sz());
    List methods = new List(Sys.MethodType, slots.sz());
    for (int i=0; i<slots.sz(); ++i)
    {
      Slot slot = (Slot)slots.get(i);
      if (slot instanceof Field)
        fields.add(slot);
      else
        methods.add(slot);
    }
    this.slots       = slots.trim().ro();
    this.fields      = fields.trim().ro();
    this.methods     = methods.trim().ro();
    this.slotsByName = nameToSlot;
    this.facets      = Facets.make(fpod, qname, null, ftype.facets);
  }

  /**
   * Merge the inherit's slots into my slot maps.
   *  slots:       Slot[] by order
   *  nameToSlot:  String name -> Slot
   *  nameToIndex: String name -> Int index of slots
   */
  private void merge(Type inheritedType, List slots, HashMap nameToSlot, HashMap nameToIndex)
  {
    if (inheritedType == null) return;
    List inheritedSlots = inheritedType.reflect().slots;
    for (int i=0; i<inheritedSlots.sz(); ++i)
      merge((Slot)inheritedSlots.get(i), slots, nameToSlot, nameToIndex);
  }

  /**
   * Merge the inherited slot into my slot maps.  Assume this slot
   * trumps any previous definition (because we process inheritance
   * and my slots in the right order)
   *  slots:       Slot[] by order
   *  nameToSlot:  String name -> Slot
   *  nameToIndex: String name -> Int index of slots
   */
  private void merge(Slot slot, List slots, HashMap nameToSlot, HashMap nameToIndex)
  {
    // skip constructors which aren't mine
    if (slot.isCtor().val && slot.parent != this) return;

    String name = slot.name.val;
    Int dup = (Int)nameToIndex.get(name);
    if (dup != null)
    {
      // check if this is a Getter or Setter, in which case the Field
      // trumps and we need to cache the method on the Field
      // Note: this works because we assume the compiler always generates
      // the field before the getter and setter in fcode
      if ((slot.flags & (FConst.Getter|FConst.Setter)) != 0)
      {
        Field field = (Field)slots.get(dup);
        if ((slot.flags & FConst.Getter) != 0)
          field.getter = (Method)slot;
        else
          field.setter = (Method)slot;
        return;
      }

      nameToSlot.put(name, slot);
      slots.set(dup, slot);
    }
    else
    {
      nameToSlot.put(name, slot);
      slots.add(slot);
      nameToIndex.put(name, Int.make(slots.sz()-1));
    }
  }

  /**
   * Map fcode field to a sys::Field.
   */
  private Field map(FPod fpod, FField f)
  {
    Str name = Str.make(fpod.name(f.name)).intern();
    Type fieldType = pod.findType(f.type);
    Facets facets = Facets.make(fpod, qname, name, f.facets);
    return new Field(this, name, f.flags, facets, fieldType);
  }

  /**
   * Map fcode method to a sys::Method.
   */
  private Method map(FPod fpod, FMethod m)
  {
    Str name = Str.make(fpod.name(m.name)).intern();
    Type returnType = pod.findType(m.ret);
    Type inheritedReturnType = pod.findType(m.inheritedRet);
    List params = new List(Sys.MethodParamType, m.paramCount);
    for (int j=0; j<m.paramCount; ++j)
    {
      FMethodVar p = m.vars[j];
      int pflags = (p.def == null) ? 0 : MethodParam.HAS_DEFAULT;
      params.add(new MethodParam(Str.make(fpod.name(p.name)).intern(), pod.findType(p.type), pflags));
    }
    Facets facets = Facets.make(fpod, qname, name, m.facets);
    return new Method(this, name, m.flags, facets, returnType, inheritedReturnType, params);
  }

//////////////////////////////////////////////////////////////////////////
// Emit
//////////////////////////////////////////////////////////////////////////

  /**
   * Emit to a Java class.
   */
  public synchronized Class emit()
  {
    if (cls == null)
    {
      if (Debug) System.out.println("-- emit:   " + qname);

      // make sure we have reflected to setup slots
      reflect();

      // if sys class, just load it by name
      String podName = pod.name.val;
      if (podName.equals("sys") || Sys.usePrecompiledOnly)
      {
        try
        {
          this.cls = Class.forName("fan." + podName + "." + name.val);
        }
        catch (Exception e)
        {
          e.printStackTrace();
          throw Err.make("Cannot load precompiled class: " + qname, e).val;
        }
      }

      // otherwise we need to emit it
      else
      {
        try
        {
          Class[] classes = FTypeEmit.emitAndLoad(this, ftype);
          this.cls = classes[0];
          if (classes.length > 1)
            this.auxCls = classes[1];
        }
        catch (Exception e)
        {
          e.printStackTrace();
          throw Err.make("Cannot emit: " + qname, e).val;
        }
      }

      // we are done with our ftype now, gc it
      this.ftype = null;
    }
    return cls;
  }

  /**
   * This is called to map a class if it has been precompiled
   */
  public void precompiled(Class cls)
  {
    this.cls = cls;
    try
    {
      if (isMixin().val)
        this.auxCls = cls.getClassLoader().loadClass(cls.getName()+"$");
      else if (is(Sys.ErrType))
        this.auxCls = cls.getClassLoader().loadClass(cls.getName()+"$Val");
    }
    catch (Exception e)
    {
      System.out.println("ERROR: Invalid precompiled class missing aux: " + qname);
      e.printStackTrace();
    }
  }

//////////////////////////////////////////////////////////////////////////
// Finish
//////////////////////////////////////////////////////////////////////////

  /**
   * Finish ensures we have reflected and emitted, then does
   * the final binding between slots and Java members
   */
  public synchronized void finish()
  {
    if (finished) return;
    try
    {
      // ensure reflected and emitted
      reflect();
      emit();
      finished = true;

      // map Java members to my slots for reflection; if
      // mixin then we do this for both the interface and
      // the static methods only of the implementation class
      finishSlots(cls, false);
      if (isMixin().val) finishSlots(auxCls, true);
    }
    catch (Exception e)
    {
      e.printStackTrace();
      throw Err.make("Cannot emitFinish: " + qname, e).val;
    }
  }

  /**
   * Map the Java members of the specified
   * class to my slots for reflection.
   */
  private void finishSlots(Class cls, boolean staticOnly)
  {
    // map the class's fields to my slots
    java.lang.reflect.Field[] fields = cls.getDeclaredFields();
    for (int i=0; i<fields.length; ++i)
      finishField(fields[i]);

    // map the class's methods to my slots
    java.lang.reflect.Method[] methods = cls.getDeclaredMethods();
    for (int i=0; i<methods.length; ++i)
      finishMethod(methods[i], staticOnly);
  }

  private void finishField(java.lang.reflect.Field f)
  {
    Slot slot = slot(f.getName(), false);
    if (slot == null || !(slot instanceof Field)) return;
    Field field = (Field)slot;
    if (field.reflect != null) return; // first one seems to give us most specific binding
    f.setAccessible(true);
    field.reflect = f;
  }

  private void finishMethod(java.lang.reflect.Method m, boolean staticOnly)
  {
    Slot slot = slot(m.getName(), false);
    if (slot == null) return;
    if (staticOnly && !slot.isStatic().val) return;
    m.setAccessible(true);
    if (slot instanceof Method)
    {
      Method method = (Method)slot;

      // alloc java.lang.reflect.Method[] array big enough
      // to handle all the versions with default parameters
      if (method.reflect == null)
      {
        int n = 1;
        for (int j=method.params.sz()-1; j>=0; --j)
        {
          if (((MethodParam)method.params.get(j)).hasDefault().val) n++;
          else break;
        }
        method.reflect = new java.lang.reflect.Method[n];
      }

      // get parameters, if sys we need to skip the
      // methods that use non-Fan signatures
      Class[] params = m.getParameterTypes();
      if (pod == Sys.SysPod && !checkAllFan(params)) return;

      // zero index is full signature up to using max defaults
      method.reflect[method.params.sz()-params.length] = m;
    }
    else
    {
      Field field = (Field)slot;
      if (m.getReturnType() == void.class)
        field.setter.reflect = new java.lang.reflect.Method[] { m };
      else
        field.getter.reflect = new java.lang.reflect.Method[] { m };
    }
  }

  // used by JStub only
  public FTypeEmit[] emitToClassFiles()
    throws Exception
  {
    reflect();
    return FTypeEmit.emit(this, ftype);
  }

  boolean checkAllFan(Class[] params)
  {
    for (int i=0; i<params.length; ++i)
      if (!params[i].getName().startsWith("fan."))
        return false;
    return true;
  }

//////////////////////////////////////////////////////////////////////////
// Stub
//////////////////////////////////////////////////////////////////////////

  // Stub is used when sys.pod is not found, so that we can stub
  // enough to do self tests and compile sys itself.  We only use
  // stubs for types in the sys module itself.  Since we don't have
  // fcode to use for reflect, we use real java.lang.reflection to
  // to figure out which Java methods look like Fan methods.

  static Type stub(Pod pod, String name, Type base)
  {
    Type type = new Type(pod, name, FConst.Public, null);
    type.base = base;
    type.isStub = true;
    return type;
  }

  private void stubReflect()
  {
    try
    {
      if (Debug) System.out.println("-- stub:   " + qname);
      if (name.val.equals("Obj"))
        cls = Class.forName("fan.sys.FanObj");
      else
        cls = Class.forName("fan.sys." + name);
      stubSlots(cls);
      return;
    }
    catch (Exception e)
    {
      e.printStackTrace();
      throw Err.make(e).val;
    }
  }

  protected void stubSlots(Slot[] x)
  {
    // map by name first to weed out ok duplicates
    slotsByName = new HashMap(x.length*3);
    for (int i=0; i<x.length; ++i)
    {
      Slot slot = x[i];
      Slot dup = (Slot)slotsByName.get(slot.name().val);
      if (dup != null)
      {
        // one likely reason for dups is that we have a method accessor
        // for a field - so update the field's getter/setter cache (but
        // leave the field as the primary slot in the name hashmap
        if (dup instanceof Field && slot instanceof Method)
        {
          if ((((Method)slot).flags & FConst.Getter) != 0)
            ((Field)dup).getter = (Method)slot;
          else if ((((Method)slot).flags & FConst.Setter) != 0)
            ((Field)dup).setter = (Method)slot;
          else
            throw new IllegalStateException();
          continue;
        }

        // use one with more params (we don't handle defaults, therefore
        // in stub mode the self tests must use full blown signature)
        Method a = (Method)dup;
        Method b = (Method)slot;
        if (a.params().sz() > b.params.sz())
          slot = a;
        else
          slot = b;
      }
      slotsByName.put(slot.name().val, slot);
    }

    // put the rest into lists
    fields  = new List(Sys.FieldType,  slotsByName.size());
    methods = new List(Sys.MethodType, slotsByName.size());
    slots   = new List(Sys.SlotType,   slotsByName.size());
    x = (Slot[])slotsByName.values().toArray(new Slot[slotsByName.size()]);
    for (int i=0; i<x.length; ++i)
    {
      Slot slot = x[i];
      if (slot.isField().val) fields.add(slot);
      else methods.add(slot);
      slots.add(slot);
    }
  }

  private void stubSlots(Class cls)
    throws Exception
  {
    ArrayList acc = new ArrayList();
    java.lang.reflect.Method[] methods = cls.getMethods();
    for (int i=0; i<methods.length; ++i)
    {
      try
      {
        Slot slot = stubSlot(methods[i]);
        if (slot != null) acc.add(slot);
      }
      catch (Exception e)
      {
        // skip slots which don't use pure Fan type system
      }
    }
    stubSlots((Slot[])acc.toArray(new Slot[acc.size()]));
  }

  private Slot stubSlot(java.lang.reflect.Method method)
  {
    if (method.getDeclaringClass() == Object.class) return null;
    if (method.getDeclaringClass() == Obj.class && Modifier.isStatic(method.getModifiers()) && !method.getName().equals("echo")) return null;
    return new Method(this, Str.make(method.getName()),
      stubFlags(method.getModifiers()), null,
      stubType(method.getReturnType()),
      stubType(method.getReturnType()),
      stubParams(method.getParameterTypes()));
  }

  private int stubFlags(int modifiers)
  {
    int flags = 0;
    if (Modifier.isPublic(modifiers))    flags |= FConst.Public;
    if (Modifier.isProtected(modifiers)) flags |= FConst.Protected;
    if (Modifier.isPrivate(modifiers))   flags |= FConst.Private;
    if (Modifier.isStatic(modifiers))    flags |= FConst.Static;
    if (Modifier.isAbstract(modifiers))  flags |= FConst.Abstract;
    if (!Modifier.isFinal(modifiers))    flags |= FConst.Virtual;
    return flags;
  }

  private List stubParams(Class[] cls)
  {
    List params = new List(Sys.MethodParamType, cls.length);
    for (int i=0; i<cls.length; ++i)
      params.add(new MethodParam(Str.make("p" + i), stubType(cls[i]), 0));
    return params;
  }

  private Type stubType(Class cls)
  {
    if (cls == void.class) return Sys.VoidType;
    String name = cls.getName();
    if (!name.startsWith("fan.sys.")) throw new RuntimeException();
    String typeName = name.substring("fan.sys.".length());
    return Sys.SysPod.findType(typeName, true);
  }

//////////////////////////////////////////////////////////////////////////
// Compiler Support
//////////////////////////////////////////////////////////////////////////

  public boolean isObj()   { return this == Sys.ObjType;  }
  public boolean isBool()  { return this == Sys.BoolType; }
  public boolean isInt()   { return this == Sys.IntType;  }
  public boolean isFloat() { return this == Sys.FloatType; }
  public boolean isStr()   { return this == Sys.StrType;  }
  public boolean isVoid()  { return this == Sys.VoidType; }

  public Field[] fieldArr()   { return (Field[])fields().toArray(new Field[fields.sz()]); }
  public Method[] methodArr() { return (Method[])methods().toArray(new Method[methods.sz()]); }
  public Slot[] slotArr()     { return (Slot[])slots().toArray(new Slot[slots.sz()]); }

  public void dump()
  {
    System.out.println("Type " + qname() + " extends " + base());
    for (int i=0; i<slots().sz(); ++i)
      System.out.println("  " + ((Slot)slots().get(i)).signature());
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  static final boolean Debug = false;
  static Object noParams;

  // available when hollow
  final Pod pod;
  final Str name;
  final Str qname;
  final int flags;
  Facets facets;
  Type base;
  List mixins;
  List inheritance;
  FType ftype;      // we only keep this around for memory compiles
  boolean isStub;   // sys stub type used to bootstrap compiling of sys itself
  boolean docLoaded;
  public Str doc;

  // available when reflected
  List fields;
  List methods;
  List slots;
  HashMap slotsByName;  // String:Slot

  // available when emitted
  Class cls;       // main Java class representation
  Class auxCls;    // implementation Java class if mixin/Err

  // flags to ensure we finish only once
  boolean finished;

  // misc
  Type listOf;

}