//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   28 Jun 06  Brian Frank  Creation
//
package fan.sys;

import java.net.*;

/**
 * Uri is used to immutably represent a Universal Resource Identifier.
 */
public final class Uri
  extends FanObj
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public static Uri parse(String s) { return parse(Str.make(s)); }
  public static Uri parse(Str s)
  {
    try
    {
      return new Uri(s);
    }
    catch (Exception e)
    {
      throw ArgErr.make("Invalid Uri: `" + s + "`").val;
    }
  }

  public Uri(Str s)
    throws Exception
  {
    this(s, new URI(s.val));
  }

  public Uri(URI uri)
  {
    this(Str.make(uri.toString()), uri);
  }

  private Uri(Str s, URI u)
  {
    str      = s;
    uri      = u;
    scheme   = lower(u.getScheme());
    host     = Str.make(u.getHost());
    pathStr  = Str.make(u.getPath());
    path     = parsePath(pathStr);
    queryStr = Str.make(u.getQuery());
    query    = parseQuery(queryStr);
  }

  static Str lower(String s)
  {
    if (s == null) return null;
    return Str.make(Str.lower(s));
  }

  static List parsePath(Str pathStr)
  {
    if (pathStr == null) return null;
    List list = new List(Sys.StrType);

    try
    {
      String p = pathStr.val;
      int start = 0, len = p.length();
      for (int i=0; i<len; ++i)
      {
        int ch = p.charAt(i);
        if (ch != '/') continue;
        if (start < i)
          list.add(Str.make(p.substring(start, i)));
        start = i+1;
      }
      if (start < len)
        list.add(Str.make(p.substring(start, len)));
    }
    catch (Exception e)
    {
      // don't let internal error bring down whole uri
      e.printStackTrace();
    }

    return list.trim().ro();
  }

  static Map parseQuery(Str queryStr)
  {
    if (queryStr == null) return null;
    Map map = new Map(Sys.StrType, Sys.StrType);

    try
    {
      String q = queryStr.val;
      int start = 0, eq = 0, len = q.length();
      for (int i=0; i<len; ++i)
      {
        int ch = q.charAt(i);
        if (ch == '=') eq = i;
        if (ch != '&' && ch != ';') continue;

        if (start < i)
        {
          if (start == eq)
            map.set(Str.make(q.substring(start, i)), null);
          else
            map.set(Str.make(q.substring(start, eq)), Str.make(q.substring(eq+1, i)));
        }

        start = eq = i+1;
      }

      if (start < len)
      {
        if (start == eq)
          map.set(Str.make(q.substring(start, len)), null);
        else
          map.set(Str.make(q.substring(start, eq)), Str.make(q.substring(eq+1, len)));
      }
    }
    catch (Exception e)
    {
      // don't let internal error bring down whole uri
      e.printStackTrace();
    }

    return map.ro();
  }

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

  public final Bool equals(Obj obj)
  {
    if (obj instanceof Uri)
    {
      return str.equals(((Uri)obj).str);
    }
    return Bool.False;
  }

  public int hashCode()
  {
    return str.hashCode();
  }

  public Int hash()
  {
    return str.hash();
  }

  public Str toStr()
  {
    return str;
  }

  public Type type()
  {
    return Sys.UriType;
  }

//////////////////////////////////////////////////////////////////////////
// Components
//////////////////////////////////////////////////////////////////////////

  public Bool isAbs()
  {
    return scheme != null ? Bool.True : Bool.False;
  }

  public Bool isRel()
  {
    return scheme == null ? Bool.True : Bool.False;
  }

  public Bool isDir()
  {
    if (pathStr != null)
    {
      String p = pathStr.val;
      int len = p.length();
      if (len > 0 && p.charAt(len-1) == '/')
        return Bool.True;
    }
    return Bool.False;
  }

  public Str scheme()
  {
    return scheme;
  }

  public Str body()
  {
    if (scheme == null) return str;
    return Str.make(str.val.substring(scheme.val.length()+1));
  }

  public Str auth()
  {
    if (host == null) return null;
    int port = uri.getPort();
    String userInfo = uri.getUserInfo();
    if (port < 0)
    {
      if (userInfo == null) return host;
      else return Str.make(userInfo + '@' + host.val);
    }
    else
    {
      if (userInfo == null) return Str.make(host.val + ':' + port);
      else return Str.make(userInfo + '@' + host.val + ':' + port);
    }
  }

  public Str host()
  {
    return host;
  }

  public Str userInfo()
  {
    return Str.make(uri.getUserInfo());
  }

  public Int port()
  {
    int port = uri.getPort();
    return port < 0 ? null : Int.make(port);
  }

  public List path()
  {
    return path;
  }

  public Str pathStr()
  {
    return pathStr;
  }

  public Bool isPathAbs()
  {
    if (pathStr == null || pathStr.val.length() == 0)
      return Bool.False;
    else
      return pathStr.val.charAt(0) == '/' ? Bool.True : Bool.False;
  }

  public Str name()
  {
    if (path == null || path.sz() == 0) return null;
    return (Str)path.last();
  }

  public Str basename()
  {
    Str name = name();
    if (name == null) return null;
    String n = name.val;
    int dot = n.lastIndexOf('.');
    if (dot < 2)
    {
      if (dot < 0) return name;
      if (n.equals(".")) return name;
      if (n.equals("..")) return name;
    }
    return Str.make(n.substring(0, dot));
  }

  public Str ext()
  {
    Str name = name();
    if (name == null) return null;
    String n = name.val;
    int dot = n.lastIndexOf('.');
    if (dot < 2)
    {
      if (dot < 0) return null;
      if (n.equals(".")) return null;
      if (n.equals("..")) return null;
    }
    return Str.make(n.substring(dot+1));
  }

  public Map query()
  {
    return query;
  }

  public Str queryStr()
  {
    return queryStr;
  }

  public Str frag()
  {
    return Str.make(uri.getFragment());
  }

//////////////////////////////////////////////////////////////////////////
// Manipulation
//////////////////////////////////////////////////////////////////////////

  public Uri parent()
  {
    // if no path bail
    if (path == null || path.sz() == 0) return null;

    // if just a simple filename, then no parent
    String p = pathStr.val;
    if (path.sz() == 1 && !isPathAbs().val && !isDir().val) return null;

    // build parent uri string
    StringBuilder s = new StringBuilder();
    if (scheme != null) s.append(scheme).append(':');
    if (host != null) s.append("//").append(auth());
    s.append(p.substring(0, p.lastIndexOf('/', p.length()-2)+1));

    return Uri.parse(s.toString());
  }

  public Uri normalize()
  {
    URI norm = uri.normalize();
    if (norm == uri) return this;
    return new Uri(norm);
  }

  public Uri plus(Uri join)
  {
    return new Uri(uri.resolve(join.uri));
  }

  public Uri minus(Uri base)
  {
    return new Uri(base.uri.relativize(uri));
  }

  public Uri tail() { return tail(Int.One); }
  public Uri tail(Int levels)
  {
    int lvl = (int)levels.val;
    if (path == null || path.sz() < lvl)
      throw ArgErr.make("Cannot call tail(" + lvl + ") on `" + str + "`").val;
    StringBuilder s = new StringBuilder();
    for (int i=lvl; i<path.sz(); ++i)
    {
      if (i > lvl) s.append('/');
      s.append(path.get(i));
    }
    return parse(Str.make(s.toString()));
  }

//////////////////////////////////////////////////////////////////////////
// Resolution
//////////////////////////////////////////////////////////////////////////

  public File toFile()
  {
    return File.make(this);
  }

  public Obj get() { return get(null); }
  public Obj get(Obj base)
  {
    if (scheme != null)
    {
      throw UnsupportedErr.make("Absolute Uris not supported yet: " + str).val;
    }

    if (base == null)
      throw ArgErr.make("Must specify base for `" + str + "`").val;

    return base.trapUri(this);
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  final URI uri;
  final Str str;
  final Str scheme;
  final Str host;
  final List path;
  final Str pathStr;
  final Map query;
  final Str queryStr;

}