//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   3 Sep 05  Brian Frank  Creation
//
package fanx.compiler;

import java.io.File;
import java.util.*;
import fan.sys.*;
import fanx.fcode.*;
import fanx.util.*;
import fanx.compiler.ast.*;
import fanx.compiler.steps.*;

/**
 * Compiler manages the top level process of the compiler pipeline.
 * There are a couple different "pipelines" used to accomplish
 * various twists on compiling Fan code (from memory, files, etc).
 * The pipelines are implemented as discrete CompilerSteps.
 * As the steps are executed, the Compiler instance itself stores
 * the state as we move from files -> ast -> resolved ast -> code.
 *
 * Error reporting is managed via the Compiler.errors list.  If
 * the compiler encounters problems it accumulates the errors as
 * CompileExceptions in this list, then raises the first exception
 * to the caller.  All errors go thru the CompilerSupport.err()
 * methods for logging.  To log an error and continue we simply
 * call err().  To fail fast, we code something like: throw err().
 * Or at the end of a step we may call bombIfErr() which throws the
 * first exception if any errors have accumulated.
 */
public class Compiler
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  /**
   * Construct with reasonable defaults
   */
  public Compiler()
  {
    this.log    = new Log();
    this.outDir = Sys.PodsDir;
    this.errors = new ArrayList();
  }

//////////////////////////////////////////////////////////////////////////
// Compile
//////////////////////////////////////////////////////////////////////////

  /**
   * Generate a jar/dll from fan source files inside specified source dir.
   */
  public void compile(File srcDir)
  {
    // setup
    this.srcDir = srcDir;
    log.message("Compile", srcDir);
    log.indent();

    // pipeline
    new InitShimPod(this).run();
    new FindSourceFiles(this).run();
    new ParseFiles(this).run();
    pipeline();
    new WritePod(this).run();

    // cleanup
    log.unindent();
  }

  /**
   * Dynamically compile the source file directly to a new
   * memory only pod with the specified unique name.
   */
  public Pod define(String podName, String source, String filename)
    throws Exception
  {
    // pipeline
    new InitShimPod(this, podName).run();
    pod.isScript = true;
    new ParseString(this, source, filename).run();
    pipeline();

    return Pod.find(podName, false, fpod);
  }

  /**
   * Run the standard set of compiler steps in the pipeline which
   * occur between parsing and assembler.
   */
  private void pipeline()
  {
//System.out.println("===============================");
//dumpPretty();
//dumpTree();
    new InitClosures(this).run();
    new InitShimTypes(this).run();
    new InitNameSpace(this).run();
    new DefaultCtor(this).run();
    new InitEnum(this).run();
    new ResolveImports(this).run();
    new ResolveTypeRefs(this).run();
    new TypeInference(this).run();
    new FieldAccessors(this).run();
    new InitShimExtends(this).run();
    new InitShimSlots(this).run();
    new Normalize(this).run();
    new ResolveExpr(this).run();
    new ClosureVars(this).run();
    new CheckErrors(this).run();
    new CheckMethodParamDefs(this).run();
    new Assemble(this).run();
//System.out.println("-------------------------------");
//dumpPretty();
//dumpTree();
//new FPrinter(fpod).types();
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  public void dumpPretty(Node node)
  {
    node.pretty(new AstWriter(3));
  }

  public void dumpPretty()
  {
    for (int i=0; i<classDefs.length; ++i)
      dumpPretty(classDefs[i]);
  }

  public void dumpTree()
  {
    for (int i=0; i<classDefs.length; ++i)
      dumpTree(classDefs[i]);
  }

  public void dumpTree(Node node)
  {
    node.tree(new AstWriter(3));
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public Log log;                   // set externally
  public File outDir;               // override externally
  public boolean includeDoc;        // include fandoc in output pod
  public boolean includeSrc;        // include source code in output pod
  public ArrayList errors;          // CompileException[]
  public File srcDir;               // compile()
  public ShimPod pod;               // InitShimPod
  public boolean isSys;             // InitShimPod
  public File[] srcFiles;           // FindSourceFiles
  public File[] resFiles;           // FindSourceFiles
  public CompilationUnit[] units;   // Parse
  public ClassDef[] classDefs;      // Parse
  public NameSpace ns;              // InitNameSpace
  public FPod fpod;                 // Assemble

}