//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler;

import java.io.*;
import java.util.*;
import java.util.zip.*;
import fanx.util.*;
import fanx.compiler.ast.*;

/**
 * CompilerSupport provides lots of convenience methods.
 */
public abstract class CompilerSupport
{

//////////////////////////////////////////////////////////////////////////
// Constructors
//////////////////////////////////////////////////////////////////////////

  public CompilerSupport(Compiler compiler)
  {
    this.compiler = compiler;
  }

//////////////////////////////////////////////////////////////////////////
// Abtract Syntax Tree
//////////////////////////////////////////////////////////////////////////

  public void dumpPretty(Node node)
  {
    node.pretty(new AstWriter(3));
  }

  public void dumpTree(Node node)
  {
    node.tree(new AstWriter(3));
  }

//////////////////////////////////////////////////////////////////////////
// FileUtil Wrappers
//////////////////////////////////////////////////////////////////////////

  public void delete(File f)
  {
    try
    {
      log("Delete", f);
      FileUtil.delete(f);
    }
    catch (IOException e)
    {
      throw err("Cannot delete", f, e);
    }
  }

  public void mkdir(File f)
  {
    try
    {
      if (f.exists() && f.isDirectory()) return;
      log("Makedir", f);
      FileUtil.mkdir(f);
    }
    catch (IOException e)
    {
      throw err("Cannot make dir", f, e);
    }
  }

  public void zip(File src, File zip)
  {
    try
    {
      log("Zip", zip);
      FileUtil.zip(src, zip);
    }
    catch (IOException e)
    {
      throw err("Cannot make zip", zip, e);
    }
  }

  public ZipOutputStream openZip(File zip)
  {
    try
    {
      verbose("OpenZip", zip);
      return new ZipOutputStream(new BufferedOutputStream(new FileOutputStream(zip)));
    }
    catch (IOException e)
    {
      throw err("Cannot make zip", zip, e);
    }
  }

  public void closeZip(File zip, ZipOutputStream out)
  {
    try
    {
      verbose("CloseZip", zip);
      out.close();
    }
    catch (IOException e)
    {
      throw err("Cannot close zip", zip, e);
    }
  }

//////////////////////////////////////////////////////////////////////////
// Log
//////////////////////////////////////////////////////////////////////////

  public Log log()
  {
    return compiler.log;
  }

  public boolean isVerbose()
  {
    return log().verbose;
  }

  public void log(String msg, Object src)
  {
    log().message(msg, src);
  }

  public void verbose(String msg, Object src)
  {
    log().verbose(msg, src);
  }

//////////////////////////////////////////////////////////////////////////
// Exceptions
//////////////////////////////////////////////////////////////////////////

  public CompileException err(String msg, Object file)
  {
    return err(new CompileException(msg, file));
  }

  public CompileException err(String msg, Object file, Throwable cause)
  {
    return err(new CompileException(msg, file, cause));
  }

  public CompileException err(String msg, ILocation loc)
  {
    return err(new CompileException(msg, loc));
  }

  public CompileException err(String msg, ILocation loc, Throwable cause)
  {
    return err(new CompileException(msg, loc, cause));
  }

  public CompileException err(CompileException e)
  {
    if (compiler != null)
    {
      compiler.log.error(e);
      compiler.errors.add(e);
    }
    return e;
  }

  public void bombIfErr()
  {
    if (compiler.errors.size() > 0)
      throw (CompileException)compiler.errors.get(compiler.errors.size()-1);
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public final Compiler compiler;

}