//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.ast;

import java.util.*;
import fan.sys.*;
import fanx.util.*;
import fanx.compiler.*;
import fanx.compiler.parser.*;

/**
 * CallExpr invokes a method.
 */
public class CallExpr
  extends Expr
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public CallExpr(ILocation location, Expr target, String name)
  {
    super(location);
    this.target = target;
    this.name = name;
    args = new ArrayList();
  }

  public CallExpr(ILocation location, Expr target, Method method)
  {
    this(location, target, method.name().val);
    this.method = method;
  }

  public CallExpr(ILocation location)
  {
    super(location);
    args = new ArrayList();
  }

//////////////////////////////////////////////////////////////////////////
// Access
//////////////////////////////////////////////////////////////////////////

  public Expr[] getArgs()
  {
    return (Expr[])args.toArray(new Expr[args.size()]);
  }

  public Type[] getArgTypes()
  {
    Type[] types= new Type[args.size()];
    for (int i=0; i<types.length; ++i)
      types[i] = ((Expr)args.get(i)).type;
    return types;
  }

//////////////////////////////////////////////////////////////////////////
// Expr
//////////////////////////////////////////////////////////////////////////

  public int id()
  {
    return CALL;
  }

//////////////////////////////////////////////////////////////////////////
// Tree
//////////////////////////////////////////////////////////////////////////

  protected void walkChildren(Visitor v)
  {
    target = walk(v, target);
    args = walk(v, args);
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  public final String toString()
  {
    StringBuilder s = new StringBuilder();
    if (target != null)
      s.append(target).append(isDynamic ? "->" : ".");
    s.append(name);
    s.append("(");
    for (int i=0; i<args.size(); ++i)
    {
      if (i > 0) s.append(", ");
      s.append(args.get(i));
    }
    s.append(")");
    return s.toString();
  }

  public void tree(AstWriter out)
  {
    out.w(StrUtil.getName(getClass()) + " " + name);
    if (this instanceof ShortcutExpr)
    {
      ShortcutExpr cut = (ShortcutExpr)this;
      out.w(" ").w(Token.toString(cut.opToken));
    }
    out.w(" " + leaveToString());
    out.nl();
    out.indent();
    if (target != null) { out.w("target: "); target.tree(out); }
    for (int i=0; i<args.size(); ++i)
    {
      out.w("arg[" + i + "]: ");
      ((Expr)args.get(i)).tree(out);
    }
    out.unindent();
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public Expr target;          // target containing method
  public String name;          // name of method to invoke, new if constructor
  public ArrayList args;       // Expr[] arguments to pass
  public Method method;        // resolved Method
  public boolean isDynamic;    // true if this is a -> dynamic call
  public boolean isCtorChain;  // true if this is MethodDef.ctorChain call

}