//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.ast;

import java.util.*;
import fan.sys.*;
import fanx.fcode.*;
import fanx.util.*;
import fanx.compiler.*;

/**
 * MethodDef models a method definition - signature and executable code.
 */
public class MethodDef
  extends SlotDef
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public static MethodDef makeStaticInit(Block block)
  {
    MethodDef def = new MethodDef(block);
    def.name   = "static$init";
    def.flags  = FConst.Public | FConst.Static | FConst.Synthetic;
    def.ret    = new TypeRef(block, Sys.VoidType);
    def.code   = block;
    return def;
  }

  public static MethodDef makeInstanceInit(Block block)
  {
    MethodDef def = new MethodDef(block);
    def.name   = "instance$init";
    def.flags  = FConst.Public | FConst.Synthetic;
    def.ret    = new TypeRef(block, Sys.VoidType);
    def.code   = block;
    return def;
  }

  public MethodDef(ILocation location)
  {
    super(location);
    this.params = new ArrayList(4);
    this.vars   = new ArrayList(8);
  }

//////////////////////////////////////////////////////////////////////////
// Methods
//////////////////////////////////////////////////////////////////////////

  public Slot slot() { return method; }

  /**
   * Is this a constructor (marked via new flag)
   */
  public boolean isCtor()
  {
    return (flags & FConst.Ctor) != 0;
  }

  /**
   * Return if this a static initializer block.
   */
  public boolean isStaticInit()
  {
    return name.equals("static$init");
  }

  /**
   * Return if getter/setter for FieldDef.
   */
  public boolean isFieldAccessor()
  {
    return accessorFor != null;
  }

  /**
   * Get method parameter types as array.
   */
  public Type[] getParamTypes()
  {
    Type[] types= new Type[params.size()];
    for (int i=0; i<types.length; ++i)
      types[i] = ((MethodParamDef)params.get(i)).type.type;
    return types;
  }

  /**
   * Return if this method has one or more local variables
   * which is used by a closure.
   */
  public boolean hasLocalVarUsedInClosure()
  {
    // if the method doesn't contain any closures, then easy
    if (closureCount == 0) return false;

    // walk the local vars list
    for (int i=0; i<vars.size(); ++i)
      if (((MethodVar)vars.get(i)).usedInClosure)
        return true;

    return false;
  }

//////////////////////////////////////////////////////////////////////////
// Tree
//////////////////////////////////////////////////////////////////////////

  public Node walk(Visitor v)
  {
    v.enterMethodDef(this);
    ret = walk(v, ret);
    params = walk(v, params);
    ctorChain = (CallExpr)walk(v, ctorChain);
    code = walk(v, code);
    v.exitMethodDef(this);
    return v.visit(this);
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  public void tree(AstWriter out)
  {
    out.w("MethodDef " + name + " " + StrUtil.flagsToString(flags)).nl();
    out.indent();

    if (method != null) out.w("method: " + method.signature()).nl();

    for (int i=0; i<params.size(); ++i)
    {
      out.w("params[" + i + "]: ");
      ((MethodParamDef)params.get(i)).tree(out);
    }

    if (true && vars.size() > 0)
    {
      out.w("vars:").nl();
      out.indent();
      for (int i=0; i<vars.size(); ++i)
      {
        ((MethodVar)vars.get(i)).tree(out);
        out.nl();
      }
      out.unindent();
    }

    out.w("ret: ");
    ret.tree(out);

    if (ctorChain != null) { out.w("ctorChain:"); ctorChain.tree(out);  }

    if (code != null) { out.w("code:"); code.tree(out); }

    out.unindent();
  }

  public void pretty(AstWriter out)
  {
    out.w("" + StrUtil.flagsToString(flags) + ret + " " + name + "(");
    for (int i=0; i<params.size(); ++i)
    {
      if (i > 0) out.w(", ");
      ((MethodParamDef)params.get(i)).pretty(out);
    }
    out.w(")").nl();

    if (ctorChain != null) { out.w(" : "); ctorChain.pretty(out); out.nl(); }

    if (code != null) code.pretty(out);
    out.nl();
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public TypeRef ret;          // return type
  public ArrayList params;     // MethodParamDef (name & type)
  public ArrayList vars;       // MethodVar (name, type, register, scope)
  public CallExpr ctorChain;   // this or super ctor call
  public Method method;        // resolved method
  public Block code;           // code block
  public int closureCount;     // number of closures in my block - used for unique naming by InitClosures
  public FieldDef accessorFor; // if getter/setter for specified field

}