//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.ast;

import java.util.*;
import fanx.compiler.*;

/**
 * Node is the base class of all classes which represent a node
 * in the abstract syntax tree generated by the parser.
 */
public abstract class Node
  implements ILocation
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public Node(ILocation location)
  {
    if (location == null)
      throw new IllegalArgumentException("null location");

    this.location = location;
  }

//////////////////////////////////////////////////////////////////////////
// Tree
//////////////////////////////////////////////////////////////////////////

  /**
   * Walk thru the node tree calling IVisitor.visit().
   * The walk must be children first, self last.  The
   * visitor gets a chance to replace any node via the
   * visit() return value.
   */
  public abstract Node walk(Visitor visitor);

  /**
   * Recursively walk() thru the specified node.
   */
  public static Node    walk(Visitor visitor, Node node)    { if (node == null) return null; return node.walk(visitor); }
  public static TypeRef walk(Visitor visitor, TypeRef node) { if (node == null) return null; return (TypeRef)node.walk(visitor); }
  public static Block   walk(Visitor visitor, Block node)   { if (node == null) return null; return (Block)node.walk(visitor); }
  public static Stmt    walk(Visitor visitor, Stmt node)    { if (node == null) return null; return (Stmt)node.walk(visitor); }
  public static Expr    walk(Visitor visitor, Expr node)    { if (node == null) return null; return (Expr)node.walk(visitor); }

  /**
   * Recursively walk() thru the specified list of nodes.
   */
  public static ArrayList walk(Visitor visitor, ArrayList nodes)
  {
    for (int i=0; i<nodes.size(); ++i)
    {
      Node orig = (Node)nodes.get(i);
      Node alt  = walk(visitor, orig);
      if (orig != alt) nodes.set(i, alt);
    }
    return nodes;
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  /**
   * Debug dump the tree structure format.
   */
  public abstract void tree(AstWriter out);

  /**
   * Debug dump in pretty format.
   */
  public abstract void pretty(AstWriter out);

  /**
   * Get a hash id string.
   */
  public String hashId()
  {
    return Integer.toString(System.identityHashCode(this), 36);
  }

//////////////////////////////////////////////////////////////////////////
// ILocation
//////////////////////////////////////////////////////////////////////////

  public Object file()   { return location.file(); }
  public int line()      { return location.line(); }
  public int col()       { return location.col();  }
  public String format() { return location.format();  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public static final Node[] none = new Node[0];

  public ILocation location;

}