//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.ast;

import fanx.compiler.*;
import fanx.compiler.parser.*;

/**
 * UnknownVarExpr is a place holder in the AST for a variable until
 * we can figure out what it references: field, prop, or class.
 */
public class UnknownVarExpr
  extends Expr
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public UnknownVarExpr(ILocation location, Expr target, String name)
  {
    super(location);
    this.target = target;
    this.name   = name;
  }

  public UnknownVarExpr(ILocation location, String name)
  {
    this(location, null, name);
  }

//////////////////////////////////////////////////////////////////////////
// Expr
//////////////////////////////////////////////////////////////////////////

  public int id()
  {
    return UNKNOWN_VAR;
  }

  public String toSimpleName()
  {
    if (target != null) return null;
    return name;
  }

  public TypeRef toTypeRef()
  {
    // if this {something}.id then it can't be a type ref
    if (target != null) return null;

    // hack for Java version of compiler
    if (!Parser.isTypeId(name)) return null;

    return new TypeRef(location, name);
  }

//////////////////////////////////////////////////////////////////////////
// Tree
//////////////////////////////////////////////////////////////////////////

  protected void walkChildren(Visitor v)
  {
    target = walk(v, target);
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    StringBuilder s = new StringBuilder();
    if (target != null) s.append(target).append('.');
    s.append(name).append('?');
    return s.toString();
  }

  public void tree(AstWriter out)
  {
    out.w("UnknownVar " + name).nl();
    out.indent();
    if (target != null) { out.w("target: "); target.tree(out); }
    out.unindent();
  }

  public void pretty(AstWriter out)
  {
    if (target != null) { target.pretty(out); out.w("."); }
    out.w(name);
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public Expr target;
  public String name;

}