//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.parser;

import java.io.*;
import java.util.*;
import fan.sys.*;
import fanx.fcode.*;
import fanx.util.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.ast.*;

/**
 * Parser's job in life is to process a source file into
 * it's abstract syntax tree.
 */
public class Parser
  extends CompilerSupport
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public Parser(Compiler compiler, char[] buf, String source)
  {
    super(compiler);
    this.source    = source;
    this.tokenizer = new Tokenizer(compiler, buf, source);
    this.isSys     = compiler.isSys;
  }

//////////////////////////////////////////////////////////////////////////
// Access
//////////////////////////////////////////////////////////////////////////

  /**
   * Parse the character buffer into it's abstract syntax tree.
   */
  public CompilationUnit parse()
  {
    readTokens();

    unit = new CompilationUnit(cur, compiler.pod.name().val);

    // add implicit "import fan.sys"
    if (!isSys) unit.imports.add(new Import(unit, "sys"));

    // import declarations
    imports();

    // classes
    while (curt != Token.EOF)
    {
      ClassDef def = classDef();
      if (def != null) unit.classes.add(def);
    }

    return unit;
  }

  /**
   * Read all the tokens into memory and store in the tokens field.
   */
  private void readTokens()
  {
    this.tokens = tokenizer.tokenize();
    this.numTokens = tokens.length;

    // initialize cur, peek for position 0
    reset(0);
  }

//////////////////////////////////////////////////////////////////////////
// Imports
//////////////////////////////////////////////////////////////////////////

  void imports()
  {
    while (curt == Token.USING)
    {
      Import i = new Import(cur);
      consume(Token.USING);
      i.podName = consumeId();
      endOfStmt();
      unit.imports.add(i);
    }
  }

//////////////////////////////////////////////////////////////////////////
// Class
//////////////////////////////////////////////////////////////////////////

  ClassDef classDef()
  {
    String doc = null;
    while (curt == Token.DOC_COMMENT)
      doc = (String)consume(Token.DOC_COMMENT).val;

    if (curt == Token.EOF) return null;

    Token loc = cur;

    int flags = flags();
    boolean mixin = false;
    boolean isEnum = false;

    if (curt == Token.MIXIN)
    {
      flags |= FConst.Mixin | FConst.Abstract;
      mixin = true;
      consume();
    }
    else if (curt == Token.ENUM)
    {
      flags |= FConst.Enum | FConst.Final;
      isEnum = true;
      consume();
    }
    else
    {
      consume(Token.CLASS);
    }

    String name = consumeId();

    // init class and name
    ClassDef cls = new ClassDef(loc, unit, name, flags);
    cls.doc = doc;
    if (isEnum)
      cls.base = new TypeRef(loc, Sys.EnumType);

    if (curt == Token.EXTENDS || curt == Token.MIXIN)
    {
      throw err("Must use : instead of extends or mixin", cur);
    }

    if (curt == Token.COLON)
    {
      consume();
      cls.bases.add(type());
      while (curt == Token.COMMA)
      {
        consume();
        cls.bases.add(type());
      }
    }

    // extends
    /*
    if (curt == Token.EXTENDS)
    {
      if (mixin)
      {
        cls.base = new TypeRef(loc, Sys.ObjType);
        consume();
        cls.mixins.add( type() );
        while (curt == Token.COMMA)
        {
          consume();
          cls.mixins.add( type() );
        }
      }
      else
      {
        if (isEnum) throw err("An enum can only implement mixins (no extends)", cur);
        consume();
        cls.base = type();
      }
    }
    else
    {
      if (isEnum)
        cls.base = new TypeRef(loc, Sys.EnumType);
      else if (!cls.qname.equals("sys::Obj"))
        cls.base = new TypeRef(loc, Sys.ObjType);
    }

    // mixins
    if (curt == Token.MIXIN)
    {
      if (mixin) throw err("A mixin can only extend other mixins", cur);
      consume();
      cls.mixins.add( type() );
      while (curt == Token.COMMA)
      {
        consume();
        cls.mixins.add( type() );
      }
    }
    */

    // start class body
    consume(Token.LBRACE);

    // if enum, parse values
    if (isEnum) enumDefs(cls);

    // slots
    while (curt != Token.RBRACE)
    {
      SlotDef slot = slotDef(cls);
      if (slot != null) cls.slots.add(slot);
    }

    // end of class body
    consume(Token.RBRACE);

    return cls;
  }

//////////////////////////////////////////////////////////////////////////
// Enum
//////////////////////////////////////////////////////////////////////////

  void enumDefs(ClassDef cls)
  {
    int ordinal = 0;
    cls.enums.add(enumDef(ordinal++));
    while (curt == Token.COMMA)
    {
      consume(Token.COMMA);
      cls.enums.add(enumDef(ordinal++));
    }
    endOfStmt();
  }

  EnumDef enumDef(int ordinal)
  {
    ILocation loc = cur;
    String name = consumeId();
    EnumDef def = new EnumDef(loc, ordinal, name);

    // optional ctor args
    if (curt == Token.LPAREN)
    {
      consume(Token.LPAREN);
      if (curt != Token.RPAREN)
      {
        while (true)
        {
          def.ctorArgs.add( expr() );
          if (curt == Token.RPAREN) break;
          consume(Token.COMMA);
        }
      }
      consume(Token.RPAREN);
    }

    return def;
  }

//////////////////////////////////////////////////////////////////////////
// Slots
//////////////////////////////////////////////////////////////////////////

  SlotDef slotDef(ClassDef cls)
  {
    // fan doc comment
    String doc = null;
    while (curt == Token.DOC_COMMENT)
      doc = (String)consume(Token.DOC_COMMENT).val;

    // end of class
    if (curt == Token.RBRACE) return null;

    // check for static {} class initialization
    if (curt == Token.STATIC && peekt == Token.LBRACE)
    {
      consume(Token.STATIC);
      return MethodDef.makeStaticInit(block(true));
    }

    // all members start with flags
    ILocation loc = cur;
    int flags   = flags();

    // check if this is a Java style constructor
    if (curt == Token.ID && cur.val.equals(cls.name) && peekt == Token.LPAREN)
      throw err("Invalid constructor syntax - use new keyword", loc);

    // check for inferred typed field
    if (curt == Token.ID && peekt == Token.ASSIGN) throw err("Must use := for declaration assignments", cur);
    if (curt == Token.ID && peekt == Token.DEF_ASSIGN)
    {
      String name = consumeId();
      return fieldDef(loc, doc, flags, null, name);
    }

    // check for constructor
    if ((flags & FConst.Ctor) != 0)
    {
      TypeRef type = new TypeRef(loc, Sys.VoidType);
      String name  = consumeId();
      return methodDef(loc, doc, flags, type, name);
    }

    // otherwise must be field or method
    TypeRef type  = type();
    String name   = consumeId();
    if (curt == Token.LPAREN)
    {
      return methodDef(loc, doc, flags, type, name);
    }
    else
    {
      return fieldDef(loc, doc, flags, type, name);
    }
  }

  int flags()
  {
    int flags = 0;
    boolean protection = false;
    while (true)
    {
           if (curt == Token.ABSTRACT)  { flags |= FConst.Abstract; }
      else if (curt == Token.FINAL)     { flags |= FConst.Final;    }
      else if (curt == Token.CONST)     { flags |= FConst.Const;    }
      else if (curt == Token.INTERNAL)  { flags |= FConst.Internal;  protection = true; }
      else if (curt == Token.NATIVE)    { flags |= FConst.Native;   }
      else if (curt == Token.NEW)       { flags |= FConst.Ctor;      }
      else if (curt == Token.OVERRIDE)  { flags |= FConst.Override; }
      else if (curt == Token.PRIVATE)   { flags |= FConst.Private;   protection = true; }
      else if (curt == Token.PROTECTED) { flags |= FConst.Protected; protection = true; }
      else if (curt == Token.PUBLIC)    { flags |= FConst.Public;    protection = true; }
      else if (curt == Token.READONLY)  { flags |= Readonly;  }
      else if (curt == Token.STATIC)    { flags |= FConst.Static;   }
      else if (curt == Token.VIRTUAL)   { flags |= FConst.Virtual;  }
      else break;
      consume();
    }
    if (!protection) flags |= FConst.Public;
    if ((flags & FConst.Abstract) != 0) flags |= FConst.Virtual;
    return flags;
  }

//////////////////////////////////////////////////////////////////////////
// FieldDef
//////////////////////////////////////////////////////////////////////////

  FieldDef fieldDef(ILocation loc, String doc, int flags, TypeRef type, String name)
  {
    FieldDef field = new FieldDef(loc);

    field.doc      = doc;
    field.flags    = flags;
    field.getFlags = flags;
    field.setFlags = flags;
    field.type     = type;
    field.name     = name;

    if (curt == Token.ASSIGN)
      throw err("Must use := for declaration assignments", cur);

    if (curt == Token.DEF_ASSIGN)
    {
      consume();
      field.init = expr();
    }

    // TODO need to support comma: int x, y, z

    if (curt == Token.LBRACE)
    {
      consume(Token.LBRACE);
      getOrSet(field);
      getOrSet(field);
      consume(Token.RBRACE);
    }

    // readonly is syntatic sugar for { private set }
    if ((flags & Readonly) != 0)
    {
      if (field.set != null)
        throw err("Cannot declare a setter for a readonly field");
      field.setFlags = (field.setFlags & ProtectionMask) | FConst.Private;
    }

    endOfStmt();
    return field;
  }

  private void getOrSet(FieldDef field)
  {
    int flags = getOrSetFlags(field.flags);
    if (curt == Token.ID)
    {
      // get or set
      String id = consumeId();

      // { ...block... }
      Block block = null;
      if (curt == Token.LBRACE)
        block = block(id.equals("set"));
      else
        endOfStmt();

      // map to get or set on FieldDef
      if (id.equals("get"))
      {
        field.getFlags = flags;
        field.get = block;
      }
      else if (id.equals("set"))
      {
        field.setFlags = flags;
        field.set = block;
      }
      else
      {
        throw err("Expected get or set", cur);
      }
    }
  }

  int getOrSetFlags(int defFlags)
  {
    int flags = defFlags;
    boolean protection = false;
    while (true)
    {
           if (curt == Token.INTERNAL)  { flags &= ProtectionMask; flags |= FConst.Internal; }
      else if (curt == Token.PRIVATE)   { flags &= ProtectionMask; flags |= FConst.Private; }
      else if (curt == Token.PROTECTED) { flags &= ProtectionMask; flags |= FConst.Protected; }
      else if (curt == Token.PUBLIC)    { flags &= ProtectionMask; flags |= FConst.Public; }
      else break;
      consume();
    }
    return flags;
  }

//////////////////////////////////////////////////////////////////////////
// MethodDef
//////////////////////////////////////////////////////////////////////////

  MethodDef methodDef(ILocation loc, String doc, int flags, TypeRef ret, String name)
  {
    MethodDef method = new MethodDef(loc);

    method.doc   = doc;
    method.flags = flags;
    method.ret   = ret;
    method.name  = name;

    consume(Token.LPAREN);
    if (curt != Token.RPAREN)
    {
      while (true)
      {
        MethodParamDef param = new MethodParamDef(cur);
        param.type = type();
        param.name = consumeId();
        if (curt == Token.ASSIGN) throw err("Must use := for parameter default", cur);
        if (curt == Token.DEF_ASSIGN)
        {
          consume();
          param.def = expr();
        }
        method.params.add(param);
        if (curt == Token.RPAREN) break;
        consume(Token.COMMA);
      }
    }
    consume(Token.RPAREN);

    if ((flags & FConst.Abstract) != 0 || (flags & FConst.Native) != 0 || isSys)
    {
      if (curt == Token.LBRACE) throw err("Abstract and native methods cannot have body", cur);
      endOfStmt();
    }
    else
    {
      // ctor chain
      if (((flags & FConst.Ctor) != 0) && (curt == Token.COLON))
        method.ctorChain = ctorChain(method);

      // body
      if (curt != Token.LBRACE) throw err("Expecting method body", cur);
      method.code = block(ret.isVoid());
    }

    return method;
  }

  CallExpr ctorChain(MethodDef method)
  {
    inCtorChain = true;

    consume(Token.COLON);
    ILocation loc = cur;

    CallExpr call = new CallExpr(loc);
    call.isCtorChain = true;
    switch (curt)
    {
      case Token.SUPER: call.target = superExpr(); break;
      case Token.THIS:  call.target = thisExpr(); break;
      default: throw err("Expecting this or super for constructor chaining", loc);
    }

    // we can omit name if super
    if (call.target.id() == Expr.SUPER && curt != Token.DOT)
    {
      call.name = method.name;
    }
    else
    {
      consume(Token.DOT);
      call.name = consumeId();
    }

    // TODO: omit args if pass thru
    callArgs(call);
    inCtorChain = false;
    return call;
  }

//////////////////////////////////////////////////////////////////////////
// Block & Statements
//////////////////////////////////////////////////////////////////////////

  private Block block(boolean inVoid)
  {
    this.inVoid = inVoid;
    verify(Token.LBRACE);
    return stmtOrBlock();
  }

  private Block stmtOrBlock()
  {
    Block block = new Block(cur);

    if (curt != Token.LBRACE)
    {
      block.statements.add( stmt() );
    }
    else
    {
      consume(Token.LBRACE);
      while (curt != Token.RBRACE)
        block.statements.add( stmt() );
      consume(Token.RBRACE);
    }

    return block;
  }

  private Stmt stmt()
  {
    // check for statement keywords
    switch (curt)
    {
      case Token.BREAK:    return breakStmt();
      case Token.CONTINUE: return continueStmt();
      case Token.FOR:      return forStmt();
      case Token.IF:       return ifStmt();
      case Token.RETURN:   return returnStmt();
      case Token.SWITCH:   return switchStmt();
      case Token.THROW:    return throwStmt();
      case Token.TRY:      return tryStmt();
      case Token.WHILE:    return whileStmt();
    }

    // at this point we either have an expr or local var declaration
    return exprOrLocalVarDefStmt(true);
  }

  private Stmt exprOrLocalVarDefStmt(boolean endOfStmt)
  {
    ILocation loc = cur;

    // at this point we either have an expr or local var declaration
    Expr expr = expr();

    // if the next token is an identifer and the expression could be
    // a type declaration, then we can assume this is a local variable def
    if (curt == Token.ID)
    {
      TypeRef typeRef = expr.toTypeRef();
      if (typeRef != null) return localVarDefStmt(expr.location, typeRef, consumeId(), endOfStmt);
    }

    // check for an inferred type local variable declaration
    if (curt == Token.ASSIGN) throw err("Must use := for declaration assignments", cur);
    if (curt == Token.DEF_ASSIGN)
    {
      String name = expr.toSimpleName();
      if (name == null) throw err("Expecting name for local variable declaration", loc);
      return localVarDefStmt(expr.location, null, name, endOfStmt);
    }

    // otherwise assume it's a stand alone expression statement
    if (endOfStmt) endOfStmt();
    return new ExprStmt(expr.location, expr);
  }

  private IfStmt ifStmt()
  {
    IfStmt stmt = new IfStmt(cur);
    consume(Token.IF);
    consume(Token.LPAREN);
    stmt.condition = expr();
    consume(Token.RPAREN);
    stmt.trueBlock = stmtOrBlock();
    if (curt == Token.ELSE)
    {
      consume(Token.ELSE);
      stmt.falseBlock = stmtOrBlock();
    }
    return stmt;
  }

  private ReturnStmt returnStmt()
  {
    ReturnStmt stmt = new ReturnStmt(cur);
    consume(Token.RETURN);
    if (!inVoid && curt != Token.SEMICOLON && curt != Token.RBRACE) stmt.expr = expr();
    endOfStmt();
    return stmt;
  }

  private ThrowStmt throwStmt()
  {
    ThrowStmt stmt = new ThrowStmt(cur);
    consume(Token.THROW);
    stmt.exception = expr();
    endOfStmt();
    return stmt;
  }

  private WhileStmt whileStmt()
  {
    WhileStmt stmt = new WhileStmt(cur);
    consume(Token.WHILE);
    consume(Token.LPAREN);
    stmt.condition = expr();
    consume(Token.RPAREN);
    stmt.block = stmtOrBlock();
    return stmt;
  }

  private ForStmt forStmt()
  {
    ForStmt stmt = new ForStmt (cur);
    consume(Token.FOR);
    consume(Token.LPAREN);

    if (curt != Token.SEMICOLON) stmt.init = exprOrLocalVarDefStmt(false);
    consume(Token.SEMICOLON);

    if (curt != Token.SEMICOLON) stmt.condition = expr();
    consume(Token.SEMICOLON);

    if (curt != Token.RPAREN) stmt.update = expr();
    consume(Token.RPAREN);

    stmt.block = stmtOrBlock();
    return stmt;
  }

  private BreakStmt breakStmt()
  {
    BreakStmt stmt = new BreakStmt(cur);
    consume(Token.BREAK);
    endOfStmt();
    return stmt;
  }

  private ContinueStmt continueStmt()
  {
    ContinueStmt stmt = new ContinueStmt(cur);
    consume(Token.CONTINUE);
    endOfStmt();
    return stmt;
  }

  private TryStmt tryStmt()
  {
    TryStmt stmt = new TryStmt(cur);
    consume(Token.TRY);
    verify(Token.LBRACE);
    stmt.block = block(inVoid);
    if (curt != Token.CATCH && curt != Token.FINALLY)
      throw err("Expecting catch or finally block", cur);
    while (curt == Token.CATCH)
    {
      stmt.catches.add(tryCatch());
    }
    if (curt == Token.FINALLY)
    {
      consume();
      verify(Token.LBRACE);
      stmt.finallyBlock = block(inVoid);
    }
    return stmt;
  }

  private TryCatch tryCatch()
  {
    TryCatch c = new TryCatch(cur);
    consume(Token.CATCH);

    if (curt == Token.LPAREN)
    {
      consume(Token.LPAREN);
      c.errType = type();
      c.errVariable = consumeId();
      consume(Token.RPAREN);
    }

    verify(Token.LBRACE);
    c.block = block(inVoid);

    // insert implicit local variable declaration
    if (c.errVariable != null)
    {
      LocalVarDefStmt stmt = new LocalVarDefStmt(c.location, c.errType, c.errVariable, null);
      stmt.catchVar = true;
      c.block.add(0, stmt);
    }

    return c;
  }

  private SwitchStmt switchStmt()
  {
    SwitchStmt stmt = new SwitchStmt(cur);
    consume(Token.SWITCH);
    consume(Token.LPAREN);
    stmt.condition = expr();
    consume(Token.RPAREN);
    consume(Token.LBRACE);
    while (curt != Token.RBRACE)
    {
      if (curt == Token.CASE)
      {
        SwitchCase c = new SwitchCase(cur);
        while (curt == Token.CASE)
        {
          consume();
          c.cases.add(termExpr());
          consume(Token.COLON);
        }
        c.block = switchBlock();
        stmt.cases.add(c);
      }
      else if (curt == Token.DEFAULT)
      {
        consume();
        consume(Token.COLON);
        stmt.defaultBlock = switchBlock();
      }
      else
      {
        throw err("Expected case or default statement");
      }
    }
    consume(Token.RBRACE);
    endOfStmt();
    return stmt;
  }

  private Block switchBlock()
  {
    Block block = null;
    while (curt != Token.CASE && curt != Token.DEFAULT && curt != Token.RBRACE)
    {
      if (block == null) block = new Block(cur);
      block.statements.add( stmt() );
    }
    return block;
  }

  private LocalVarDefStmt localVarDefStmt(ILocation loc, TypeRef type, String name, boolean endOfStmt)
  {
    LocalVarDefStmt stmt = new LocalVarDefStmt(loc);
    stmt.type = type;
    stmt.name = name;

    if (curt == Token.ASSIGN)
      throw err("Must use := for declaration assignments", cur);

    if (curt == Token.DEF_ASSIGN)
    {
      consume();
      stmt.init = expr();
    }

    if (endOfStmt) endOfStmt();   // TODO: int x, y, ...
    return stmt;
  }

//////////////////////////////////////////////////////////////////////////
// Expr
//////////////////////////////////////////////////////////////////////////

  private Expr expr()
  {
    return assignExpr();
  }

  private Expr assignExpr()
  {
    // this tree is built to the right (others to the left)
    Expr expr = ternary();
    if (cur.isAssign())
    {
      if (curt == Token.ASSIGN)
        return new BinaryExpr(expr, consume(), expr, assignExpr());
      else
        return new ShortcutExpr(expr, consume(), expr, assignExpr());
    }
    return expr;
  }

  private Expr ternary()
  {
    Expr expr = condOrExpr();
    if (curt == Token.QUESTION)
    {
      Expr condition = expr;
      consume(Token.QUESTION);
      Expr trueExpr  = condOrExpr();
      consume(Token.COLON);
      Expr falseExpr = condOrExpr();
      expr = new TernaryExpr(condition.location, condition, trueExpr, falseExpr);
    }
    return expr;
  }

  private Expr condOrExpr()
  {
    Expr expr = condAndExpr();
    while (curt == Token.DOUBLE_PIPE)
      expr = new BinaryExpr(expr, consume(), expr, condAndExpr());
    return expr;
  }

  private Expr condAndExpr()
  {
    Expr expr = equalityExpr();
    while (curt == Token.DOUBLE_AMP)
      expr = new BinaryExpr(expr, consume(), expr, equalityExpr());
    return expr;
  }

  private Expr equalityExpr()
  {
    Expr expr = relationalExpr();
    while (curt == Token.EQ || curt == Token.NOT_EQ ||
           curt == Token.SAME || curt == Token.NOT_SAME)
    {
      if (curt == Token.SAME || curt == Token.NOT_SAME)
        expr = new BinaryExpr(expr, consume(), expr, relationalExpr());
      else
        expr = new ShortcutExpr(expr, consume(), expr, relationalExpr());
    }
    return expr;
  }

  private Expr relationalExpr()
  {
    Expr expr = rangeExpr();
    while (curt == Token.IS || curt == Token.AS ||
           curt == Token.LT || curt == Token.LTEQ ||
           curt == Token.GT || curt == Token.GTEQ ||
           curt == Token.COMPARE)
    {
      if (curt == Token.IS || curt == Token.AS)
        expr = new BinaryExpr(expr, consume(), expr, typeVarExpr());
      else
        expr = new ShortcutExpr(expr, consume(), expr, rangeExpr());
    }
    return expr;
  }

  private Expr rangeExpr()
  {
    Expr expr = bitOrExpr();
    if (curt == Token.DOT_DOT || curt == Token.DOT_DOT_DOT)
    {
      RangeLiteralExpr range = new RangeLiteralExpr(expr.location);
      range.start     = expr;
      range.inclusive = consume().type == Token.DOT_DOT;
      range.end       = bitOrExpr();
      return range;
    }
    return expr;
  }

  private Expr bitOrExpr()
  {
    Expr expr = bitAndExpr();
    while (curt == Token.CARET || curt == Token.PIPE)
    {
      if (peekt == Token.COMMA || sniffType()) break;  // in case pipe is a method type
      expr = new ShortcutExpr(expr, consume(), expr, bitAndExpr());
    }
    return expr;
  }

  private Expr bitAndExpr()
  {
    Expr expr = shiftExpr();
    while (curt == Token.AMP)
      expr = new ShortcutExpr(expr, consume(), expr, shiftExpr());
    return expr;
  }

  private Expr shiftExpr()
  {
    Expr expr = additiveExpr();
    while (curt == Token.LSHIFT || curt == Token.RSHIFT)
      expr = new ShortcutExpr(expr, consume(), expr, additiveExpr());
    return expr;
  }

  private Expr additiveExpr()
  {
    Expr expr = multiplicativeExpr();
    while (curt == Token.PLUS || curt == Token.MINUS)
      expr = new ShortcutExpr(expr, consume(), expr, multiplicativeExpr());
    return expr;
  }

  private Expr multiplicativeExpr()
  {
    Expr expr = parenExpr();
    while (curt == Token.STAR || curt == Token.SLASH || curt == Token.PERCENT)
      expr = new ShortcutExpr(expr, consume(), expr, parenExpr());
    return expr;
  }

  private Expr parenExpr()
  {
    if (curt != Token.LPAREN)
      return unaryExpr();

    // consume opening paren
    ILocation loc = cur;
    consume(Token.LPAREN);

    // In Fan just like C# and Java, a paren could mean either a
    // cast or a parenthesized expression, to figure this out
    // we follow the C# ECMA spec rules - see tryCastType() for a
    // more thorough description
    TypeRef castType = tryCastType();
    if (castType != null)
    {
      CastExpr cast = new CastExpr(loc);
      cast.toType = castType;
      consume(Token.RPAREN);
      cast.target = parenExpr();
      return cast;
    }

    // this is just a normal parenthesized expression
    Expr expr = expr();
    consume(Token.RPAREN);
    while (true)
    {
      Expr chained = termChainExpr(expr);
      if (chained == null) return expr;
      expr = chained;
    }
  }

  private Expr unaryExpr()
  {
    ILocation loc = cur;
    Token tok = cur;
    int tokt = curt;

    if (tokt == Token.BANG || tokt == Token.PLUS)
    {
      consume();
      return new UnaryExpr(loc, tokt, parenExpr());
    }

    if (tokt == Token.TILDE || tokt == Token.MINUS)
    {
      consume();
      return new ShortcutExpr(loc, tok, parenExpr());
    }

    if (tokt == Token.INCREMENT || tokt == Token.DECREMENT)
    {
      consume();
      return new UnaryExpr(loc, tokt == Token.INCREMENT ? Expr.PRE_INCREMENT : Expr.PRE_DECREMENT, tokt, parenExpr());
    }

    Expr expr = termExpr();

    tokt = curt;
    if (tokt == Token.INCREMENT || tokt == Token.DECREMENT)
    {
      consume();
      return new UnaryExpr(loc, tokt == Token.INCREMENT ? Expr.POST_INCREMENT : Expr.POST_DECREMENT, tokt, expr);
    }

    return expr;
  }

//////////////////////////////////////////////////////////////////////////
// Term Expr
//////////////////////////////////////////////////////////////////////////

  /**
   * A term is a base terminal such as a variable, call, or literal,
   * optionally followed by a chain of accessor expressions - such
   * as "x.y[z](a, b)".
   */
  private Expr termExpr()
  {
    Expr target = termBaseExpr();
    while (true)
    {
      Expr chained = termChainExpr(target);
      if (chained == null) return target;
      target = chained;
    }
  }

  /**
   * Parse a term base target.
   */
  private Expr termBaseExpr()
  {
    switch (curt)
    {
      case Token.ID:               return idExpr(null, false);
      case Token.INT_LITERAL:      return intLiteralExpr();
      case Token.CHAR_LITERAL:     return charLiteralExpr();
      case Token.FLOAT_LITERAL:    return floatLiteralExpr();
      case Token.STR_LITERAL:      return strLiteralExpr();
      case Token.DURATION_LITERAL: return durationLiteralExpr();
      case Token.URI_LITERAL:      return uriLiteralExpr();
      case Token.PIPE:             return methodTypeExpr();
      case Token.LBRACKET:         return collectionLiteralExpr();
      case Token.FALSE:            return boolLiteralExpr(false);
      case Token.NULL:             return nullLiteralExpr();
      case Token.SUPER:            return superExpr();
      case Token.THIS:             return thisExpr();
      case Token.TRUE:             return boolLiteralExpr(true);
    }
    throw err("Expected expression, not '" + cur + "'");
  }

  /**
   * A chain expression is a piece of a term expression that may
   * be chained together such as "call().foo->var[x]".  If the specified
   * target expression contain a chained access, then return the new
   * expression, otherwise return null.
   */
  private Expr termChainExpr(Expr target)
  {
    ILocation loc = cur;

    // if ".id" field access or ".id()" call
    if (curt == Token.DOT)
    {
      consume(Token.DOT);
      return idExpr(target, false);
    }

    // if "->id" dynamic call
    if (curt == Token.ARROW)
    {
      consume(Token.ARROW);
      return idExpr(target, true);
    }

    // if target[...]
    if (curt == Token.LBRACKET)
      return indexExpr(target);

    // if Key:Val
    if (curt == Token.COLON)
    {
      TypeRef keyType = target.toTypeRef();
      if (keyType != null)
      {
        int mark = pos;
        consume(Token.COLON);
        TypeRef valType = tryType();
        if (valType != null)
          return new TypeVarExpr(loc, new MapTypeRef(loc, keyType, valType));
        reset(mark);
      }
    }

    // otherwise the expression should be finished
    return null;
  }

//////////////////////////////////////////////////////////////////////////
// Term Expr Utils
//////////////////////////////////////////////////////////////////////////

  private Expr idExpr(Expr target, boolean dynamicCall)
  {
    if (peekt == Token.LPAREN)
    {
      CallExpr call = callExpr(target);
      call.isDynamic = dynamicCall;
      return call;
    }

    Token loc = cur;
    String name = consumeId();

    // a::b is fully qualified type ref
    /*
    if (curt == Token.POUND)
    {
      if (target != null) throw err("Unexpected fully qualified type", loc);
      consume();
      name += "::" + consumeId();
      return new TypeVarExpr(loc, new TypeRef(loc, name));
    }
    */

    // if we have a closure then this is a call with one arg of a closure
    ClosureExpr closure = tryClosure();
    if (closure != null)
    {
      CallExpr call = new CallExpr(loc);
      call.target = target;
      call.name   = name;
      call.isDynamic = dynamicCall;
      call.args.add(closure);
      return call;
    }

    // if dynamic call then we know this is a call not a field
    if (dynamicCall)
    {
      CallExpr call = new CallExpr(loc);
      call.target = target;
      call.name    = name;
      call.isDynamic = true;
      return call;
    }

    return new UnknownVarExpr(loc, target, name);
  }

  private CallExpr callExpr(Expr target)
  {
    CallExpr call = new CallExpr(cur);
    call.target = target;
    call.name   = consumeId();
    callArgs(call);
    return call;
  }

  private void callArgs(CallExpr call)
  {
    consume(Token.LPAREN);
    if (curt != Token.RPAREN)
    {
      while (true)
      {
        call.args.add( expr() );
        if (curt == Token.RPAREN) break;
        consume(Token.COMMA);
      }
    }
    consume(Token.RPAREN);

    if (!inCtorChain)
    {
      ClosureExpr closure = tryClosure();
      if (closure != null) call.args.add(closure);
    }
  }

  private Expr thisExpr()
  {
    Token loc = cur;
    consume(Token.THIS);
    return new ThisExpr(loc);
  }

  private Expr superExpr()
  {
    Token loc = cur;
    consume(Token.SUPER);
    return new SuperExpr(loc);
  }

  private Expr indexExpr(Expr target)
  {
    // TODO: this approach b/w index and lists/maps is fucked up
    ILocation loc = cur;
    consume(Token.LBRACKET);

    // check for Foo[] list type name
    if (curt == Token.RBRACKET)
    {
      TypeRef type = target.toTypeRef();
      if (type != null)
      {
        consume(Token.RBRACKET);
        return new TypeVarExpr(loc, new ListTypeRef(loc, type));
      }
    }

// check for Foo[Bar] map type name
    if (isTypeId(cur))
    {
      TypeRef valType = target.toTypeRef();
      if (valType != null)
      {
        TypeRef keyType = tryType();
        if (keyType != null)
        {
          consume(Token.RBRACKET);
          //return new TypeVarExpr(loc, new MapTypeRef(loc, keyType, valType));
// TODO
throw err("Old style map type sig Val[Key] should be Key:Val", valType);
        }
      }
    }

    // parse expression (unless this is [,] or [:])
    Expr expr = null;
    if (curt != Token.COMMA && curt != Token.COLON) expr = expr();

    // if we have a comma, then must be a list literal
    if (curt == Token.COMMA)
    {
      TypeRef type = target.toTypeRef();
      if (type == null) throw err("Invalid type for list literal, use Type[...]", target);
      return listLiteralExpr(loc, type, expr);
    }

    // if we have a colon, then must be a map literal
    if (curt == Token.COLON)
    {
      if (target instanceof TypeVarExpr)
      {
        TypeRef type = ((TypeVarExpr)target).typeRef;
        if (type instanceof MapTypeRef)
          return mapLiteralExpr(loc, (MapTypeRef)type, expr);
      }
      throw err("Invalid type for map literal, use ValType[KeyType][...]", target);
    }

    // otherwise this must be a standard single key index
    consume(Token.RBRACKET);
    return new ShortcutExpr(loc, ShortcutExpr.GET, null, target, expr);
  }

//////////////////////////////////////////////////////////////////////////
// Literal Expressions
//////////////////////////////////////////////////////////////////////////

  private LiteralExpr boolLiteralExpr(boolean val)
  {
    Token loc = cur;
    if (val) consume(Token.TRUE);
    else consume(Token.FALSE);
    return new LiteralExpr(loc, Expr.BOOL_LITERAL, val ? Boolean.TRUE : Boolean.FALSE);
  }

  private LiteralExpr intLiteralExpr()
  {
    Token loc = cur;
    Long val = (Long)consume(Token.INT_LITERAL).val;
    return new LiteralExpr(loc, Expr.INT_LITERAL, val);
  }

  private LiteralExpr charLiteralExpr()
  {
    Token loc = cur;
    Character val = (Character)consume(Token.CHAR_LITERAL).val;
    return new LiteralExpr(loc, Expr.INT_LITERAL, new Long(val.charValue()));
  }

  private LiteralExpr floatLiteralExpr()
  {
    Token loc = cur;
    Double val = (Double)consume(Token.FLOAT_LITERAL).val;
    return new LiteralExpr(loc, Expr.FLOAT_LITERAL, val);
  }

  private LiteralExpr strLiteralExpr()
  {
    Token loc = cur;
    String val = (String)consume(Token.STR_LITERAL).val;
    return new LiteralExpr(loc, Expr.STR_LITERAL,   val);
  }

  private LiteralExpr durationLiteralExpr()
  {
    Token loc = cur;
    Long val = (Long)consume(Token.DURATION_LITERAL).val;
    return new LiteralExpr(loc, Expr.DURATION_LITERAL,   val);
  }

  private LiteralExpr uriLiteralExpr()
  {
    Token loc = cur;
    String val = (String)consume(Token.URI_LITERAL).val;
    try { Uri.parse(Str.make(val)); }
    catch (Exception e) { throw err("Invalid URI '" + val + "'", loc); }
    return new LiteralExpr(loc, Expr.URI_LITERAL,   val);
  }

  private LiteralExpr nullLiteralExpr()
  {
    Token loc = cur;
    consume(Token.NULL);
    return new LiteralExpr(loc, Expr.NULL_LITERAL, null);
  }

//////////////////////////////////////////////////////////////////////////
// Collection "Literals"
//////////////////////////////////////////////////////////////////////////

  private Expr collectionLiteralExpr()
  {
    Token loc = cur;

    // just in case this is a [Key:Val], try type first
    TypeRef type = tryType();
    if (type != null)
      return new TypeVarExpr(loc, type);

    // opening bracket
    consume(Token.LBRACKET);

    // empty list?
    if (curt == Token.COMMA)
    {
      consume(Token.COMMA);
      consume(Token.RBRACKET);
      return new ListLiteralExpr(loc, null);
    }

    // empty map?
    if (curt == Token.COLON)
    {
      consume(Token.COLON);
      consume(Token.RBRACKET);
      return new MapLiteralExpr(loc, null);
    }

    // [] is error
    if (curt == Token.RBRACKET)
    {
      throw err("Invalid list literal; use '[,]' for empty Obj[] list");
    }

    // read first expression
    Expr first = expr();

    // at this point we can determine if it is a list or a map
    if (curt == Token.COLON)
      return mapLiteralExpr(loc, null, first);
    else
      return listLiteralExpr(loc, null, first);
  }

  /**
   * Parse List literal, cur must be on comma after first item
   */
  private ListLiteralExpr listLiteralExpr(ILocation loc, TypeRef type, Expr first)
  {
    ListLiteralExpr list = new ListLiteralExpr(loc, type);
    if (first != null) list.items.add(first);
    while (curt == Token.COMMA)
    {
      consume(Token.COMMA);
      if (curt == Token.RBRACKET) break; // allow extra trailing comma
      list.items.add( expr() );
    }
    consume(Token.RBRACKET);
    return list;
  }

  /**
   * Parse Map literal, cur must be on colon of first pair
   */
  private MapLiteralExpr mapLiteralExpr(ILocation loc, MapTypeRef mapType, Expr first)
  {
    MapLiteralExpr map = new MapLiteralExpr(loc, mapType);
    if (first == null)
    {
      consume(Token.COLON);
      consume(Token.RBRACKET);
      return map;
    }

    map.keys.add(first);
    consume(Token.COLON);
    map.vals.add( expr() );
    while (curt == Token.COMMA)
    {
      consume(Token.COMMA);
      if (curt == Token.RBRACKET) break; // allow extra trailing comma
      map.keys.add( expr() );
      consume(Token.COLON);
      map.vals.add( expr() );
    }
    consume(Token.RBRACKET);
    return map;
  }

//////////////////////////////////////////////////////////////////////////
// Closure
//////////////////////////////////////////////////////////////////////////

  /**
   * Called when we encounter || or |...| as start of expression:
   *   - Type literal: |...|.type
   *   - Closure:      |...| { ... }
   */
  private Expr methodTypeExpr()
  {
    ILocation loc = cur;

    MethodTypeRef methodType = methodType();

    // at this point there are a couple obvious cases where
    // this is a type literal and not a closure:
    //   |xxx|.type
    //   |xxx| var
    if (curt == Token.DOT || curt == Token.ID)
      return new TypeVarExpr(loc, methodType);

    // otherwise assume a closure body
    ClosureExpr c = new ClosureExpr(loc);
    c.signature = methodType;
    c.code = closureBody(methodType);
    return c;
  }

  /**
   * Attempt to parse a closure expression or return null
   * if we aren't positioned at the start of a closure expression.
   */
  private ClosureExpr tryClosure()
  {
    ILocation loc = cur;

    // No args:  |,| { ... }
    if (curt == Token.PIPE && peekt == Token.COMMA)
    {
      consume(Token.PIPE);
      consume(Token.COMMA);
      consume(Token.PIPE);
      ClosureExpr c = new ClosureExpr(loc);
      c.signature = new MethodTypeRef(loc);
      c.signature.ret = new TypeRef(loc, Sys.VoidType);
      c.code = closureBody(c.signature);
      return c;
    }

    // gotta have pipe now
    if (curt != Token.PIPE) return null;

    // try to parse |A a, B b -> R|
    int mark = pos;
    MethodTypeRef type = (MethodTypeRef)tryType();
    if (type == null) return null;

    // if we don't see opening brace for body - no go
    if (curt != Token.LBRACE) { reset(mark); return null; }

    // we should have a closure now
    ClosureExpr c = new ClosureExpr(loc);
    c.signature = type;
    c.code = closureBody(c.signature);
    return c;
  }

  /**
   * Parse a closure body: { stmt* }
   */
  private Block closureBody(MethodTypeRef closureType)
  {
    boolean oldInVoid = inVoid;
    if (curt != Token.LBRACE)
      throw err("Expected '{' for start of closure body, not '" + cur + "'");

    Block body = block(closureType.ret.isVoid());

    inVoid = oldInVoid;
    return body;
  }

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  /**
   * Parse a type into a TypeVarExpr
   */
  private TypeVarExpr typeVarExpr()
  {
    return new TypeVarExpr(cur, type());
  }

  /**
   * Return true if the current stream of tokens can
   * be parsed as a valid type production.  No actual
   * tokens are consumed by this method.
   */
  private boolean sniffType()
  {
    suppressErrors = true;
    int mark = pos;
    TypeRef type = null;
    try
    {
      type = type();
    }
    catch (SuppressedException e)
    {
    }
    suppressErrors = false;
    reset(mark);
    return type != null;
  }

  /**
   * If the current stream of tokens can be parsed as a
   * valid type production return it.  Otherwise leave
   * the parse positioned on the current token.
   */
  private TypeRef tryType()
  {
    suppressErrors = true;
    int mark = pos;
    TypeRef type = null;
    try
    {
      type = type();
    }
    catch (SuppressedException e)
    {
    }
    suppressErrors = false;
    if (type == null) reset(mark);
    return type;
  }

  /**
   * Parse a type production.
   */
  private TypeRef type()
  {
    Token loc = cur;

    // check for id
    if (isTypeId(cur))
    {
      String id = consumeId();
      TypeRef type = new TypeRef(loc, id);

// TODO
if (curt == Token.LBRACKET && isTypeId(peek))
  throw err("Old style map type sig Val[Key] should be Key:Val", cur);

      type = listType(type);

      // Foo:Bar - parameterized Map
      if (curt == Token.COLON && isTypeId(peek))
      {
        consume(Token.COLON);
        TypeRef key = type;
        TypeRef val = type();
        type = new MapTypeRef(loc, key, val);
      }

      return listType(type);
    }

    // [K:V] - parameterized Map
    TypeRef type;
    if (curt == Token.LBRACKET)
    {
      consume(Token.LBRACKET);
      type = type();
      consume(Token.RBRACKET);
      return listType(type);
    }

    // |A a, B b -> R| - parameterized Method
    if (curt == Token.PIPE)
      return listType(methodType());

    throw err("Expected type identifier, not '" + cur + "'");
  }

  private TypeRef listType(TypeRef type)
  {
    // Foo[] - parameterized List
    while (curt == Token.LBRACKET && peekt == Token.RBRACKET)
    {
      consume(Token.LBRACKET);
      consume(Token.RBRACKET);
      type = type.toListOf();
    }
    return type;
  }

  private boolean isTypeId(Token tok) { return tok.type == Token.ID && isTypeId(tok.val.toString()); }
  public static boolean isTypeId(String id)
  {
    // TODO: hack hack hack!
    // Use capitalization to check for type identifer versus
    // some other identifier - this needs to get replaced with
    // a two pass parser (1st scan tokens looking for class X)
    return id.charAt(0) < 'a';
  }

  /**
   * Parser a parameterized method type production.
   */
  private MethodTypeRef methodType()
  {
    MethodTypeRef type = new MethodTypeRef(cur);

    consume(Token.PIPE);

    if (curt == Token.COMMA)
    {
      consume(Token.COMMA);
      consume(Token.PIPE);
      type.ret = new TypeRef(type.location, Sys.VoidType);
      return type;
    }

    while (true)
    {
      // if we see a pipe, this could be either an method type
      // argument or the closing pipe with a default ->Void;
      // note we optimize for the common case of |...| {...}
      if (curt == Token.PIPE)
      {
        if (peekt == Token.LBRACE || !sniffType())
        {
          type.ret = new TypeRef(type.location, Sys.VoidType);
          consume();
          break;
        }
      }

      // once we see an arrow, only "-> ret |" is
      // allowed to be left
      if (curt == Token.ARROW)
      {
        consume();
        type.ret = type();
        consume(Token.PIPE);
        break;
      }

      if (type.paramTypes.size() > 0) consume(Token.COMMA);
      type.paramTypes.add( type() );
      type.paramNames.add( consumeId() );
    }

    return type;
  }

  /**
   * Try to parse a cast expression type production, and if the current
   * sequence of tokens is not a type production, then leave the tokens
   * untouched and return null.  We assume the open paren has been consumed
   * and if successful the cur token should be left on the close paren.
   * NOTE: this method must be kept in sync with type()
   */
  private TypeRef tryCastType()
  {
    // In Fan just like C# and Java, a paren could mean either a
    // cast or a parenthesized expression, to figure this out
    // we follow the C# ECMA spec (14.6.6 Cast expressions)...
    //
    // To resolve cast expression ambiguities, the following rule exists: a
    // sequence of one or more tokens enclosed on parentheses is considered
    // the start of a cast expression only if at least one of the following
    // are true:
    //  - the sequence of tokens is a correct grammar for a type, but not for
    //    an expression (for example a::b)
    //  - the sequence of tokens is a correct grammar for a type (for example
    //    a or a::b), and the token immediately following the closing parentheses
    //    is the token an identifier, ~, !, (, a literal, or any keyword except
    //    as and is
    //  casts: (x)y  (x)(y) and (x)(-y)
    //  not casts: (x)-y   (as long as x is not a type keyword)

    // attempt to parse a type production
    int mark = pos;
    suppressErrors = true;
    TypeRef type = null;
    try
    {
      type = type();
    }
    catch (SuppressedException e)
    {
    }
    suppressErrors = false;

    // ok at this point we must have a close paren for this to be a cast
    if (type != null && curt == Token.RPAREN)
    {
      // we can meet rule two by checking the next token
      if (peekt == Token.ID   || peekt == Token.TILDE  ||
          peekt == Token.BANG || peekt == Token.LPAREN ||
          peek.isLiteral()    ||
          (peek.isKeyword() && peekt != Token.AS && peekt != Token.IS))
        return type;
    }

    // if we haven't returned by this point, it means we have consumed
    // tokens but discovered that this isn't a cast operation so reset
    // the parser back to the token we were on entering this method
    reset(mark);
    return null;
  }

//////////////////////////////////////////////////////////////////////////
// Errors
//////////////////////////////////////////////////////////////////////////

  CompileException err(String msg)
  {
    if (suppressErrors) throw new SuppressedException();
    return err(msg, cur);
  }

  static class SuppressedException extends RuntimeException {}

//////////////////////////////////////////////////////////////////////////
// Tokens
//////////////////////////////////////////////////////////////////////////

  /**
   * Verify current is an identifier, consume it, and return it.
   */
  String consumeId()
  {
    if (curt != Token.ID)
      throw err("Expected identifier, not '" + cur + "'");
    return (String)consume().val;
  }

  /**
   * Check that the current token matches the specified
   * type, and then consume it.  Return consumed token.
   */
  Token consume(int type)
  {
    verify(type);
    return consume();
  }

  /**
   * Check that the current token matches the specified
   * type, but do not consume it.
   */
  void verify(int type)
  {
    if (curt != type)
      throw err("Expected '" + Token.toString(type) + "', not '" + cur + "'");
  }

  /**
   * Consume the current token.  Return consumed token.
   */
  Token consume()
  {
    Token result = cur;

    // get the next token from the buffer, if pos is past numTokens,
    // then always use the last token which will be eof
    Token next;
    pos++;
    if (pos+1 < numTokens)
      next = tokens[pos+1];  // next peek is cur+1
    else
      next = tokens[numTokens-1];

    cur   = peek;
    peek  = next;
    curt  = cur.type;
    peekt = peek.type;

    return result;
  }

  /**
   * Statements can be terminated with a semicolon, end of line or } end of block.
   */
  void endOfStmt()
  {
    if (curt == Token.SEMICOLON) { consume(); return; }
    if (cur.newline) return;
    if (curt == Token.RBRACE) return;
    throw err("Expected end of statement: semicolon, newline, or end of block; not '" + cur + "'");
  }


  /**
   * Reset the current position to the specified tokens index.
   */
  void reset(int pos)
  {
    this.pos   = pos;
    this.cur   = tokens[pos];
    this.peek  = pos+1 < numTokens ? tokens[pos+1] : tokens[pos];
    this.curt  = cur.type;
    this.peekt = peek.type;
  }

//////////////////////////////////////////////////////////////////////////
// Parser Flags
//////////////////////////////////////////////////////////////////////////

  // These are flags used only by the parser we merge with FConst
  // flags by starting from most significant bit and working down
  public static final int Readonly = 0x80000000;

  // Bitwise and this mask to clear all protection scope flags
  public static final int ProtectionMask = ~(FConst.Public|FConst.Protected|FConst.Private|FConst.Internal);

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  String source;           // name of source to use for error reporting
  Tokenizer tokenizer;     // tokenizer
  Token[] tokens;          // tokens all read in
  int numTokens;           // number of tokens
  int pos;                 // offset into tokens for cur
  Token cur;               // current token
  int curt;                // current token type
  Token peek;              // next token
  int peekt;               // next token type
  CompilationUnit unit;    // compilation unit to generate
  boolean isSys;           // is this the sys pod
  boolean suppressErrors;  // throw SuppressedException rather than CompileException
  boolean inCtorChain;     // are we currently in a constructor chain call
  boolean inVoid;          // are we currently in a void method

}