//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   3 Sep 05  Brian Frank  Creation
//
package fanx.compiler.parser;

import java.io.*;
import java.util.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.util.*;

/**
 * Tokenizer inputs a Str and output a list of Tokens
 */
public class Tokenizer
  extends CompilerSupport
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  /**
   * Construct with characters of source file.  The buffer
   * passed must be normalized in that all newlines must be
   * represented strictly as \n and not \r or \r\n (see
   * SourceReader).  The buffer must be sized exactly
   * correct -- this let's us take advantage of implicit
   * array checking without having to access a separate
   * length variable.
   */
  public Tokenizer(Compiler compiler, char[] buf, String fileName)
  {
    super(compiler);
    this.buf      = buf;
    this.fileName = fileName;
    this.isDoc    = compiler != null ? compiler.includeDoc : false;
    this.tokens   = new ArrayList();

    // initialize cur and peek
    cur = peek = ' ';
    if (buf.length > 0) cur  = buf[0];
    if (buf.length > 1) peek = buf[1];
    curt  = cur  < 128 ? charMap[cur]  : ALPHA;
    peekt = peek < 128 ? charMap[peek] : ALPHA;
    pos = 0;

    // if first line starts with #, then treat it like an end of
    // line, so that Unix guys can specify the executable to run
    if (cur == '#')
    {
      while (true)
      {
        if (cur == '\n') { consume(); break; }
        if (cur == 0) break;
        consume();
      }
    }
  }

////////////////////////////////////////////////////////////////
// Access
////////////////////////////////////////////////////////////////

  /**
   * Tokenize the buffer
   */
  public Token[] tokenize()
  {
    while (true)
    {
      Token tok = next();
      tokens.add(tok);
      if (tok.type == Token.EOF) break;
    }
    return (Token[])tokens.toArray(new Token[tokens.size()]);
  }

  /**
   * Return the next token in the buffer.
   */
  private Token next()
  {
    while (true)
    {
      int line = this.line;
      int col  = this.col;

      Token tok = find();

      if (tok == null) continue;

      tok.file    = fileName;
      tok.line    = line;
      tok.col     = col;
      tok.newline = lastLine < line;

      lastLine = line;

      return tok;
    }
  }

  private Token find()
  {
    // skip whitespace
    if (curt == SPACE) { consume(); return null; }

    // alpha means keyword or identifier
    if (curt == ALPHA) return word();

    // number, .number (note that + and - are handled as unary operator)
    if (curt == DIGIT) return number();
    if (cur == '.' && (peekt == DIGIT)) return number();

    // str literal
    if (cur == '"')  return str();
    if (cur == '\'') return ch();
    if (cur == '`')  return uri();

    // comments
    if (cur == '*' && peek == '*') return docComment();
    if (cur == '/' && peek == '/') return skipCommentSL();
    if (cur == '/' && peek == '*') return skipCommentML();

    // symbols
    return symbol();
  }

//////////////////////////////////////////////////////////////////////////
// Word
//////////////////////////////////////////////////////////////////////////

  /**
   * Parse a word token: alpha (alpha|number)*
   * Words are either keywords or identifiers
   */
  private Token word()
  {
    // store starting position of word
    int start = pos;

    // find end of word to compute length
    while (curt == ALPHA || curt == DIGIT) consume();
    int len = pos-start;

    // create String (gc note this string now references buf)
    String word = new String(buf, start, len);

    // check keywords
    Integer keyword= (Integer)Token.keywords.get(word);
    if (keyword != null)
      return new Token(keyword.intValue());

    // otherwise this is a normal identifier
    return new Token(Token.ID, word);
  }

//////////////////////////////////////////////////////////////////////////
// Number
//////////////////////////////////////////////////////////////////////////

  /**
   * Parse a number literal token.
   */
  private Token number()
  {
    boolean neg = false;
    long whole = 0;
    int wholeCount = 0;
    double fraction = 0;
    int exp = 0;
    int type = Token.INT_LITERAL;  // int or float literal

    // check for hex value
    if (cur == '0' && peek == 'x')
      return hex();

    // if we are starting with a sign
    /*
    if (cur == '-') { consume(); neg = true; }
    else if (cur == '+') { consume(); }
    */

    // read whole part
    while (curt == DIGIT)
    {
      whole = whole*10 + digit(cur);
      consume();
      // it would be nice to actually trap max long value, but
      // as a simpler catch all we know decimal numbers must
      // never be bigger than 19 digits
      wholeCount++;
      if (wholeCount > 19) throw err("Numeric literal too big");
      if (cur == '_') consume();
    }

    // if dot then read fraction
    if (cur == '.' && peekt == DIGIT)
    {
      consume();
      for (double m=0.1; curt == DIGIT; m/=10)
      {
        fraction = fraction + digit(cur)*m;
        consume();
        if (cur == '_') consume();
      }
      type = Token.FLOAT_LITERAL;
    }

    // check for exponent
    if (cur == 'e' || cur == 'E')
    {
      consume();
      boolean negExp = false;
      if (cur == '-') { consume(); negExp = true; }
      else if (cur == '+') { consume(); }
      if (curt != DIGIT) throw err("Expected exponent digits");
      while (curt == DIGIT)
      {
        exp = exp*10 + digit(cur);
        consume();
        if (cur == '_') consume();
      }
      if (negExp) exp = -exp;
      type = Token.FLOAT_LITERAL;
    }

    // check for F suffix
    if (cur == 'f' || cur == 'F')
    {
      consume();
      type = Token.FLOAT_LITERAL;
    }

    // check if a duration
    long dur = -1;
    if (cur == 'n' && peek == 's') { consume(); consume(); dur = 1L; }
    if (cur == 'm' && peek == 's') { consume(); consume(); dur = 1000000L; }
    if (cur == 's' && peek == 'e') { consume(); consume(); if (cur != 'c') throw err("Expected 'sec' in Duration literal"); consume(); dur = 1000000000L; }
    if (cur == 'm' && peek == 'i') { consume(); consume(); if (cur != 'n') throw err("Expected 'min' in Duration literal"); consume(); dur = 60000000000L; }
    if (cur == 'h' && peek == 'r') { consume(); consume(); dur = 3600000000000L; }
    if (cur == 'd' && peek == 'a') { consume(); consume(); if (cur != 'y' || peek != 's') throw err("Expected 'days' in Duration literal"); consume(); consume(); dur = 86400000000000L; }

    // int literal
    if (type == Token.INT_LITERAL)
    {
      if (fraction != 0 || exp != 0) throw err("Int literals may not have fractional or exponent components");
      long val = whole;
      if (neg) val = -val;
      if (dur > 0)
        return new Token(Token.DURATION_LITERAL, new Long(val*dur));
      else
        return new Token(type, new Long(val));
    }

    // float literal
    if (type == Token.FLOAT_LITERAL)
    {
      double val = whole + fraction;
      if (exp != 0) val = val * Math.pow(10, exp);
      if (neg) val = -val;
      if (dur > 0)
        return new Token(Token.DURATION_LITERAL, new Long((long)(val*dur)));
      else
        return new Token(type, new Double(val));
    }

    throw err("Interenal error");
  }

  /**
   * Process hex int/long literal starting with 0x
   */
  Token hex()
  {
    consume(); // 0
    consume(); // x

    // read first hex
    int type = Token.INT_LITERAL;
    long val = hex(cur);
    if (val < 0) throw err("Expecting hex number");
    consume();
    int nibCount = 1;
    while (true)
    {
      int nib = hex(cur);
      if (nib < 0)
      {
        if (cur == '_') { consume(); continue; }
        break;
      }
      nibCount++;
      if (nibCount > 16) throw err("Hex literal too big");
      val = (val << 4) + nib;
      consume();
    }

    return new Token(type, new Long(val));
  }

  static int digit(int c)
  {
    return c - '0';
  }

  static int hex(int c)
  {
    if ('0' <= c && c <= '9') return c - '0';
    if ('a' <= c && c <= 'f') return c - 'a' + 10;
    if ('A' <= c && c <= 'F') return c - 'A' + 10;
    return -1;
  }

//////////////////////////////////////////////////////////////////////////
// String
//////////////////////////////////////////////////////////////////////////

  /**
   * Parse a string literal token.
   */
  private Token str()
  {
    // consume opening quote
    consume();

    // store starting position
    StringBuilder s = new StringBuilder();

    // loop until we find end of string
    boolean interpolated = false;
    while (true)
    {
      if (cur == '"') { consume(); break; }
      if (cur == 0) throw err("Unexpected end of string");
      if (cur == '$')
      {
        // if we have detected an interpolated string, then
        // insert opening paren to treat whole string atomically
        if (!interpolated)
        {
          interpolated = true;
          tokens.add(makeToken(Token.LPAREN, null));
        }

        if (!strInterpolation(s.toString()))
        {
          tokens.add(makeToken(Token.RPAREN, null));
          return null;
        }

        s.setLength(0);
      }
      else if (cur == '\\')
      {
        s.append(escape());
      }
      else
      {
        s.append((char)cur);
        consume();
      }
    }

    if (interpolated)
    {
      tokens.add(makeToken(Token.STR_LITERAL, s.toString()));
      tokens.add(makeToken(Token.RPAREN, null));
      return null;
    }
    return new Token(Token.STR_LITERAL, s.toString());
  }

  /**
   * When we hit a $ inside a string it indicates
   * an embedded expression.  We make this look like
   * a stream of tokens such that:
   *   "a ${b} c" -> ("a " + b + " c")
   * Return true if more in the string literal.
   */
  private boolean strInterpolation(String s)
  {
    consume(); // $
    tokens.add(makeToken(Token.STR_LITERAL, s));
    tokens.add(makeToken(Token.PLUS, null));

    // if { we allow an expression b/w {...}
    if (cur == '{')
    {
      tokens.add(makeToken(Token.LPAREN, null));
      consume();
      while (true)
      {
        if (cur == '"')
          throw err("Unexpected end of string, missing }");
        Token tok = next();
        if (tok.type == Token.RBRACE) break;
        tokens.add(tok);
      }
      tokens.add(makeToken(Token.RPAREN, null));
    }

    // else also allow a single identifier
    else
    {
      Token tok = next();
      if (tok.type != Token.ID) throw err("Expected identifier after $");
      tokens.add(tok);
      while (true)
      {
        if (cur != '.') break;
        tokens.add(next()); // dot
        tok = next();
        if (tok.type != Token.ID) throw err("Expected identifier");
        tokens.add(tok);
      }
    }

    // if not at end of string, then add plus
    if (cur == '\"')
    {
      consume();
      return false;
    }

    // add plus and keep chugging
    tokens.add(makeToken(Token.PLUS, null));
    return true;
  }

  /**
   * Create an implicit + operator for string interpolation.
   */
  private Token makeToken(int type, Object value)
  {
    Token tok = new Token(type, value);
    tok.file  = fileName;
    tok.line  = line;
    tok.col   = col;
    return tok;
  }

//////////////////////////////////////////////////////////////////////////
// Character
//////////////////////////////////////////////////////////////////////////

  /**
   * Parse a char literal token.
   */
  private Token ch()
  {
    // consume opening quote
    consume();

    // if \ then process as escape
    char c;
    if (cur == '\\')
    {
      c = escape();
    }
    else
    {
      c = (char)cur;
      consume();
    }

    // expecting ' quote
    if (cur != '\'') throw err("Expecting ' close of char literal");
    consume();

    return new Token(Token.CHAR_LITERAL, new Character(c));
  }

  /**
   * Parse an escapse sequence which starts with a \
   */
  char escape()
  {
    // consume slash
    if (cur != '\\') throw err("Internal error");
    consume();

    // check basics
    switch (cur)
    {
      case 'b':   consume(); return '\b';
      case 'f':   consume(); return '\f';
      case 'n':   consume(); return '\n';
      case 'r':   consume(); return '\r';
      case 't':   consume(); return '\t';
      case '$':   consume(); return '$';
      case '"':   consume(); return '"';
      case '\'':  consume(); return '\'';
      case '\\':  consume(); return '\\';
    }

    // check for uxxxx
    if (cur == 'u')
    {
      consume();
      int n3 = hex(cur); consume();
      int n2 = hex(cur); consume();
      int n1 = hex(cur); consume();
      int n0 = hex(cur); consume();
      if (n3 < 0 || n2 < 0 || n1 < 0 || n0 < 0) throw err("Invalid hex value for \\uxxxx");
      return (char)((n3 << 12) | (n2 << 8) | (n1 << 4) | n0);
    }

    throw err("Invalid escape sequence");
  }

//////////////////////////////////////////////////////////////////////////
// Uri
//////////////////////////////////////////////////////////////////////////

  /**
   * Parse a uri literal token.
   */
  private Token uri()
  {
    // consume opening tick
    consume();

    // store starting position
    StringBuilder s = new StringBuilder();

    // loop until we find end of string
    while (true)
    {
      if (cur == '`') { consume(); break; }
      if (cur == 0) throw err("Unexpected end of uri");
      s.append((char)cur);
      consume();
    }

    return new Token(Token.URI_LITERAL, s.toString());
  }

//////////////////////////////////////////////////////////////////////////
// Comments
//////////////////////////////////////////////////////////////////////////

  /**
   * Skip a single line // comment
   */
  private Token skipCommentSL()
  {
    consume(); // first slash
    consume(); // next slash
    while (true)
    {
      if (cur == '\n') { consume(); break; }
      if (cur == 0) break;
      consume();
    }
    return null;
  }

  /**
   * Skip a multi line /* comment.  Note unlike C/Java,
   * slash/star comments can be nested.
   */
  private Token skipCommentML()
  {
    consume(); // first slash
    consume(); // next slash
    int depth = 1;
    while (true)
    {
      if (cur == '*' && peek == '/') { consume(); consume(); depth--; if (depth <= 0) break; }
      if (cur == '/' && peek == '*') { consume(); consume(); depth++; continue; }
      if (cur == 0) break;
      consume();
    }
    return null;
  }

  /**
   * Parse a Javadoc style comment into a documentation comment token.
   */
  public Token docComment()
  {
    // if doc is off, then just skip the line and be done
    if (!isDoc) { skipCommentSL(); return null; }

    while (cur == '*') consume();
    if (cur == ' ') consume();

    // parse comment
    StringBuilder s = new StringBuilder();
    while (cur > 0)
    {
      // add to buffer and advance
      int c = cur;
      s.append((char)c);
      consume();

      // if not at newline, then loop
      if (c != '\n') continue;

      // we at a newline, check for leading whitespace(0+)/star(2+)/whitespace(1)
      while (cur == ' ' || cur == '\t') consume();
      if (cur != '*' || peek != '*') break;
      while (cur == '*') consume();
      if (cur == ' ' || cur == '\t') consume();
    }

    String doc = s.toString().trim();
    return new Token(Token.DOC_COMMENT, doc);
  }

//////////////////////////////////////////////////////////////////////////
// Symbol
//////////////////////////////////////////////////////////////////////////

  /**
   * Parse a symbol token (typically into an operator).
   */
  private Token symbol()
  {
    int c = cur;
    consume();
    switch (c)
    {
      case 0:
        return new Token(Token.EOF);
      case '\r':
        throw err("Carriage return \\r not allowed in source");
      case '!':
        if (cur == '=')
        {
          consume();
          if (cur == '=') { consume(); return new Token(Token.NOT_SAME); }
          return new Token(Token.NOT_EQ);
        }
        return new Token(Token.BANG);
      case '#':
        return new Token(Token.POUND);
      case '%':
        if (cur== '=') { consume(); return new Token(Token.ASSIGN_PERCENT); }
        return new Token(Token.PERCENT);
      case '&':
        if (cur == '=') { consume(); return new Token(Token.ASSIGN_AMP); }
        if (cur == '&') { consume(); return new Token(Token.DOUBLE_AMP); }
        return new Token(Token.AMP);
      case '(':
        return new Token(Token.LPAREN);
      case ')':
        return new Token(Token.RPAREN);
      case '*':
        if (cur == '=') { consume(); return new Token(Token.ASSIGN_STAR); }
        return new Token(Token.STAR);
      case '+':
        if (cur == '=') { consume(); return new Token(Token.ASSIGN_PLUS); }
        if (cur == '+') { consume(); return new Token(Token.INCREMENT); }
        return new Token(Token.PLUS);
      case ',':
        return new Token(Token.COMMA);
      case '-':
        if (cur == '>') { consume(); return new Token(Token.ARROW); }
        if (cur == '-') { consume(); return new Token(Token.DECREMENT); }
        if (cur == '=') { consume(); return new Token(Token.ASSIGN_MINUS); }
        return new Token(Token.MINUS);
      case '.':
        if (cur == '.')
        {
          consume();
          if (cur == '.') { consume(); return new Token(Token.DOT_DOT_DOT); }
          return new Token(Token.DOT_DOT);
        }
        return new Token(Token.DOT);
      case '/':
        if (cur == '=') { consume(); return new Token(Token.ASSIGN_SLASH); }
        return new Token(Token.SLASH);
      case ':':
        if (cur == '=') { consume(); return new Token(Token.DEF_ASSIGN); }
        return new Token(Token.COLON);
      case ';':
        return new Token(Token.SEMICOLON);
      case '<':
        if (cur == '=')
        {
          consume();
          if (cur == '>') { consume(); return new Token(Token.COMPARE); }
          return new Token(Token.LTEQ);
        }
        if (cur == '<')
        {
          consume();
          if (cur == '=') { consume(); return new Token(Token.ASSIGN_LSHIFT); }
          return new Token(Token.LSHIFT);
        }
        return new Token(Token.LT);
      case '=':
        if (cur == '=')
        {
          consume();
          if (cur == '=') { consume(); return new Token(Token.SAME); }
          return new Token(Token.EQ);
        }
        return new Token(Token.ASSIGN);
      case '>':
        if (cur == '=') { consume(); return new Token(Token.GTEQ); }
        if (cur == '>')
        {
          consume();
          if (cur == '=') { consume(); return new Token(Token.ASSIGN_RSHIFT); }
          return new Token(Token.RSHIFT);
        }
        return new Token(Token.GT);
      case '?':
        return new Token(Token.QUESTION);
      case '[':
        return new Token(Token.LBRACKET);
      case ']':
        return new Token(Token.RBRACKET);
      case '^':
        if (cur == '=') { consume(); return new Token(Token.ASSIGN_CARET); }
        return new Token(Token.CARET);
      case '{':
        return new Token(Token.LBRACE);
      case '|':
        if (cur == '|') { consume(); return new Token(Token.DOUBLE_PIPE); }
        if (cur == '=') { consume(); return new Token(Token.ASSIGN_PIPE); }
        return new Token(Token.PIPE);
      case '}':
        return new Token(Token.RBRACE);
      case '~':
        return new Token(Token.TILDE);
    }

    throw err("Unexpected symbol: " + (char)c + " (0x" + Integer.toHexString(c) + ")");
  }

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  /**
   * Dump state in a one line summary.
   */
  private void dump()
  {
    System.out.println(line + ":" + col + "  cur=" + str(cur) + " [" + curt + "] peek=" + str(peek) + " [" + peekt + "]");
  }

  /**
   * Get a suitable debug string for the specified char.
   */
  private String str(int c)
  {
    if (c == 0)    return "eof";
    if (c == ' ')  return "_";
    if (c == '\n') return "\\n";
    return String.valueOf((char)c);
  }

  /**
   * Return a CompileException for current location in source.
   */
  public CompileException err(String msg)
  {
    return err(msg, new Location(fileName, line, col));
  }

////////////////////////////////////////////////////////////////
// Read
////////////////////////////////////////////////////////////////

  /**
   * Consume the cur char and advance to next char in buffer:
   *  - updates cur, curt, peek, and peekt fields
   *  - updates the line and col count
   *  - end of file, sets fields to 0
   */
  private void consume()
  {
    // if cur is a line break, then advance line number,
    // because the char we are getting ready to make cur
    // is the first char on the next line
    if (cur == '\n')
    {
      line++;
      col = 1;
    }
    else
    {
      col++;
    }

    // get the next character from the buffer, any
    // problems mean that we have read past the end
    int c = 0;
    pos++;
    try
    {
      c = buf[pos+1];  // next peek is cur+1
    }
    catch (Exception e)
    {
      c = 0;
    }

    // update fields
    cur   = peek;
    curt  = peekt;
    peek  = c;
    peekt = c < 128 ? charMap[c] : ALPHA;
  }

////////////////////////////////////////////////////////////////
// Char Map
////////////////////////////////////////////////////////////////

  /**
   * Is the specified char a valid name identifier character.
   */
  public static boolean isIdChar(int ch)
  {
    int type = ch < 128 ? charMap[ch] : ALPHA;
    return type == ALPHA || type == DIGIT;
  }

  private static final byte[] charMap = new byte[128];
  private static final int SPACE = 1;
  private static final int ALPHA = 2;
  private static final int DIGIT = 3;
  static
  {
    // space characters; note \r is error in symbol()
    charMap[' ']  = SPACE;
    charMap['\n'] = SPACE;
    charMap['\t'] = SPACE;

    // alpha characters
    for (int i='a'; i<='z'; ++i) charMap[i] = ALPHA;
    for (int i='A'; i<='Z'; ++i) charMap[i] = ALPHA;
    charMap['_'] = ALPHA;

    // digit characters
    for (int i='0'; i<='9'; ++i) charMap[i] = DIGIT;
  }

//////////////////////////////////////////////////////////////////////////
// Main
//////////////////////////////////////////////////////////////////////////

  public static void main(String[] args)
    throws Exception
  {
    /*
    char[] src = FileUtil.read(new File(args[0]));
    Tokenizer tokenizer = new Tokenizer(null, src, args[0], false);
    Token[] tokens = tokenizer.tokenize();
    for (int i=0; i<tokens.length; ++i)
      System.out.println(tokens[i]);
    */

    /*
    long t1 = System.currentTimeMillis();
    long n1 = System.nanoTime();
    File f = new File("/dev/fan/src/sysTest/fan");
    File[] list = f.listFiles();
    for (int i=0; i<list.length; ++i)
    {
      char[] src = FileUtil.read(list[i]);
      Token[] toks = new Tokenizer(null, src, list[i].getName(), false).tokenize();
      System.out.println(list[i] + " [" + toks.length + "]");
    }
    long t2 = System.currentTimeMillis();
    long n2 = System.nanoTime();
    System.out.println("Time: " + (t2-t1) + "ms");
    System.out.println("Time: " + (n2-n1) + "ns");
    */
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  char[] buf;       // buffer
  int pos;          // index into buf for cur
  boolean isDoc;    // return documentation comments or if false ignore them
  String fileName;  // source file name
  int line = 1;     // current line number
  int col = 1;      // current column number
  int cur;          // current char
  int curt;         // current charMap type
  int peek;         // next char
  int peekt;        // next charMap type
  int lastLine;     // line number of last token returned from next()
  ArrayList tokens; // accumlated tokens
}