//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   2 Dec 05  Brian Frank  Creation
//
package fanx.compiler.steps;

import java.util.*;
import fan.sys.*;
import fan.sys.List;
import fanx.fcode.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.ast.*;
import fanx.compiler.parser.*;

/**
 * Normalize the abstract syntax tree:
 *   - Collapse multiple static new blocks
 *   - Init fields in static new block
 *   - Init fields in instance new block
 *   - Add implicit super() in constructors
 *   - Add implicit return in methods
 */
public class Normalize
  extends CompilerStep
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  public Normalize(Compiler compiler)
  {
    super(compiler);
  }

//////////////////////////////////////////////////////////////////////////
// Normalize ClassDef
//////////////////////////////////////////////////////////////////////////

  public void run()
  {
    for (int i=0; i<compiler.classDefs.length; ++i)
      normalize(compiler.classDefs[i]);
  }

  void normalize(ClassDef cls)
  {
    staticInit(cls);
    instanceInit(cls);

    MethodDef[] methods = cls.methods();
    for (int i=0; i<methods.length; ++i)
      normalize(cls, methods[i]);
  }

//////////////////////////////////////////////////////////////////////////
// Static Initializer
//////////////////////////////////////////////////////////////////////////

  void staticInit(ClassDef cls)
  {
    Block init = new Block(cls);

    // walk thru all the members (fields _and_ methods) in
    // order and assemble a single cinit method definition
    for (int i=0; i<cls.slots.size(); ++i)
    {
      SlotDef def = (SlotDef)cls.slots.get(i);

      // if a static field with initialization, then add
      // it's initialization assignment statement
      if (def instanceof FieldDef)
      {
        FieldDef f = (FieldDef)def;
        if (f.isStatic() && f.init != null)
          init.add(fieldInitStmt(cls, f));
      }

      // if a static initializer, append it
      else
      {
        MethodDef m = (MethodDef)def;
        if (m.isStaticInit())
          init.add(m.code);
      }
    }

    // if init block is empty, then nothing to generate
    if (init.isEmpty()) return;

    // now remove all the individual init methods (from
    // end to start since we are removing indices)
    for (int i=cls.slots.size()-1; i>=0; --i)
    {
      if (cls.slots.get(i) instanceof MethodDef)
      {
        MethodDef def = (MethodDef)cls.slots.get(i);
        if (def.isStaticInit()) cls.slots.remove(i);
      }
    }

    // now add our flattened static init method
    cls.slots.add(MethodDef.makeStaticInit(init));
  }

//////////////////////////////////////////////////////////////////////////
// Instance Initializer
//////////////////////////////////////////////////////////////////////////

  void instanceInit(ClassDef cls)
  {
    Block init = new Block(cls);

    // add instance based initialization
    FieldDef[] defs = cls.instanceFields();
    for (int i=0; i<defs.length; ++i)
    {
      FieldDef def = defs[i];
      if (def.init != null)
        init.add(fieldInitStmt(cls, def));
    }

    // if init block is empty, then nothing to generate
    if (init.isEmpty()) return;

    // now add our flattened instance init method
    cls.slots.add(MethodDef.makeInstanceInit(init));
  }

//////////////////////////////////////////////////////////////////////////
// Field Init
//////////////////////////////////////////////////////////////////////////

  ExprStmt fieldInitStmt(ClassDef cls, FieldDef def)
  {
    if (def.init == null) throw new IllegalStateException();

    ILocation loc = def.location;

    // figure out if target is type or this
    Expr target;
    if (def.isStatic())
      target = new TypeVarExpr(loc, cls.type);
    else
      target = new ThisExpr(loc);

    // lhs is target.field, rhs is init expression
    Expr lhs = new FieldExpr(loc, def.field, target);
    Expr rhs = def.init;

    return new ExprStmt(loc, new BinaryExpr(loc, new Token(Token.ASSIGN), lhs, rhs));
  }

//////////////////////////////////////////////////////////////////////////
// Method
//////////////////////////////////////////////////////////////////////////

  void normalize(ClassDef cls, MethodDef m)
  {
    Block code = m.code;
    if (code == null) return;

    // implicit returns
    checkReturns(m, code);
  }

  void checkReturns(MethodDef m, Block block)
  {
    // TODO - temp hack - we need to evaluate all return paths
    if (block.lastIsMethodExit()) return;

    // add return
    if (!m.ret.type.isVoid()) throw err("Must return value from non-void method", m);
    block.statements.add(new ReturnStmt(block));
  }

}