//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Dec 05  Brian Frank  Creation
//
package fanx.compiler.steps;

import java.io.*;
import java.util.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.ast.*;
import fanx.compiler.parser.*;
import fanx.compiler.resolver.*;

/**
 * TypeInference is the process of infering a variable declaration by
 * the type of the initialization expression.  This compiler step
 * is responsible for top level type inference of FieldDefs:
 *    x := 3    ->  Int x := 3
 *    y := "x"  ->  Str x := 3
 *
 * Note: LocalVarDefStmts also support type inference, but we do that
 *  in ResolveExpr since we have a much better ability to infer fields
 *  and other variables in the block scope.
 */
public class TypeInference
  extends CompilerStep
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  public TypeInference(Compiler compiler)
  {
    super(compiler);
  }

//////////////////////////////////////////////////////////////////////////
// Visitor Callbacks
//////////////////////////////////////////////////////////////////////////

  public void run()
  {
    for (int i=0; i<compiler.classDefs.length; ++i)
      infer(compiler.classDefs[i]);
  }

  void infer(ClassDef cls)
  {
    FieldDef[] fields = cls.fields();
    for (int i=0; i<fields.length; ++i)
      infer(cls, fields[i]);
  }

//////////////////////////////////////////////////////////////////////////
// Infer
//////////////////////////////////////////////////////////////////////////

  void infer(ClassDef cls, FieldDef field)
  {
    if (field.type != null) return;

    if (field.init == null)
      throw err("Cannot infer type without := declaration assignment", field);

    try
    {
      ResolveExpr exprResolver = new ResolveExpr(compiler);
      exprResolver.init(cls);
      exprResolver.walk(field.init);
      field.type = new TypeRef(field.init.location, field.init.type);
    }
    catch (Exception e)
    {
      //e.printStackTrace();
      throw err("Cannot infer type; use explicit type declaration", field);
    }
  }

}