//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.test;

import java.lang.reflect.*;
import fan.sys.*;
import fan.sys.Type;
import fanx.util.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;


/**
 * CompileTest is the base class for tests which feed a source file to
 * the Compiler get back a Class which may be tested via reflection.
 */
public abstract class CompileTest
  extends Test
{

//////////////////////////////////////////////////////////////////////////
// Main
//////////////////////////////////////////////////////////////////////////

  public boolean skip()
  {
    // can't run compiler tests if dynamic class loading not supported
    return Sys.usePrecompiledOnly;
  }

  public final void run()
    throws Exception
  {
    compiler  = new Compiler();
    compiler.log.silent  = true;
    compiler.log.verbose = verbose;
    doRun();
  }

//////////////////////////////////////////////////////////////////////////
// Overrides
//////////////////////////////////////////////////////////////////////////

  public abstract void doRun()
    throws Exception;

//////////////////////////////////////////////////////////////////////////
// Harness
//////////////////////////////////////////////////////////////////////////

  /**
   * Verify no arg method "f" returns the specified result.
   */
  public Object verify(String func, Object result)
    throws Exception
  {
    return verify(func, new Object[0], result);
  }

  /**
   * Verify the source code for method "f" returns the specified result.
   * If the function is static return the generated class, otherwise return
   * the actual instance used (to test for side effects).
   */
  public Object verify(String func, Object[] args, Object result)
    throws Exception
  {
    String src =
    imports + "\n" +
    "class " + genClassName() + "\n" +
    "{\n" +
    members + "\n" +
    "  " + func + "\n" +
    "}\n";

    Class cls = compileToClass(src);
    java.lang.reflect.Method method = findMethod(cls, "f");

    Object instance = Modifier.isStatic(method.getModifiers()) ? null : cls.newInstance();
    Object actual = null;
    try
    {
//long t1 = System.nanoTime();
//for (int i=0; i<10000000; ++i)
      actual = method.invoke(instance, args);
//long t2 = System.nanoTime();
//System.out.println("------> " + (t2-t1)/1000000 + "ms");
    }
    catch (java.lang.reflect.InvocationTargetException e)
    {
      throw (Exception)e.getTargetException();
    }
//System.out.println("-- " + actual  + " ?= " + result);
    verify(actual, result);

    if (instance == null)
      return cls;
    else
      return instance;
  }

  public void verifyErr(String func, String msg)
    throws Exception
  {
    CompileException ex = null;
    try
    {
      verify(func, null, null);
    }
    catch (CompileException e)
    {
      ex = e;
    }
    verify(ex != null);
    if (!ex.msg.equals(msg)) System.out.println(ex + " != " + msg);
    verify(ex.msg.equals(msg));
  }

  public void verifyThrows(String func, Object[] args, Class cls)
    throws Exception
  {
    Exception ex = null;
    try
    {
      verify(func, args, null);
    }
    catch (Exception e)
    {
      ex = e;
    }
//System.out.println(ex + " ?= " + cls);
    verify(ex != null);
    verify(ex.getClass() == cls);
  }

//////////////////////////////////////////////////////////////////////////
// Compile
//////////////////////////////////////////////////////////////////////////

  public Pod compileToPod(String src)
    throws Exception
  {
    compiler.errors.clear();
    return compiler.define(genPodName(), src, "Memory");
  }

  public Type[] compileToTypes(String src)
    throws Exception
  {
    List types = compileToPod(src).types();
    return (Type[])types.toArray(new Type[types.sz()]);
  }

  public Type compileToType(String src)
    throws Exception
  {
    return compileToTypes(src)[0];
  }

  public Class compileToClass(String src)
    throws Exception
  {
    return compileToType(src).emit();
  }

  String genPodName()
  {
    if (nextPodName == null) nextPodName = testName();
    String podName = nextPodName;
    nextPodName = testName() + (podNameCount++);
    return podName;
  }

  String genClassName()
  {
    if (nextClassName == null) nextClassName = "Bar";
    String className = nextClassName;
    nextClassName = "Bar" + (classNameCount++);
    return className;
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  String nextPodName;       // next pod name used
  String nextClassName;     // next class name used
  int classNameCount;       // for unique class naming
  int podNameCount;         // for unique pod naming

  String imports = "";      // imports to add to next class tested
  String members = "";      // other stuff to insert inside class definition
  Compiler compiler;        // compiler currently being tested
  Pod pod;                  // dynamic pod to scope test classes

}