//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.test;

import java.io.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.ast.*;
import fanx.compiler.parser.*;

/**
 * ParserTest
 */
public class ParserTest
  extends Test
{

//////////////////////////////////////////////////////////////////////////
// Main
//////////////////////////////////////////////////////////////////////////

  public void run()
    throws Exception
  {
    verifyClassDef();
    verifyStmt();
  }

//////////////////////////////////////////////////////////////////////////
// ClassDef
//////////////////////////////////////////////////////////////////////////

  public void verifyClassDef()
  {
    // TODO this test doesn't run until base types are
    // resolved with the change to : instead of extends/mixin
    if (true) return;

    String x;
    int line = 1;
    ClassDef cls;

    // simple
    x =
    "class Foo {}";
    cls = parse(x);
    verify(cls.podName.equals("acmePod"));
    verify(cls.name.equals("Foo"));
    verify(cls.qname.equals("acmePod::Foo"));
    verify(cls.base.toString().endsWith("Obj"));
    verify(cls.mixins.size() == 0);

    // extends only
    x =
    "class Foo\n" +
    "  : Bar {}";
    cls = parse(x);
    verify(cls.podName.equals("acmePod"));
    verify(cls.name.equals("Foo"));
    verify(cls.qname.equals("acmePod::Foo"));
    verify(cls.base.toString().equals("Bar?"));
    verify(cls.mixins.size() == 0);

    // extends and mixins
    x =
    "class Foo : Bar, IAlpha, IBeta, IGamma {}";
    cls = parse(x);
    verify(cls.podName.equals("acmePod"));
    verify(cls.name.equals("Foo"));
    verify(cls.qname.equals("acmePod::Foo"));
    verify(cls.base.toString().equals("Bar?"));
    verify(cls.mixins.size() == 3);
    verify(cls.mixins.get(0).toString().equals("IAlpha?"));
    verify(cls.mixins.get(1).toString().equals("IBeta?"));
    verify(cls.mixins.get(2).toString().equals("IGamma?"));

    //
    // errors
    //

    // class
    x = "#!/bin/sh\n\n\n"; line = 4;
    verifyErr(x+"cool",        "Expected 'class', not 'cool'",   line, 1);
    verifyErr(x+"class {",     "Expected identifier, not '{'", line, 7);
    verifyErr(x+"class Foo y", "Expected '{', not 'y'",          line, 11);

    // implements/extends
    /*
    x += "class Foo\n"; line++;
    verifyErr(x+"mixin ~",   "Expected type identifier, not '~'", line, 7);
    verifyErr(x+"mixin X,{", "Expected type identifier, not '{'", line, 9);
    verifyErr(x+": *",      "Expected type identifier, not '*'", line, 9);
    //verifyErr(x+" X,Y",    "Expected '{', not ','",               line, 10);
    */

    // method defs
    x  = "class Foo {\n";
    x += "  Foo() {} }";
    verifyErr(x, "Invalid constructor syntax - use new keyword", 2, 3);
    x  = "class Foo {\n";
    x += "  protected Foo() {} }";
    verifyErr(x, "Invalid constructor syntax - use new keyword", 2, 3);
  }

//////////////////////////////////////////////////////////////////////////
// verifyStmt
//////////////////////////////////////////////////////////////////////////

  public void verifyStmt()
  {
    String err = "Expected end of statement: semicolon, newline, or end of block; not";
    verifyStmt("{ return 5; }");
    verifyStmt("{ return 5 }");
    verifyStmt("{ if (true) return 5 return 6}", err+" 'return'", 5, 22);
    verifyStmt("{ if (true) return null; else return this; }");
    verifyStmt("{ if (true)\nreturn null\nelse\nreturn this\n}");
    verifyStmt("{ if (true) return null else return this}", err+" 'else'", 5, 25);
  }

  void verifyStmt(String source)
  {
    verifyStmt(source, null, 0, 0);
  }

  void verifyStmt(String block, String err, int line, int col)
  {
    String x = "class Foo {\n" +
    "Obj f()\n" + block + "\n" +
    "}\n";

    if (err == null) parse(x);
    else verifyErrNs(x, err, line, col);
  }

//////////////////////////////////////////////////////////////////////////
// Util
//////////////////////////////////////////////////////////////////////////

  public ClassDef parse(String source)
  {
    // always prefix with com.acme namespace
    source = "#class NoWay!/*\n\n\n" + source;

    Compiler compiler = new Compiler();
    compiler.pod = new ShimPod("acmePod");
    CompilationUnit unit = new Parser(compiler, source.toCharArray(), "Test").parse();

    if (verbose) unit.pretty(new AstWriter());

    return unit.topClass();
  }

  public void verifyErrNs(String src, String msg, int line, int col)
  {
    src = "#/user/fan/bin/fan.exe\n\n" + src;
    verifyErr(src, msg, line, col);
  }

  public void verifyErr(String src, String msg, int line, int col)
  {
    CompileException ex = null;
    try
    {
      // parse
      Compiler compiler = new Compiler();
      compiler.log.silent = true;
      compiler.pod = new ShimPod("acmePod");
      new Parser(compiler, src.toCharArray(), "Test").parse();
    }
    catch (CompileException e)
    {
      if (verbose)
      {
        println(e);
        println(msg + " [Test:" + line + ":" + col + "]");
      }
      ex = e;
    }
    verify(ex != null);
    verify(ex.msg.equals(msg));
    verify(ex.location.line() == line);
    verify(ex.location.col() == col);
  }

}