//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   18 Oct 06  Andy Frank  Creation
//

using System;

namespace Fan.Sys
{
  /// <summary>
  /// Duration
  /// </summary>
  public sealed class Duration : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    public static Duration Make(Int ticks) { return Make(ticks.val); }
    public static Duration Make(long ticks)
    {
      if (ticks == 0) return Zero;
      return new Duration(ticks);
    }

    public static Duration now()
    {
      return new Duration(-1); //System.nanoTime()); TODO!!!!
    }

    private Duration(long ticks)
    {
      this.ticks = ticks;
    }

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override bool Equals(Object obj)
    {
      return ticks == (obj as Duration).ticks;
    }

    public override Bool Equals(Obj obj)
    {
      if (obj is Duration)
        return ticks == ((Duration)obj).ticks ? Bool.True : Bool.False;
      else
        return Bool.False;
    }

    public override Int Compare(Obj obj)
    {
      long that = ((Duration)obj).ticks;
      if (ticks < that) return Int.LT; return ticks  == that ? Int.EQ : Int.GT;
    }

    public override int GetHashCode()
    {
      //return (int)(ticks ^ (ticks >>> 32));
      return (int)(ticks ^ (ticks >> 32));
    }

    public override Int Hash()
    {
      return Int.Make(ticks);
    }

    public Int Ticks()
    {
      return Int.Make(ticks);
    }

    public override Type Type()
    {
      return Sys.DurationType;
    }

  //////////////////////////////////////////////////////////////////////////
  // Operators
  //////////////////////////////////////////////////////////////////////////

    public Duration Negate()
    {
      return Make(-ticks);
    }

    public Duration Plus(Duration x)
    {
      return Make(ticks + x.ticks);
    }

    public Duration Minus(Duration x)
    {
      return Make(ticks - x.ticks);
    }

    public Duration Star(Float x)
    {
      return Make((long)(ticks * x.val));
    }

    public Duration Slash(Float x)
    {
      return Make((long)(ticks / x.val));
    }

    public Duration Percent(Float x)
    {
      return Make((long)(ticks % x.val));
    }

  //////////////////////////////////////////////////////////////////////////
  // Conversion
  //////////////////////////////////////////////////////////////////////////

    public override Str ToStr()
    {
      if (ticks == 0) return ZeroStr;

      // if clean millisecond boundary
      long ns = ticks;
      if (ns % 1000000L == 0)
      {
        if (ns % 86400000000000L == 0)
        {
          return Str.Make(ns/86400000000000L + "days");
        }
        if (ns % 3600000000000L == 0)
        {
          return Str.Make(ns/3600000000000L + "hr");
        }
        if (ns % 60000000000L == 0)
        {
          return Str.Make(ns/60000000000L + "min");
        }
        if (ns % 1000000000L == 0)
        {
          return Str.Make(ns/1000000000L + "sec");
        }
        return Str.Make(ns/1000000L + "ms");
      }

      // return in nanoseconds
      return Str.Make(ns + "ns");
    }

    public Int toMillis()
    {
      return Int.Make(ticks/1000000L);
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public static readonly Duration Zero = new Duration(0);
    public static readonly Str ZeroStr = Str.Make("0ns");

    public readonly long ticks;

  }
}