//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   13 Sep 06  Andy Frank  Creation
//

using System;

namespace Fan.Sys
{
  ///
  /// Float is a 64-bit floating point value.
  ///
  public sealed class Float : Num
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    public static Float Parse(Str s) { return Parse(s, Bool.True); }
    public static Float Parse(Str s, Bool check)
    {
      try
      {
        string x = s.val;
        if (x == "NaN")  return NaN;
        if (x == "+Inf") return PosInf;
        if (x == "-Inf") return NegInf;
        return Make(Double.Parse(x));
      }
      catch (FormatException)
      {
        if (!check.val) return null;
        throw ArgErr.Make("Invalid Float: '" + s + "'").val;
      }
    }

    public static Float Make(double val)
    {
      if (val == 0) return Zero;
      return new Float(val);
    }

    public static Float MakeBits(Int bits)
    {
      return Make(BitConverter.DoubleToInt64Bits(bits.val));
    }

    /*
    public static Float makeBits32(Int bits)
    {
      return Make(BitConverter.FloatToInt32Bits((int)bits.val));
    }
    */

    /*
    public static Float PosInf() { return PosInf; }

    public static Float NegInf() { return NegInf; }

    public static Float NaN() { return NaN; }

    public static Float Pi() { return Pi; }

    public static Float E() { return E; }
    */

    private Float(double val) { this.val = val; }

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override bool Equals(Object obj)
    {
      return val == (obj as Float).val;
    }

    public override Bool Equals(Obj obj)
    {
      if (obj is Float)
      {
        double x = ((Float)obj).val;
        if (Double.IsNaN(val)) return Bool.Make(Double.IsNaN(x));
        return val == x ? Bool.True : Bool.False;
      }
      return Bool.False;
    }

    /*
    public Bool Approx(Float that) { return approx(that, null); }
    public Bool Approx(Float that, Float tolerance)
    {
      // need this to check +inf, -inf, and nan
      if (equals(that).val) return Bool.True;

      double t;
      if (tolerance == null)
        t = Math.min( Math.abs(val/1e6), Math.abs(that.val/1e6) );
      else
        t = tolerance.val;
      return Math.abs(val - that.val) <= t ? Bool.True : Bool.False;
    }
    */

    public override Int Compare(Obj obj)
    {
      double that = ((Float)obj).val;
      if (Double.IsNaN(val))
      {
        return (Double.IsNaN(that)) ? Int.EQ : Int.LT;
      }
      else if (Double.IsNaN(that))
      {
        return Int.GT;
      }
      else
      {
        if (val < that) return Int.LT; return val == that ? Int.EQ : Int.GT;
      }
    }

    public override int GetHashCode()
    {
      long hash = BitConverter.DoubleToInt64Bits(val);
      //return (int)(hash ^ (hash >>> 32));
      return (int)(hash ^ (hash >> 32));
    }

    public override Int Hash()
    {
      return Bits();
    }

    public Int Bits()
    {
      return Int.Make(BitConverter.DoubleToInt64Bits(val));
    }

    /*
    public Int Bits32()
    {
      return Int.make(BitConverter.FloatToInt32Bits((float)val) & 0xFFFFFFFFL);
    }
    */

    public override Type Type()
    {
      return Sys.FloatType;
    }

  //////////////////////////////////////////////////////////////////////////
  // Operators
  //////////////////////////////////////////////////////////////////////////

    public Float Negate() { return Make(-val); }
    public Float Star      (Float x) { return Make(val * x.val); }
    public Float Slash     (Float x) { return Make(val / x.val); }
    public Float Percent   (Float x) { return Make(val % x.val); }
    public Float Plus      (Float x) { return Make(val + x.val); }
    public Float Minus     (Float x) { return Make(val - x.val); }
    public Float Increment ()        { return Make(val+1); }
    public Float Decrement ()        { return Make(val-1); }

  //////////////////////////////////////////////////////////////////////////
  // Num
  //////////////////////////////////////////////////////////////////////////

    public override Int ToInt() { return Int.Make((long)val); }

    public override Float ToFloat() { return this; }

  //////////////////////////////////////////////////////////////////////////
  // Math
  //////////////////////////////////////////////////////////////////////////

    /*
    public Float Abs()
    {
      if (val >= 0) return this;
      return make(-val);
    }

    public Float min(Float that)
    {
      if (val <= that.val) return this;
      return that;
    }

    public Float max(Float that)
    {
      if (val >= that.val) return this;
      return that;
    }

    public Float ceil()
    {
      return make(Math.ceil(val));
    }

    public Float floor()
    {
      return make(Math.floor(val));
    }

    public Float round()
    {
      return make(Math.rint(val));
    }

    public Float exp()
    {
      return make(Math.exp(val));
    }

    public Float log()
    {
      return make(Math.log(val));
    }

    public Float log10()
    {
      return make(Math.log10(val));
    }

    public Float pow(Float pow)
    {
      return make(Math.pow(val, pow.val));
    }

    public Float sqrt()
    {
      return make(Math.sqrt(val));
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Trig
  //////////////////////////////////////////////////////////////////////////

    /*
    public Float acos()
    {
      return make(Math.acos(val));
    }

    public Float asin()
    {
      return make(Math.asin(val));
    }

    public Float atan()
    {
      return make(Math.atan(val));
    }

    public static Float atan2(Float y, Float x)
    {
      return make(Math.atan2(y.val, x.val));
    }

    public Float cos()
    {
      return make(Math.cos(val));
    }

    public Float cosh()
    {
      return make(Math.cosh(val));
    }

    public Float sin()
    {
      return make(Math.sin(val));
    }

    public Float sinh()
    {
      return make(Math.sinh(val));
    }

    public Float tan()
    {
      return make(Math.tan(val));
    }

    public Float tanh()
    {
      return make(Math.tanh(val));
    }

    public Float toDegrees()
    {
      return make(Math.toDegrees(val));
    }

    public Float toRadians()
    {
      return make(Math.toRadians(val));
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Conversion
  //////////////////////////////////////////////////////////////////////////

    public override Str ToStr()
    {
      if (Double.IsNaN(val)) return NaNStr;
      if (val == Double.PositiveInfinity) return PosInfStr;
      if (val == Double.NegativeInfinity) return NegInfStr;
      string s = "" + val;
      if (s.IndexOf('.') == -1) s += ".0";  // to match java behavior
      return Str.Make(s);
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public static readonly Float Zero   = new Float(0);
    public static readonly Float PosInf = new Float(Double.PositiveInfinity);
    public static readonly Float NegInf = new Float(Double.NegativeInfinity);
    public static readonly Float NaN    = new Float(Double.NaN);
    public static readonly Float E      = new Float(Math.E);
    public static readonly Float Pi     = new Float(Math.PI);
    public static readonly Str PosInfStr = Str.Make("+Inf");
    public static readonly Str NegInfStr = Str.Make("-Inf");
    public static readonly Str NaNStr    = Str.Make("NaN");

    public readonly double val;
  }
}