//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Sep 06  Andy Frank  Creation
//

using System;
using System.Diagnostics;
using Fanx.Util;

namespace Fan.Sys
{
  /// <summary>
  /// List.
  /// </summary>
  public class List : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////

    public static List Make(Type of, Int capacity)
    {
      return new List(of, (int)capacity.val);
    }

    public static List MakeObj(Int capacity)
    {
      return new List(Sys.ObjType, (int)capacity.val);
    }

    public List(Type of, Obj[] values)
    {
      if (of == null) { Console.WriteLine(new StackTrace(true)); throw new NullErr().val; }
      this.of = of;
      this.values = values;
      this.size = values.Length;
    }

    public List(Type of, int capacity)
    {
      if (of == null) { Console.WriteLine(new StackTrace(true)); throw new NullErr().val; }
      this.of = of;
      this.values = capacity == 0 ? empty : new Obj[capacity];
    }

    public List(Type of)
    {
      if (of == null) { Console.WriteLine(new StackTrace(true)); throw new NullErr().val; }
      this.of = of;
      this.values = empty;
    }

    /*
    public List(Type of, Collection collection)
    {
      if (of == null) { Thread.dumpStack(); throw new NullErr().val; }
      this.of = of;
      this.size = collection.size();
      this.values = (Obj[])collection.toArray(new Obj[size]);
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override Type Type()
    {
      return of.ToListOf();
    }

  //////////////////////////////////////////////////////////////////////////
  // Access
  //////////////////////////////////////////////////////////////////////////

    public Bool IsEmpty()
    {
      return size == 0 ? Bool.True : Bool.False;
    }

    public Int Size()
    {
      return Int.Pos(size);
    }

    /*
    public void Size(Int s)
    {
      Modify();
      int newSize = (int)s.val;
      if (newSize > size)
      {
        Obj[] temp = new Obj[newSize];
        Array.Copy(values, temp, size);
        values = temp;
        size = newSize;
      }
      else
      {
        Obj[] temp = new Obj[newSize];
        Array.Copy(values, temp, size);
        values = temp;
        size = newSize;
      }
    }
    */

    public Int Capacity()
    {
      return Int.Pos(values.Length);
    }

    /*
    public void Capacity(Int c)
    {
      modify();
      int newCapacity = (int)c.val;
      if (newCapacity < size) throw ArgErr.Make("capacity < size").val;
      Obj[] temp = new Obj[newCapacity];
      Array.Copy(values, temp, size);
      values = temp;
    }
    */

    public Obj Get(Int index)
    {
      try
      {
        int i = (int)index.val;
        if (i < 0) i = size + i;
        if (i >= size) throw IndexErr.Make(index).val;
        return values[i];
      }
      catch (IndexOutOfRangeException)
      {
        throw IndexErr.Make(index).val;
      }
    }

    /*
    public final List slice(Range r)
    {
      try
      {
        int s = r.start(size);
        int e = r.end(size);
        int n = e - s + 1;
        if (n < 0) throw IndexErr.Make(r).val;

        List acc = new List(of, n);
        acc.size = n;
        System.arraycopy(values, s, acc.values, 0, n);
        return acc;
      }
      catch (IndexOutOfRangeException)
      {
        throw IndexErr.Make(r).val;
      }
    }
    */

    public Bool Contains(Obj val)
    {
      return Bool.Make(Index(val) != null);
    }

    public Int Index(Obj val) { return Index(val, Int.Zero); }
    public Int Index(Obj val, Int off)
    {
      if (size == 0) return null;
      int start = (int)off.val;
      if (start < 0) start = size + start;
      if (start >= size) throw IndexErr.Make(off).val;

      try
      {
        if (val == null)
        {
          for (int i=start; i<size; i++)
            if (values[i] == null)
              return Int.Pos(i);
        }
        else
        {
          for (int i=start; i<size; i++)
          {
            Obj obj = values[i];
            if (obj != null && obj.Equals(val).val)
              return Int.Pos(i);
          }
        }
        return null;
      }
      catch (IndexOutOfRangeException)
      {
        throw IndexErr.Make(off).val;
      }
    }

    public Obj First()
    {
      if (size == 0) return null;
      return values[0];
    }

    public Obj Last()
    {
      if (size == 0) return null;
      return values[size-1];
    }

    public override Int Hash()
    {
      long hash = 33;
      for (int i=0; i<size; ++i)
      {
        Obj obj = values[i];
        if (obj != null) hash ^= obj.Hash().val;
      }
      return Int.Make(hash);
    }

    public override Bool Equals(Obj that)
    {
      if (that is List)
      {
        List x = (List)that;
        if (!of.Equals(x.of).val) return Bool.False;
        if (size != x.size) return Bool.False;
        for (int i=0; i<size; ++i)
          if (!OpUtil.CompareEQ(values[i], x.values[i]).val) return Bool.False;
        return Bool.True;
      }
      return Bool.False;
    }

  //////////////////////////////////////////////////////////////////////////
  // Modification
  //////////////////////////////////////////////////////////////////////////

    public List Set(Int index, Obj val)
    {
      Modify();
      try
      {
        int i = (int)index.val;
        if (i < 0) i = size + i;
        if (i >= size) throw IndexErr.Make(index).val;
        values[i] = val;
        return this;
      }
      catch (IndexOutOfRangeException)
      {
        throw IndexErr.Make(index).val;
      }
    }

    public List Add(Obj val)
    {
      // modify in Insert(int, Obj)
      return Insert(size, val);
    }

    /*
    public List AddAll(List list)
    {
      // modify in InsertAll(int, List)
      return insertAll(size, list);
    }
    */

    public List Insert(Int index, Obj val)
    {
      // modify in Insert(int, Obj)
      int i = (int)index.val;
      if (i < 0) i = size + i;
      if (i > size) throw IndexErr.Make(index).val;
      return Insert(i, val);
    }

    private List Insert(int i, Obj val)
    {
      Modify();
      if (values.Length <= size)
        Grow(size+1);
      if (i < size)
        Array.Copy(values, i, values, i+1, size-i);
      values[i] = val;
      size++;
      return this;
    }

    /*
    public List InsertAll(Int index, List list)
    {
      // modify in insertAll(int, List)
      int i = (int)index.val;
      if (i < 0) i = size + i;
      if (i > size) throw IndexErr.Make(index).val;
      return insertAll(i, list);
    }

    private List insertAll(int i, List list)
    {
      modify();
      if (list.size == 0) return this;
      if (values.length < size+list.size)
        grow(size+list.size);
      if (i < size)
        System.arraycopy(values, i, values, i+list.size, size-i);
      System.arraycopy(list.values, 0, values, i, list.size);
      size+=list.size;
      return this;
    }
    */

    public Obj Remove(Obj val)
    {
      // modify in removeAt(Int)
      Int index = Index(val);
      if (index.val < 0) return null;
      return RemoveAt(index);
    }

    public Obj RemoveAt(Int index)
    {
      Modify();
      int i = (int)index.val;
      if (i < 0) i = size + i;
      if (i >= size) throw IndexErr.Make(index).val;
      Obj old = values[i];
      if (i < size-1)
        Array.Copy(values, i+1, values, i, size-i-1);
      size--;
      return old;
    }

    private void Grow(int desiredSize)
    {
      int desired = (int)desiredSize;
      if (desired < 1) throw Err.Make("desired " + desired + " < 1").val;
      int newSize = Math.Max(desired, size*2);
      if (newSize < 10) newSize = 10;
      Obj[] temp = new Obj[newSize];
      Array.Copy(values, temp, size);
      values = temp;
    }

    public List Trim()
    {
      Modify();
      if (size == 0)
      {
        values = empty;
      }
      else
      {
        Obj[] temp = new Obj[size];
        Array.Copy(values, temp, size);
        values = temp;
      }
      return this;
    }

    public List Clear()
    {
      Modify();
      for (int i=0; i<size; ++i)
        values[i] = null;
      size = 0;
      return this;
    }

  //////////////////////////////////////////////////////////////////////////
  // Stack
  //////////////////////////////////////////////////////////////////////////

    /*
    public Obj Peek()
    {
      if (size == 0) return null;
      return values[size-1];
    }

    public Obj Pop()
    {
      // modify in removeAt()
      if (size == 0) return null;
      return RemoveAt(Int.NegOne);
    }

    public List Push(Obj obj)
    {
      // modify in add()
      return Add(obj);
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Iterators
  //////////////////////////////////////////////////////////////////////////

    /*
    public final void each(Method m)
    {
      for (int i=0; i<size; ++i)
        m.call2(values[i], Int.pos(i));
    }

    public final void eachr(Method m)
    {
      for (int i=size-1; i>=0; --i)
        m.call2(values[i], Int.pos(i));
    }

    public final Obj find(Method m)
    {
      for (int i=0; i<size; ++i)
        if (m.call2(values[i], Int.pos(i)) == Bool.True)
          return values[i];
      return null;
    }

    public final List findAll(Method m)
    {
      List acc = new List(of, size);
      for (int i=0; i<size; ++i)
        if (m.call2(values[i], Int.pos(i)) == Bool.True)
          acc.add(values[i]);
      return acc;
    }

    public final List findType(Type t)
    {
      List acc = new List(t, size);
      for (int i=0; i<size; ++i)
      {
        Obj item = values[i];
        if (item != null && item.type().is(t))
          acc.add(item);
      }
      return acc;
    }

    public final List exclude(Method m)
    {
      List acc = new List(of, size);
      for (int i=0; i<size; ++i)
        if (m.call2(values[i], Int.pos(i)) != Bool.True)
          acc.add(values[i]);
      return acc;
    }

    public final Bool any(Method m)
    {
      for (int i=0; i<size; ++i)
        if (m.call2(values[i], Int.pos(i)) == Bool.True)
          return Bool.True;
      return Bool.False;
    }

    public final Bool all(Method m)
    {
      for (int i=0; i<size; ++i)
        if (m.call2(values[i], Int.pos(i)) != Bool.True)
          return Bool.False;
      return Bool.True;
    }

    public final Obj reduce(Obj reduction, Method m)
    {
      for (int i=0; i<size; ++i)
        reduction = m.call3(reduction, values[i], Int.pos(i));
      return reduction;
    }

    public final List map(List acc, Method m)
    {
      if (acc.size == 0) acc.capacity(size());
      for (int i=0; i<size; ++i)
        acc.add(m.call2(values[i], Int.pos(i)));
      return acc;
    }

    public final Obj max() { return max(null); }
    public final Obj max(Method m)
    {
      if (size == 0) return null;
      Comparator c = toComparator(m);
      Obj max = values[0];
      for (int i=1; i<size; ++i)
        if (c.compare(values[i], max) > 0)
          max = values[i];
      return max;
    }

    public final Obj min() { return min(null); }
    public final Obj min(Method m)
    {
      if (size == 0) return null;
      Comparator c = toComparator(m);
      Obj min = values[0];
      for (int i=1; i<size; ++i)
        if (c.compare(values[i], min) < 0)
          min = values[i];
      return min;
    }

    public final List unique()
    {
      HashMap dups = new HashMap(size*3);
      List acc = new List(of, size);
      for (int i=0; i<size; ++i)
      {
        Obj v = values[i];
        if (dups.get(v) == null)
        {
          dups.put(v, v);
          acc.add(v);
        }
      }
      return acc;
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Utils
  //////////////////////////////////////////////////////////////////////////

    /*
    public final List sort() { return sort(null); }
    public final List sort(final Method m)
    {
      modify();
      Arrays.sort(values, 0, size, toComparator(m));
      return this;
    }

    public final List sortr() { return sortr(null); }
    public final List sortr(final Method m)
    {
      modify();
      Arrays.sort(values, 0, size, toReverseComparator(m));
      return this;
    }

    public final List reverse()
    {
      modify();
      Obj[] values = this.values;
      int size = this.size;
      int mid   = size/2;
      for (int i=0; i<mid; ++i)
      {
        Obj a = values[i];
        Obj b = values[size-i-1];
        values[i] = b;
        values[size-i-1] = a;
      }
      return this;
    }

    public final List swap(Int a, Int b)
    {
      // modify in set()
      Obj temp = get(a);
      set(a, get(b));
      set(b, temp);
      return this;
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Conversion
  //////////////////////////////////////////////////////////////////////////

    /*
    public final Str join() { return join(Str.Empty, null); }
    public final Str join(Str sep) { return join(sep, null); }
    public final Str join(Str sep, Method m)
    {
      StringBuilder s = new StringBuilder(32+size*32);
      for (int i=0; i<size; ++i)
      {
        if (i > 0) s.append(sep.val);
        if (m == null)
          s.append(values[i]);
        else
          s.append(m.call2(values[i], Int.pos(i)));
      }
      return Str.Make(s.toString());
    }

    public final Str toStr()
    {
      if (size == 0) return Str.Make("[,]");
      StringBuilder s = new StringBuilder(32+size*32);
      s.append("[");
      for (int i=0; i<size; ++i)
      {
        if (i > 0) s.append(", ");
        s.append(values[i]);
      }
      s.append("]");
      return Str.Make(s.toString());
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Runtime Utils
  //////////////////////////////////////////////////////////////////////////

    public int Sz()
    {
      return size;
    }

    public Obj Get(int i)
    {
      try
      {
        if (i >= size) throw IndexErr.Make(""+i).val;
        return values[i];
      }
      catch (IndexOutOfRangeException)
      {
        throw IndexErr.Make(""+i).val;
      }
    }

    public object[] ToArray()
    {
      if (values.Length == size) return values;
      object[] r = new object[size];
      Array.Copy(values, 0, r, 0, size);
      return r;
    }

    public object[] ToArray(object[] a)
    {
      try
      {
        Array.Copy(values, 0, a, 0, size);
        return a;
      }
      catch (IndexOutOfRangeException)
      {
        throw IndexErr.Make().val;
      }
    }

    public object[] ToArray(object[] a, int start, int len)
    {
      try
      {
        Array.Copy(values, start, a, 0, len);
        return a;
      }
      catch (IndexOutOfRangeException)
      {
        throw IndexErr.Make().val;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Comparators
  //////////////////////////////////////////////////////////////////////////
    /*
    static Comparator toComparator(final Method m)
    {
      if (m == null) return defaultComparator;
      return new Comparator()
      {
        public int compare(Object a, Object b) { return (int)((Int)m.call2((Obj)a, (Obj)b)).val; }
      };
    }
    static final Comparator defaultComparator = new Comparator()
    {
      public int compare(Object a, Object b) { return (int)OpUtil.compare((Obj)a, (Obj)b).val; }
    };

    static Comparator toReverseComparator(final Method m)
    {
      if (m == null) return defaultReverseComparator;
      return new Comparator()
      {
        public int compare(Object a, Object b) { return (int)((Int)m.call2((Obj)b, (Obj)a)).val; }
      };
    }
    static final Comparator defaultReverseComparator = new Comparator()
    {
      public int compare(Object a, Object b) { return (int)OpUtil.compare((Obj)b, (Obj)a).val; }
    };
    */
  //////////////////////////////////////////////////////////////////////////
  // Readonly
  //////////////////////////////////////////////////////////////////////////

    public Bool IsRW()
    {
      return isReadonly ? Bool.False : Bool.True;
    }

    public Bool IsRO()
    {
      return isReadonly ? Bool.True : Bool.False;
    }

    public List RW()
    {
      if (!isReadonly) return this;

      Obj[] temp = new Obj[size];
      Array.Copy(values, temp, size);

      List rw = new List(of);
      rw.values       = temp;
      rw.size         = size;
      rw.isReadonly   = false;
      rw.readonlyList = this;
      return rw;
    }

    public List RO()
    {
      if (isReadonly) return this;
      if (readonlyList == null)
      {
        List ro = new List(of);
        ro.values     = values;
        ro.size       = size;
        ro.isReadonly = true;
        readonlyList  = ro;
      }
      return readonlyList;
    }

    private void Modify()
    {
      // if readonly then throw readonly exception
      if (isReadonly)
        throw ReadonlyErr.Make("List is readonly").val;

      // if we have a cached readonlyList, then detach
      // it so it remains immutable
      if (readonlyList != null)
      {
        Obj[] temp = new Obj[size];
        Array.Copy(values, temp, size);
        readonlyList.values = temp;
        readonlyList = null;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    private static readonly Obj[] empty = new Obj[0];

    private Type of;
    private Obj[] values;
    private int size;
    private bool isReadonly;
    private List readonlyList;

  }
}