//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Dec 06  Andy Frank  Creation
//

using System.Text;

namespace Fan.Sys
{
  /// <summary>
  /// MethodType is a parameterized type for Methods.
  /// </summary>
  public class MethodType : GenericType
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    public MethodType(Type[] pars, Type ret) : base(Sys.MethodType)
    {
      this.pars = pars;
      this.ret  = ret;

      // I am a generic parameter type if any my args or
      // return type are generic parameter types.
      this.genericParameterType |= ret.IsGenericParameter();
      for (int i=0; i<pars.Length; i++)
        this.genericParameterType |= pars[i].IsGenericParameter();
    }

  //////////////////////////////////////////////////////////////////////////
  // Type
  //////////////////////////////////////////////////////////////////////////

    public override Int Hash()
    {
      return Signature().Hash();
    }

    public override Bool Equals(Obj obj)
    {
      if (obj is MethodType)
      {
        MethodType x = (MethodType)obj;
        if (pars.Length != x.pars.Length) return Bool.False;
        for (int i=0; i<pars.Length; i++)
          if (!pars[i].Equals(x.pars[i]).val) return Bool.False;
        return ret.Equals(x.ret);
      }
      return Bool.False;
    }

    public override Type Base()
    {
      return Sys.MethodType;
    }

    public override Str Signature()
    {
      if (sig == null)
      {
        StringBuilder s = new StringBuilder();
        s.Append('|');
        for (int i=0; i<pars.Length; i++)
        {
          if (i > 0) s.Append(',');
          s.Append(pars[i].Signature().val);
        }
        s.Append('-').Append('>');
        s.Append(ret.Signature().val);
        s.Append('|');
        sig = Str.Make(s.ToString());
      }
      return sig;
    }

    public override bool Is(Type type)
    {
      if (this == type) return true;
      if (type is MethodType)
      {
        MethodType t = (MethodType)type;

        // match return type
        if (!ret.Is(t.ret)) return false;

        // match pars - it is ok for me to have less than
        // the type pars (if I want to ignore them), but I
        // must have no more
        if (pars.Length > t.pars.Length) return false;
        for (int i=0; i<pars.Length; i++)
          if (!t.pars[i].Is(pars[i])) return false;

        // this method works for the specified method type
        return true;
      }
      return Base().Is(type);
    }

    internal override Map MakeParams()
    {
      Map map = new Map(Sys.StrType, Sys.TypeType);
      for (int i=0; i<pars.Length; i++)
        map.Set(Str.ascii['A'+i], pars[i]);
      return map.Set(Str.ascii['R'], ret).RO();
    }

  //////////////////////////////////////////////////////////////////////////
  // GenericType
  //////////////////////////////////////////////////////////////////////////

    public override Type GetRawType()
    {
      return Sys.MethodType;
    }

    public override bool IsGenericParameter()
    {
      return genericParameterType;
    }

    protected override Type DoParameterize(Type t)
    {
      // return
      if (t == Sys.RType) return ret;

      // if A-H maps to avail pars
      int name = t.name.val[0] - 'A';
      if (name < pars.Length) return pars[name];

      // otherwise let anything be used
      return Sys.ObjType;
    }

  //////////////////////////////////////////////////////////////////////////
  // Method Support
  //////////////////////////////////////////////////////////////////////////

    internal List ToMethodParams()
    {
      MethodParam[] p = new MethodParam[pars.Length];
      for (int i=0; i<p.Length; i++)
        p[i] = new MethodParam(Str.Make("p" + i), pars[i], 0);
      return new List(Sys.MethodParamType, p);
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public readonly Type[] pars;
    public readonly Type ret;
    private Str sig;
    private bool genericParameterType;
  }
}