//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   14 Sep 06  Andy Frank  Creation
//

using System;
using System.Collections;
using System.IO;
using System.Reflection;
using System.Runtime.CompilerServices;
using Fanx.Emit;
using Fanx.Fcode;
using Fanx.Util;

namespace Fan.Sys
{
  ///
  /// Type models a static type definition for an Obj class.  A Type lifecycle:
  ///
  ///  1) Hollow: in this state we know basic identity of the type, and
  ///     it's inheritance hierarchy.  A type is setup to be hollow during
  ///     Pod.load().
  ///  2) Reflected: in this state we read all the slot definitions from the
  ///     fcode to populate the slot tables used to for reflection.  At this
  ///     point clients can discover the signatures of the Type.
  ///  3) Emitted: the final state of loading a Type is to emit to a Java
  ///     class called "fan.{pod}.{type}".  Once emitted we can instantiate
  ///     the type or call it's methods.
  ///  4) Finished: once we have reflected the slots into memory and emitted
  ///     the Java class, the last stage is to bind the all the java.lang.reflect
  ///     representations to the Slots for dynamic dispatch.  We delay this
  ///     until needed by Method or Field for a reflection invocation
  ///
  ///  A special stub state is also used to bootstrap the compiler self tests
  ///  and to compile sys itself.  Stub types use java reflection to map their
  ///  Java API to their Fan API (rather crudely).
  ///
  public class Type : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Management
  //////////////////////////////////////////////////////////////////////////

    public static Type Find(Str sig) { return TypeParser.Load(sig.val, true, null); }
    public static Type Find(String sig) { return TypeParser.Load(sig, true, null); }
    public static Type Find(Str sig, Bool check) { return TypeParser.Load(sig.val, check.val, null); }
    public static Type Find(String sig, bool check) { return TypeParser.Load(sig, check, null); }
    public static Type Find(String podName, String typeName, bool check)
    {
      Pod pod = Fan.Sys.Pod.Find(podName, check, null);
      if (pod == null) return null;
      return pod.FindType(typeName, check);
    }

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    internal Type(Pod pod, FType ftype)
    {
      this.pod    = pod;
      this.ftype  = ftype;
      this.name   = Str.Make(pod.fpod.Name(pod.fpod.TypeRef(ftype.selft).typeName));
      this.qname  = Str.Make(pod.name.val + "::" + name.val);
      this.flags  = ftype.flags;
      if (Debug) Console.WriteLine("-- init:   " + qname);
    }

    // ShimType and parameterized type constructor
    public Type(Pod pod, String name, int flags, Facets facets)
    {
      this.pod    = pod;
      this.name   = Str.Make(name);
      this.qname  = Str.Make(pod.name.val + "::" + name);
      this.flags  = flags;
      this.facets = facets;
    }

  //////////////////////////////////////////////////////////////////////////
  // Naming
  //////////////////////////////////////////////////////////////////////////

    //public Type Type() { return Sys.TypeType; }

    public Pod Pod()   { return pod; }
    public Str Name()  { return name; }
    public Str Qname() { return qname; }
    public virtual Str Signature() { return qname; }

  //////////////////////////////////////////////////////////////////////////
  // Flags
  //////////////////////////////////////////////////////////////////////////

    public Bool IsAbstract() { return Bool.Make(flags & FConst.Abstract); }
    public Bool IsClass() { return Bool.Make((flags & (FConst.Enum|FConst.Mixin)) == 0); }
    public Bool IsEnum() { return Bool.Make(flags & FConst.Enum); }
    public Bool IsFinal() { return Bool.Make(flags & FConst.Final); }
    public Bool IsInternal() { return Bool.Make(flags & FConst.Internal); }
    public Bool IsMixin() { return Bool.Make(flags & FConst.Mixin); }
    public Bool IsPublic() { return Bool.Make(flags & FConst.Public); }
    public Bool IsSynthetic() { return Bool.Make(flags & FConst.Synthetic); }

  //////////////////////////////////////////////////////////////////////////
  // Generics
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// A generic type means that one or more of my slots contain signatures
    /// using a generic parameter (such as V or K).  Fan supports three built-in
    /// generic types: List, Map, and Method.  A generic instance (such as Str[])
    /// is NOT a generic type (all of its generic parameters have been filled in).
    /// User defined generic types are not supported in Fan.
    /// </summary>
    public bool IsGenericType()
    {
      return this == Sys.ListType || this == Sys.MapType || this == Sys.MethodType;
    }

    /// <summary>
    /// A generic instance is a type which has "instantiated" a generic type
    /// and replaced all the generic parameter types with generic argument
    /// types.  The type Str[] is a generic instance of the generic type
    /// List (V is replaced with Str).  A generic instance always has a signature
    /// which different from the qname.
    /// </summary>
    public virtual bool IsGenericInstance()
    {
      return false;
    }

    /// <summary>
    /// Return if this type is a generic parameter (such as V or K) in a
    /// generic type (List, Map, or Method).  Generic parameters serve
    /// as place holders for the parameterization of the generic type.
    /// Fan has a predefined set of generic parameters which are always
    /// defined in the sys pod with a one character name.
    /// </summary>
    public virtual bool IsGenericParameter()
    {
      return pod == Sys.SysPod && name.val.Length == 1;
    }

    /// <summary>
    /// If this type is a generic parameter (V, L, etc), then return
    /// the actual type used in the Java method.  For example V is Obj,
    /// and L is List.  This is the type we actually use when constructing
    /// a signature for the invoke opcode.
    /// </summary>
    public virtual Type GetRawType()
    {
      if (!IsGenericParameter()) return this;
      if (this == Sys.LType)  return Sys.ListType;
      if (this == Sys.MType)  return Sys.MapType;
      if (this is ListType)   return Sys.ListType;
      if (this is MapType)    return Sys.MapType;
      if (this is MethodType) return Sys.MethodType;
      return Sys.ObjType;
    }

    public Bool IsGeneric()
    {
      return IsGenericType() ? Bool.True : Bool.False;
    }

    /*
    public Map Pars()
    {
      if (noParams == null)
        noParams = new Map(Sys.StrType, Sys.TypeType).ro();
      return (Map)noParams;
    }

    public Type Parameterize(Map pars)
    {
      if (this == Sys.ListType)
      {
        Type v = (Type)params.get(Str.ascii['V']);
        if (v == null) throw ArgErr.make("List.parameterize - V undefined").val;
        return v.toListOf();
      }

      if (this == Sys.MapType)
      {
        Type v = (Type)params.get(Str.ascii['V']);
        Type k = (Type)params.get(Str.ascii['K']);
        if (v == null) throw ArgErr.make("Map.parameterize - V undefined").val;
        if (k == null) throw ArgErr.make("Map.parameterize - K undefined").val;
        return new MapType(k, v);
      }

      if (this == Sys.MethodType)
      {
        Type r = (Type)params.get(Str.ascii['R']);
        if (r == null) throw ArgErr.make("Map.parameterize - R undefined").val;
        ArrayList p = new ArrayList();
        for (int i='A'; i<='H'; ++i)
        {
          Type x = (Type)params.get(Str.ascii[i]);
          if (x == null) break;
          p.add(x);
        }
        return new MethodType((Type[])p.toArray(new Type[p.size()]), r);
      }

      throw UnsupportedErr.make("not generic: " + this).val;
    }
    */

    [MethodImpl(MethodImplOptions.Synchronized)]
    public Type ToListOf()
    {
      if (listOf == null) listOf = new ListType(this);
      return listOf;
    }

  //////////////////////////////////////////////////////////////////////////
  // Slots
  //////////////////////////////////////////////////////////////////////////

    public List Fields()  { return Reflect().fields; }
    public List Methods() { return Reflect().methods; }
    public List Slots()   { return Reflect().slots; }

    public Field Field(Str name) { return (Field)Slot(name.val, true); }
    public Field Field(Str name, Bool check) { return (Field)Slot(name.val, check.val); }
    public Field Field(string name, bool check) { return (Field)Slot(name, check); }

    public Method Method(Str name) { return (Method)Slot(name.val, true); }
    public Method Method(Str name, Bool check) { return (Method)Slot(name.val, check.val); }
    public Method Method(string name, bool check) { return (Method)Slot(name, check); }

    public Slot Slot(Str name) { return Slot(name.val, true); }
    public Slot Slot(Str name, Bool check) { return Slot(name.val, check.val); }
    public Slot Slot(string name, bool check)
    {
      Slot slot = (Slot)Reflect().slotsByName[name];
      if (slot != null) return slot;
      if (check) throw UnknownSlotErr.Make(this.qname + "." + name).val;
      return null;
    }

    public Obj Make() { return Make(null); }
    public Obj Make(List args)
    {
      return Method("make", true).Call(args);
    }

  //////////////////////////////////////////////////////////////////////////
  // Inheritance
  //////////////////////////////////////////////////////////////////////////

    public virtual Type Base() { return baseType; }

    public virtual List Mixins() { return mixins; }

    /*
    public List Inheritance()
    {
      if (inheritance == null)
      {
        HashMap map = new HashMap();
        List acc = new List(Sys.TypeType);

        // always add Obj first
        if (this != Sys.ObjType)
        {
          map.put("sys::Obj", this);
          acc.add(Sys.ObjType);
        }

        // add myself
        map.put(qname.val, this);
        acc.add(this);

        // add my direct inheritance inheritance
        addInheritance(base(), acc, map);
        List mixins = mixins();
        for (int i=0; i<mixins.sz(); ++i)
          addInheritance((Type)mixins.get(i), acc, map);

        inheritance = acc.trim();
      }
      return inheritance;
    }
    */

    /*
    private void AddInheritance(Type t, List acc, HashMap map)
    {
      if (t == null) return;
      List ti = t.inheritance();
      for (int i=0; i<ti.sz(); ++i)
      {
        Type x = (Type)ti.get(i);
        if (map.get(x.qname.val) == null)
        {
          map.put(x.qname.val, x);
          acc.add(x);
        }
      }
    }
    */

    //public Bool Fits(Type type) { return is(type) ? Bool.True : Bool.False; }
    public virtual bool Is(Type type)
    {
      // TODO
      //if (this.inheritance == null) Inheritance();
      //List inheritance = this.inheritance;
      //for (int i=0; i<inheritance.sz(); ++i)
      //  if (inheritance.get(i) == type) return true;
      //return false;
      return false;
    }

    /// <summary>
    /// Given a list of types, compute the most specific type which they all share,
    /// or at worst return sys::Obj.  This method does not take into account
    /// interfaces, only extends class inheritance.
    /// </summary>
    /*
    public static Type common(Type[] types)
    {
      if (types.length == 0) return Sys.ObjType;
      Type best = types[0];
      for (int i=1; i<types.length; ++i)
      {
        Type t = types[i];
        while (!t.is(best))
        {
          best = best.base;
          if (best == null) return Sys.ObjType;
        }
      }
      return best;
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Facets
  //////////////////////////////////////////////////////////////////////////

    public Facets Facets()
    {
      return Reflect().facets;
    }

    /*
    public Str SourceFile()
    {
      return facets().get(Str.make(FConst.SourceFile));
    }

    public Int LineNumber()
    {
      return facets().getInt(Str.make(FConst.LineNumber));
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Documentation
  //////////////////////////////////////////////////////////////////////////

    /*
    public Str doc()
    {
      if (!docLoaded)
      {
        try
        {
          InputStream in = pod.fpod.store.read("doc/" + name + ".fandoc");
          if (in != null)
          {
            try { FDoc.read(in); } finally { in.close(); }
          }
        }
        catch(Exception e)
        {
          e.printStackTrace();
        }
        docLoaded = true;
      }
      return doc;
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Conversion
  //////////////////////////////////////////////////////////////////////////

    public override Str ToStr() { return Signature(); }

  //////////////////////////////////////////////////////////////////////////
  // Reflection
  //////////////////////////////////////////////////////////////////////////

    [MethodImpl(MethodImplOptions.Synchronized)]
    public Type Reflect() // TODO - need to call this from FTypeEmit for now
    {
      // short circuit if already reflected
      if (slotsByName != null) return this;

      // if this is a sys stub, use System.Reflection
      //if (isStub) { StubReflect(); return this; }

      if (Debug) Console.WriteLine("-- reflect: " + qname + " " + slotsByName);

      // do it
      DoReflect();

      // return this
      return this;
    }

    protected virtual void DoReflect()
    {
      // if the ftype is non-null, that means it was passed in non-hollow
      // ftype (in-memory compile), otherwise we need to read it from the pod
      FPod fpod   = this.pod.fpod;
      FType ftype = this.ftype;
      if (ftype.hollow)
      {
        try
        {
          ftype.Read();
        }
        catch (IOException e)
        {
          Console.WriteLine(e);
          throw IOErr.Make("Cannot read " + qname + " from pod", e).val;
        }
      }

      // these are working accumulators used to build the
      // data structures of my defined and inherited slots
      List slots  = new List(Sys.SlotType, 64);
      Hashtable nameToSlot  = new Hashtable();   // String -> Slot
      Hashtable nameToIndex = new Hashtable();   // Strnig -> Int

      // merge in base class and mixin classes
      //merge(base, slots, nameToSlot, nameToIndex);
      //for (int i=0; i<mixins().sz(); ++i) merge((Type)mixins.get(i), slots, nameToSlot, nameToIndex);

      // merge in all my slots
      for (int i=0; i<ftype.fields.Length; i++)
      {
        Field f = Map(fpod, ftype.fields[i]);
        Merge(f, slots, nameToSlot, nameToIndex);
      }
      for (int i=0; i<ftype.methods.Length; i++)
      {
        Method m = Map(fpod, ftype.methods[i]);
        Merge(m, slots, nameToSlot, nameToIndex);
      }

      // break out into fields and methods
      List fields  = new List(Sys.FieldType,  slots.Sz());
      List methods = new List(Sys.MethodType, slots.Sz());
      for (int i=0; i<slots.Sz(); ++i)
      {
        Slot slot = (Slot)slots.Get(i);
        if (slot is Field)
          fields.Add(slot);
        else
          methods.Add(slot);
      }
      this.slots       = slots.Trim().RO();
      this.fields      = fields.Trim().RO();
      this.methods     = methods.Trim().RO();
      this.slotsByName = nameToSlot;
      this.facets      = Fan.Sys.Facets.Make(fpod, qname, null, ftype.facets);
    }

    /// <summary>
    /// Merge the inherit's slots into my slot maps.
    ///   slots:       Slot[] by order
    ///   nameToSlot:  String name -> Slot
    ///   nameToIndex: String name -> Int index of slots
    /// </summary>
    /*
    private void Merge(Type inheritedType, List slots, Hashtable nameToSlot, Hashtable nameToIndex)
    {
      if (inheritedType == null) return;
      List inheritedSlots = inheritedType.Reflect().slots;
      for (int i=0; i<inheritedSlots.Sz(); ++i)
        Merge((Slot)inheritedSlots.Get(i), slots, nameToSlot, nameToIndex);
    }
    */

    /// <summary>
    /// Merge the inherited slot into my slot maps.  Assume this slot
    /// trumps any previous definition (because we process inheritance
    /// and my slots in the right order)
    ///   slots:       Slot[] by order
    ///   nameToSlot:  String name -> Slot
    ///   nameToIndex: String name -> Int index of slots
    /// </summary>
    private void Merge(Slot slot, List slots, Hashtable nameToSlot, Hashtable nameToIndex)
    {
      // skip constructors which aren't mine
      if (slot.IsCtor().val && slot.parent != this) return;

      String name = slot.name.val;
      Int dup = (Int)nameToIndex[name];
      if (dup != null)
      {
        // check if this is a Getter or Setter, in which case the Field
        // trumps and we need to cache the method on the Field
        // Note: this works because we assume the compiler always generates
        // the field before the getter and setter in fcode
        if ((slot.flags & (FConst.Getter|FConst.Setter)) != 0)
        {
          Field field = (Field)slots.Get(dup);
          if ((slot.flags & FConst.Getter) != 0)
            field.getter = (Method)slot;
          else
            field.setter = (Method)slot;
          return;
        }

        nameToSlot[name] = slot;
        slots.Set(dup, slot);
      }
      else
      {
        nameToSlot[name] = slot;
        slots.Add(slot);
        nameToIndex[name] = Int.Make(slots.Sz()-1);
      }
    }

    /// <summary>
    /// Map fcode field to a sys::Field.
    /// </summary>
    private Field Map(FPod fpod, FField f)
    {
      Str name = Str.Make(fpod.Name(f.name)).Intern();
      Type fieldType = pod.FindType(f.type);
      Facets facets = null; //Facets.Make(fpod, qname, name, f.facets);
      return new Field(this, name, f.flags, facets, fieldType);
    }

    /// <summary>
    /// Map fcode method to a sys::Method.
    /// </summary>
    private Method Map(FPod fpod, FMethod m)
    {
      Str name = Str.Make(fpod.Name(m.name)).Intern();
      Type returnType = pod.FindType(m.ret);
      Type inheritedReturnType = pod.FindType(m.inheritedRet);
      List pars = new List(Sys.MethodParamType, m.paramCount);
      for (int j=0; j<m.paramCount; j++)
      {
        FMethodVar p = m.vars[j];
        int pflags = (p.def == null) ? 0 : MethodParam.HAS_DEFAULT;
        pars.Add(new MethodParam(Str.Make(fpod.Name(p.name)).Intern(), pod.FindType(p.type), pflags));
      }
      Facets facets = Fan.Sys.Facets.Make(fpod, qname, name, m.facets);
      return new Method(this, name, m.flags, facets, returnType, inheritedReturnType, pars);
    }

  //////////////////////////////////////////////////////////////////////////
  // Emit
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Emit to a .NET Type.
    /// </summary>
    [MethodImpl(MethodImplOptions.Synchronized)]
    public System.Type Emit()
    {
      if (type == null)
      {
        if (Debug) Console.WriteLine("-- emit:   " + qname);

        // make sure we have reflected to setup slots
        Reflect();

        // if sys class, just load it by name
        string podName = pod.name.val;
        if (podName == "sys" || Sys.usePrecompiledOnly)
        {
          try
          {
            type = System.Type.GetType("Fan." + NameUtil.Upper(podName) + "." + name.val);
          }
          catch (Exception e)
          {
            Console.WriteLine(e);
            throw Err.Make("Cannot load precompiled class: " + qname, e).val;
          }
        }

        // otherwise we need to emit it
        else
        {
          try
          {
            this.type = FTypeEmit.EmitAndLoad(ftype);
          }
          catch (Exception e)
          {
            Console.WriteLine(e);
            throw Err.Make("Cannot emit: " + qname, e).val;
          }
        }

        // we are done with our ftype now, gc it
        this.ftype = null;
      }
      return type;
    }

  //////////////////////////////////////////////////////////////////////////
  // Finish
  //////////////////////////////////////////////////////////////////////////

    /*
    /// <summary>
    /// Finish ensures we have reflected and emitted, then does
    /// the final binding between slots and .NET members.
    /// </summary>
    [MethodImpl(MethodImplOptions.Synchronized)]
    public void Finish()
    {
      if (finished) return;
      try
      {
        // ensure reflected and emitted
        Reflect();
        Emit();
        finished = true;

        // map Java members to my slots for reflection; if
        // mixin then we do this for both the interface and
        // the static methods only of the implementation class
        FinishSlots(type, false);
        // TODO
        //if (isMixin().val) FinishSlots(auxCls, true);

      }
      catch (Exception e)
      {
        System.Console.WriteLine(e.StackTrace);
        throw Err.Make("Cannot emitFinish: " + qname, e).val;
      }
    }

    /// <summary>
    /// Map the Java members of the specified
    /// class to my slots for reflection.
    /// </summary>
    private void FinishSlots(System.Type type, bool staticOnly)
    {
      // map the class's fields to my slots
//      FieldInfo[] fields = cls.getDeclaredFields();
//      for (int i=0; i<fields.length; i++)
//        FinishField(fields[i]);

      // map the class's methods to my slots
      MethodInfo[] methods = type.GetMethods();
      for (int i=0; i<methods.Length; i++)
        FinishMethod(methods[i], staticOnly);
    }

    private void FinishField(FieldInfo f)
    {
//      Slot slot = slot(f.getName(), false);
//      if (slot == null || !(slot instanceof Field)) return;
//      Field field = (Field)slot;
//      if (field.reflect != null) return; // first one seems to give us most specific binding
//      f.setAccessible(true);
//      field.reflect = f;
    }

    private void FinishMethod(MethodInfo m, bool staticOnly)
    {
      Slot slot = Slot(m.Name, false);
      if (slot == null) return;
      if (staticOnly && !slot.IsStatic().val) return;
      //m.SetAccessible(true);
      if (slot is Method)
      {
        Method method = (Method)slot;

        // alloc System.Reflection.MethodInfo[] array big enough
        // to handle all the versions with default parameters
        if (method.reflect == null)
        {
          int n = 1;
          for (int j=method.pars.Sz()-1; j>=0; j--)
          {
            if (((MethodParam)method.pars.Get(j)).HasDefault().val) n++;
            else break;
          }
          method.reflect = new MethodInfo[n];
        }

        // get parameters, if sys we need to skip the
        // methods that use non-Fan signatures
        ParameterInfo[] pars = m.GetParameters();
        if (pod == Sys.SysPod && !CheckAllFan(pars)) return;

        // zero index is full signature up to using max defaults
        method.reflect[method.pars.Sz()-pars.Length] = m;
      }
      else
      {
        Field field = (Field)slot;
        if (m.ReturnType.ToString() == "System.Void")
          field.setter.reflect = new MethodInfo[] { m };
        else
          field.getter.reflect = new MethodInfo[] { m };
      }
    }

    bool CheckAllFan(ParameterInfo[] pars)
    {
      for (int i=0; i<pars.Length; i++)
        if (!pars[i].ParameterType.ToString().StartsWith("Fan."))
          return false;
      return true;
    }
    */
  //////////////////////////////////////////////////////////////////////////
  // Compiler Support
  //////////////////////////////////////////////////////////////////////////

    public bool IsObj()   { return this == Sys.ObjType;  }
    public bool IsBool()  { return this == Sys.BoolType; }
    public bool IsInt()   { return this == Sys.IntType;  }
    public bool IsFloat() { return this == Sys.FloatType; }
    public bool IsStr()   { return this == Sys.StrType;  }
    public bool IsVoid()  { return this == Sys.VoidType; }

    public Field[] FieldArr()   { return (Field[])Fields().ToArray(new Field[fields.Sz()]); }
    public Method[] MethodArr() { return (Method[])Methods().ToArray(new Method[methods.Sz()]); }
    public Slot[] SlotArr()     { return (Slot[])Slots().ToArray(new Slot[slots.Sz()]); }

    public void Dump()
    {
      System.Console.WriteLine("Type " + Qname() + " extends " + Base());
      for (int i=0; i<Slots().Sz(); i++)
        System.Console.WriteLine("  " + ((Slot)Slots().Get(i)).Signature());
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    internal static readonly bool Debug = false;
    //internal static Object noParams;

    // available when hollow
    internal readonly Pod pod;
    internal readonly Str name;
    internal readonly Str qname;
    internal readonly int flags;
    internal Facets facets;
    internal Type baseType = null;
    internal List mixins;
    //internal List inheritance;
    internal FType ftype;   // we only keep this around for memory compiles
    //internal bool isStub;   // sys stub type used to bootstrap compiling of sys itself
    //internal bool docLoaded;
    //public Str doc;

    // available when reflected
    internal List fields;
    internal List methods;
    internal List slots;
    internal Hashtable slotsByName;  // String:Slot

    // available when emitted
    internal System.Type type = null;

    // flags to ensure we finish only once
    //bool finished;

    // misc
    Type listOf;
  }
}