//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   14 Sep 06  Andy Frank  Creation
//

using System;
using System.Collections;
using System.IO;
using PERWAPI;
using Fanx.Util;

namespace Fanx.Emit
{
  /// <summary>
  /// Emitter.
  /// </summary>
  public class Emitter
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Convenience for init().
    /// </summary>
    public Emitter(string assemblyName)
    {
      Init(assemblyName);
    }

    /// <summary>
    /// Uninitialized Emitter - must call init().
    /// </summary>
    public Emitter() {}

    /// <summary>
    /// Initialize an Emitter to generate types for an assembly
    /// with the given name.
    /// </summary>
    public void Init(string assemblyName)
    {
      this.assemblyName = assemblyName;
      buf = new MemoryStream(1000);
      peFile = new PEFile(assemblyName + ".dll", assemblyName);
      peFile.SetOutputStream(buf);
    }

    /// <summary>
    /// Commit the current definition to the PE file.
    /// </summary>
    public byte[] Commit()
    {
      peFile.WritePEFile(true); // TODO - debug flag
      //return buf.ToArray();
      byte[] b = buf.ToArray();

      if (true) // dump to file
      {
        string path = @"c:\dev\fan\tmp\" + assemblyName + ".dll";
        BinaryWriter w = new BinaryWriter(new FileInfo(path).OpenWrite());
        w.Write(b, 0, b.Length);
        w.Flush();
        w.Close();
      }

      return b;
    }

  //////////////////////////////////////////////////////////////////////////
  // Methods
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Define a new class to emit for this assembly. This class
    /// becomes the 'current' class, where subsequent EmitMethod
    /// and EmitField calls will popuate this class.
    /// </summary>
    public void EmitClass(string baseQName, string qname, TypeAttr attr)
    {
      string[] s = NameUtil.SplitQName(qname);
      className = qname;
      classDef = peFile.AddClass(attr, s[0], s[1]);
      classDef.SuperType = FindType(baseQName) as PERWAPI.Class;
      types[qname] = classDef;
    }

    /// <summary>
    /// Define a new field to emit for the current class.
    /// </summary>
    public void EmitField(string name, string type, FieldAttr attr)
    {
      Field field = classDef.AddField(attr, name, FindType(type));
      fields[className + "/F" + name] = field;
    }

    /// <summary>
    /// Define a new method to emit for the current class.
    /// </summary>
    public CILInstructions EmitMethod(
      string name, string retType, string[] paramNames, string[] paramTypes,
      MethAttr attr, string[] localNames, string[] localTypes)
    {
      Param[] pars = new Param[paramNames.Length];
      for (int i=0; i<pars.Length; i++)
        pars[i] = new Param(ParamAttr.Default, paramNames[i], FindType(paramTypes[i]));

      methodDef = classDef.AddMethod(attr, ImplAttr.IL, name, FindType(retType), pars);

      if (attr != MethAttr.Static &&
          attr != MethAttr.PublicStatic &&
          attr != MethAttr.PublicStaticFinal)
        methodDef.AddCallConv(CallConv.Instance);

      if (localNames.Length > 0)
      {
        Local[] locals = new Local[localNames.Length];
        for (int i=0; i<locals.Length; i++)
          locals[i] = new Local(localNames[i], FindType(localTypes[i]));
        methodDef.AddLocals(locals, true);
      }

      // TODO - what the fuck should this be?
      methodDef.SetMaxStack(8 + pars.Length + localNames.Length);

      AddToMethodMap(className, name, paramTypes, methodDef);
      return methodDef.CreateCodeBuffer();
    }

    /// <summary>
    /// Add a MethodDef into the lookup table.
    /// </summary>
    public void AddToMethodMap(string className, string name, string[] paramTypes, MethodDef def)
    {
      string key = className + "/M" + name;
      for (int i=0; i<paramTypes.Length; i++)
        key += "/" + paramTypes[i];
      methods[key] = def;
    }

  //////////////////////////////////////////////////////////////////////////
  // Util
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Find the Type instance for this fully qualified type name.
    /// </summary>
    internal PERWAPI.Type FindType(string qname)
    {
      // Always convert voids to native
      if (qname == "Fan.Sys.Void") qname = "System.Void";

      PERWAPI.Type type = (PERWAPI.Type)types[qname];
      if (type == null)
      {
        string aname = NameUtil.GetPodName(qname);
        if (aname == null) aname = "mscorlib";
        if (qname.StartsWith("Fanx.")) aname = "sys";  // hack for support classes

        AssemblyRef aref = (AssemblyRef)assemblies[aname];
        if (aref == null)
        {
          aref = peFile.MakeExternAssembly(aname);
          assemblies[aname] = aref;
        }

        string[] s = NameUtil.SplitQName(qname);
        if (qname.IndexOf("/") != -1)
        {
          // Nested class
          PERWAPI.ClassRef cref = (PERWAPI.ClassRef)FindType(s[0]);
          type = cref.AddNestedClass(s[1]);
        }
        else
        {
          // Normal class
          // TODO - this broke when I moved to PERWAPI - not sure why
          // so now I have to check that it doesn't exist b/f adding
          type = aref.GetClass(s[0], s[1]);
          if (type == null) type = aref.AddClass(s[0], s[1]);
        }
        types[qname] = type;
      }
      return type;
    }

    /// <summary>
    /// Find the Field for this field on the given type.
    /// </summary>
    internal Field FindField(string qname, string fieldName, string fieldType)
    {
      string key = qname + "/F" + fieldName;
      Field field = (Field)fields[key];
      if (field == null)
      {
        ClassRef cref = FindType(qname) as ClassRef;
        PERWAPI.Type ftype = FindType(fieldType);
        field = cref.AddField(fieldName, ftype);
        fields[key] = field;
      }
      return field;
    }

    /// <summary>
    /// Find the Method for this method on the given type.
    /// </summary>
    internal Method FindMethod(string qname, string methodName, string[] paramTypes, string returnType)
    {
      string key = qname + "/M" + methodName;
       for (int i=0; i<paramTypes.Length; i++)
        key += "/" + paramTypes[i];
      Method method = (Method)methods[key];

      if (method == null)
      {
        ClassRef cref = FindType(qname) as ClassRef;
        PERWAPI.Type rtype = FindType(returnType);
        PERWAPI.Type[] pars = new PERWAPI.Type[paramTypes.Length];
        for (int i=0; i<paramTypes.Length; i++)
          pars[i] = FindType(paramTypes[i]);
        method = cref.AddMethod(methodName, rtype, pars);
        methods[key] = method;
      }
      return method;
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    private string assemblyName;  // name of this assbemly
    private MemoryStream buf;     // memory buffer for pe file
    private PEFile peFile;        // model for pe file

    private string className;     // the current class name being emitted
    internal ClassDef classDef;   // the current class being emitted
    internal MethodDef methodDef; // the current or last method emitted

    private Hashtable assemblies = new Hashtable();  // Assembly lookup table
    private Hashtable types = new Hashtable();       // Type lookup table
    private Hashtable fields = new Hashtable();      // Field lookup table
    private Hashtable methods = new Hashtable();     // Method lookup table
  }
}