//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   10 Oct 06  Andy Frank  Creation
//

using System;
using System.Collections;
using PERWAPI;
using Fanx.Fcode;
using Fanx.Util;

namespace Fanx.Emit
{
  /// <summary>
  /// FCodeEmit translates FCode fcode to IL bytecode.
  /// </summary>
  public class FCodeEmit
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    public FCodeEmit(FTypeEmit parent, FMethod fmethod, CILInstructions code)
     : this(parent, fmethod.code, code)
    {
      //this(parent, fmethod.code, code);
      this.fmethod    = fmethod;
      //code.maxLocals  = Math.max(fmethod.maxLocals(), 4); // TODO - total hack
    }

    public FCodeEmit(FTypeEmit parent, FBuf fcode, CILInstructions code)
    {
      this.pod        = parent.pod;
      this.emitter    = parent.emitter;
      this.parent     = parent;
      this.buf        = fcode.buf;
      this.len        = fcode.len;
      //this.emit       = code.emit;
      this.code       = code;
      this.podClass   = NameUtil.Upper("Fan." + pod.podName + ".$Pod");
      this.jumps      = new Jumps(code);
    }

  //////////////////////////////////////////////////////////////////////////
  // Overrides
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Translate fcode to Java bytecode.
    /// </summary>
    public void Emit()
    {
      MakeErrTable();
      EmitInstructions();
      //LineTable();
    }

    /// <summary>
    /// Map fcode instructions to Java bytecode instructions.
    /// </summary>
    private void EmitInstructions()
    {
      while (pos < len)
      {
        startPos = pos;
        int opcode = ConsumeOp();
        switch (opcode)
        {
          case FConst.Nop:                break;
          case FConst.LoadNull:           code.Inst(Op.ldnull); break;
          case FConst.LoadFalse:          LoadFalse(); break;
          case FConst.LoadTrue:           LoadTrue(); break;
          case FConst.LoadInt:            LoadInt(); break;
          case FConst.LoadFloat:          LoadFloat(); break;
          case FConst.LoadStr:            LoadStr(); break;
          case FConst.LoadDuration:       LoadDuration(); break;
          case FConst.LoadUri:            LoadUri(); break;
          case FConst.LoadType:           LoadType(); break;

          case FConst.LoadVar:            LoadVar(); break;
          case FConst.StoreVar:           StoreVar(); break;
          case FConst.LoadInstance:       LoadInstance(); break;
          case FConst.StoreInstance:      StoreInstance(); break;
          case FConst.LoadStatic:         LoadStatic(); break;
          case FConst.StoreStatic:        StoreStatic(); break;
          //case FConst.LoadMixinInstance:  LoadMixinInstance(); break;
          //case FConst.StoreMixinInstance: StoreMixinInstance(); break;
          //case FConst.LoadMixinStatic:    LoadMixinStatic(); break;
          //case FConst.StoreMixinStatic:   StoreMixinStatic(); break;

          case FConst.CallNew:            CallNew(); break;
          case FConst.CallCtor:           CallCtor(); break;
          case FConst.CallStatic:         CallStatic(); break;
          case FConst.CallVirtual:        CallVirtual(); break;
          case FConst.CallNonVirtual:     CallNonVirtual(); break;
          //case FConst.CallMixinStatic:    CallMixinStatic(); break;
          //case FConst.CallMixinVirtual:   CallMixinVirtual(); break;

          case FConst.Jump:               Jump(); break;
          case FConst.JumpTrue:           JumpTrue(); break;
          case FConst.JumpFalse:          JumpFalse(); break;

          case FConst.CompareEQ:          CompareEQ(); break;
          case FConst.CompareNE:          CompareNE(); break;
          case FConst.Compare:            Compare(); break;
          case FConst.CompareLT:          CompareLT(); break;
          case FConst.CompareLE:          CompareLE(); break;
          case FConst.CompareGE:          CompareGE(); break;
          case FConst.CompareGT:          CompareGT(); break;
          case FConst.CompareSame:        CompareSame(); break;
          //case FConst.CompareNotSame:     CompareNotSame(); break;
          case FConst.CompareNull:        CompareNull(); break;
          case FConst.CompareNotNull:     CompareNotNull(); break;

          case FConst.ReturnVoid:         code.Inst(Op.ret); break;
          case FConst.ReturnObj:          code.Inst(Op.ret); break;
          case FConst.Pop:                code.Inst(Op.pop); break;
          case FConst.Dup:                code.Inst(Op.dup); break;
          //case FConst.DupDown:            code.op(DUP_X1); break; // WON'T SUPPORT
          case FConst.Is:                 Is(); break;
          case FConst.As:                 As(); break;
          case FConst.Cast:               Cast(); break;
          //case FConst.Switch:             tableswitch(); break;

          case FConst.Throw:              DoThrow(); break;
          case FConst.Leave:              DoLeave(); break;
          case FConst.JumpFinally:        U2(); break; // not used in .NET, so just eat the branch loc
          case FConst.CatchAllStart:      CatchAllStart(); break;
          case FConst.CatchErrStart:      CatchErrStart(); break;
          case FConst.CatchEnd:           CatchEnd(); break;
          case FConst.FinallyStart:       FinallyStart(); break;
          case FConst.FinallyEnd:         FinallyEnd(); break;

          default: throw new Exception(FConst.OpNames[opcode]);
        }
      }
    }

    /// <summary>
    /// Process error table (if specified).  We handle catches of Err using
    /// a catch any (0 class index).  We also need to add extra entries into
    /// the exception table for special exceptions - for example NullErr get's
    /// mapped as fan.sys.NullErr$Val and java.lang.NullPointerException.
    /// </summary>
    private void MakeErrTable()
    {
      if (fmethod == null) return;
      FFacet ferrs = fmethod.Facet(pod, FConst.ErrTable);
      if (ferrs == null) return;

      int len = ferrs.data.len;
      byte[] buf = ferrs.data.buf;
      int count = (len-2)/8;

      tryStart = new int[count];
      tryEnd   = new int[count];
      tryJump  = new int[count];
      tryErr   = new int[count];

      for (int i=2, j=0; i<len; i+=8)
      {
        tryStart[j] = (buf[i+0] & 0xFF) << 8 | (buf[i+1] & 0xFF);
        tryEnd[j]   = (buf[i+2] & 0xFF) << 8 | (buf[i+3] & 0xFF);
        tryJump[j]  = (buf[i+4] & 0xFF) << 8 | (buf[i+5] & 0xFF);
        tryErr[j]   = (buf[i+6] & 0xFF) << 8 | (buf[i+7] & 0xFF);
        j++;
      }
    }

    /// <summary>
    /// Pocess line number table (if specified), we just reuse the
    /// fcode line table buffer and replace the fcode pc with bytecode
    /// pc since they are the same sized data structures.
    /// </summary>
    /*
    private void LineTable()
    {
      if (fmethod == null) return;
      FFacet flines = fmethod.facet(pod, LineNumbers);
      if (flines == null) return;

      int len = flines.data.len;
      byte[] buf = flines.data.buf;
      for (int i=2; i<len; i += 4)
      {
        reloc(buf, i);
      }

      AttrEmit attr = code.emitAttr("LineNumberTable");
      attr.info.len = len;
      attr.info.buf = buf;
    }
    */

    /*
    private void Reloc(byte[] buf, int offset)
    {
      int fpos = (buf[offset] & 0xFF) << 8 | (buf[offset+1] & 0xFF);
      int jpos = reloc[fpos];
      buf[offset+0] = (byte)(jpos >>> 8);
      buf[offset+1] = (byte)(jpos >>> 0);
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Load/Store
  //////////////////////////////////////////////////////////////////////////

    private void LoadFalse()
    {
      if (parent.BoolFalse == null)
        parent.BoolFalse = emitter.FindField("Fan.Sys.Bool", "False", "Fan.Sys.Bool");
      code.FieldInst(FieldOp.ldsfld, parent.BoolFalse);
    }

    private void LoadTrue()
    {
      // TODO: optimize while (true) either here or in compiler
      if (parent.BoolTrue == null)
        parent.BoolTrue = emitter.FindField("Fan.Sys.Bool", "True", "Fan.Sys.Bool");
      code.FieldInst(FieldOp.ldsfld, parent.BoolTrue);
    }

    private void LoadInt()
    {
      int index = U2();
      PERWAPI.Field field = emitter.FindField(podClass, "I" + index, "Fan.Sys.Int");
      code.FieldInst(FieldOp.ldsfld, field);
    }

    private void LoadFloat()
    {
      int index = U2();
      PERWAPI.Field field = emitter.FindField(podClass, "F" + index, "Fan.Sys.Float");
      code.FieldInst(FieldOp.ldsfld, field);
    }

    private void LoadStr()
    {
      int index = U2();
      PERWAPI.Field field = emitter.FindField(podClass, "S" + index, "Fan.Sys.Str");
      code.FieldInst(FieldOp.ldsfld, field);
    }

    private void LoadDuration()
    {
      int index = U2();
      PERWAPI.Field field = emitter.FindField(podClass, "D" + index, "Fan.Sys.Duration");
      code.FieldInst(FieldOp.ldsfld, field);
    }

    private void LoadUri()
    {
      int index = U2();
      PERWAPI.Field field = emitter.FindField(podClass, "U" + index, "Fan.Sys.Uri");
      code.FieldInst(FieldOp.ldsfld, field);
    }

    private void LoadType()
    {
      LoadType(U2());
    }

    private void LoadType(int typeRefIndex)
    {
      FTypeRef tref = pod.TypeRef(typeRefIndex);
      string podName  = pod.Name(tref.podName);
      string typeName = pod.Name(tref.typeName);

      // if pod is "sys", then we can perform a shortcut and use
      // one of the predefined fields in Sys
      if (!tref.IsGenericInstance() && podName == "sys")
      {
        PERWAPI.Field field = emitter.FindField("Fan.Sys.Sys", typeName + "Type", "Fan.Sys.Type");
        code.FieldInst(FieldOp.ldsfld, field);
        return;
      }
/*
      // lazy allocate my parent's type literal map: sig -> fieldName
      if (parent.typeLiteralFields == null) parent.typeLiteralFields= new HashMap();
      HashMap map = parent.typeLiteralFields;

      // types are lazy loaded and then cached in a private static field called
      // type$count which will get generated by FTypeEmit (we keep track of signature
      // to fieldname in the typeConstFields map)
      String sig = ref.sig(pod);
      String fieldName = (String)map.get(sig);
      if (fieldName == null)
      {
        fieldName = "type$" + map.size();
        map.put(sig, fieldName);
      }
      int fieldRef = emit.field(parent.className + "." + fieldName + ":Lfan/sys/Type;");

      code.op2(GETSTATIC, fieldRef);
      code.op(DUP);
      int nonNull = code.branch(IFNONNULL);
      code.op(POP);
      code.op2(LDC_W, emit.strConst(sig));
      code.op(ICONST_1);
      code.op2(INVOKESTATIC, parent.sysFindType());
      code.op(DUP);
      code.op2(PUTSTATIC, fieldRef);
      code.mark(nonNull);
*/
    }

    private void LoadVar()
    {
      int index = U2();

      int paramCount = fmethod.paramCount;
      if ((fmethod.flags & FConst.Static) == 0) paramCount++;

      if (index < paramCount)
      {
        switch (index)
        {
          case 0:  code.Inst(Op.ldarg_0); break;
          case 1:  code.Inst(Op.ldarg_1); break;
          case 2:  code.Inst(Op.ldarg_2); break;
          case 3:  code.Inst(Op.ldarg_3); break;
          default: code.IntInst(IntOp.ldarg, index); break;
        }
      }
      else
      {
        index -= paramCount;
        switch (index)
        {
          case 0:  code.Inst(Op.ldloc_0); break;
          case 1:  code.Inst(Op.ldloc_1); break;
          case 2:  code.Inst(Op.ldloc_2); break;
          case 3:  code.Inst(Op.ldloc_3); break;
          default: code.IntInst(IntOp.ldloc, index); break;
        }
      }
    }

    private void StoreVar()
    {
      int index = U2();

      int paramCount = fmethod.paramCount;
      if ((fmethod.flags & FConst.Static) == 0) paramCount++;

      if (index < paramCount)
      {
        code.IntInst(IntOp.starg, index);
      }
      else
      {
        index -= paramCount;
        switch (index)
        {
          case 0:  code.Inst(Op.stloc_0); break;
          case 1:  code.Inst(Op.stloc_1); break;
          case 2:  code.Inst(Op.stloc_2); break;
          case 3:  code.Inst(Op.stloc_3); break;
          default: code.IntInst(IntOp.stloc, index); break;
        }
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Field
  //////////////////////////////////////////////////////////////////////////

    private void LoadInstance()
    {
      FPod.NField f = pod.Nfield(U2(), false);
      PERWAPI.Field field = emitter.FindField(f.parentType, f.fieldName, f.fieldType);
      code.FieldInst(FieldOp.ldfld, field);
    }

    private void StoreInstance()
    {
      FPod.NField f = pod.Nfield(U2(), false);
      PERWAPI.Field field = emitter.FindField(f.parentType, f.fieldName, f.fieldType);
      code.FieldInst(FieldOp.stfld, field);
    }

    private void LoadStatic()
    {
      FPod.NField f = pod.Nfield(U2(), false);
      PERWAPI.Field field = emitter.FindField(f.parentType, f.fieldName, f.fieldType);
      code.FieldInst(FieldOp.ldsfld, field);
    }

    private void StoreStatic()
    {
      FPod.NField f = pod.Nfield(U2(), false);
      PERWAPI.Field field = emitter.FindField(f.parentType, f.fieldName, f.fieldType);
      code.FieldInst(FieldOp.stsfld, field);
    }

    /*
    private void LoadMixinInstance()
    {
      int field = emit.field(pod.jfield(u2(), true));
      code.op2(GETFIELD, field);
    }

    private void StoreMixinInstance()
    {
      int field = emit.field(pod.jfield(u2(), true));
      code.op2(PUTFIELD, field);
    }

    private void LoadMixinStatic()
    {
      int field = emit.field(pod.jfield(u2(), true));
      code.op2(GETSTATIC, field);
    }

    private void StoreMixinStatic()
    {
      int field = emit.field(pod.jfield(u2(), true));
      code.op2(PUTSTATIC, field);
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Calls
  //////////////////////////////////////////////////////////////////////////

    private void CallNew()
    {
      // constructors are implemented as static factory methods
      FPod.NMethod ncall = pod.Ncall(U2(), FConst.CallNew);
      Method method = emitter.FindMethod(ncall.parentType, ncall.methodName,
        ncall.paramTypes, ncall.returnType);
      code.MethInst(MethodOp.call, method);
    }

    private void CallCtor()
    {
      // constructor implementations (without object allow) are
      // implemented as static factory methods with "$" appended

      int index = U2();
      int[] m = pod.MethodRef(index).val;
      string parent = pod.Nname(m[0]);
      string name = NameUtil.Upper(pod.Name(m[1])) + "$";

      string[] pars = new string[m.Length-3+1];
      pars[0] = parent;
      for (int i=0; i<pars.Length-1; i++)
        pars[i+1] = pod.Nname(m[i+3]);

      Method method = emitter.FindMethod(parent, name, pars, parent);
      code.MethInst(MethodOp.call, method);
    }

    private void CallStatic()
    {
      FPod.NMethod ncall = pod.Ncall(U2(), FConst.CallStatic);
      Method method = emitter.FindMethod(ncall.parentType, ncall.methodName,
        ncall.paramTypes, ncall.returnType);
      code.MethInst(MethodOp.call, method);
    }

    private void CallVirtual()
    {
      int index = U2();
      FPod.NMethod ncall = pod.Ncall(index, FConst.CallVirtual);

      //System.Console.WriteLine("### sig: " + sig);

      // if this is a virtual invoke on Obj then we to make it an
      // interface invoke because the Java runtime models Obj
      // as an interface to deal with mixins cleanly
//      if (sig.startsWith("fan/sys/Obj."))
//      {
//        int[] m = pod.methodRef(index).val;
//        int nargs = m.length-3;
//        int method = emit.interfaceRef(sig);
//        code.op2(INVOKEINTERFACE, method);
//        code.info.u1(nargs+1);
//        code.info.u1(0);
//      }
//      else
//      {
        Method method = emitter.FindMethod(ncall.parentType, ncall.methodName,
          ncall.paramTypes, ncall.returnType);
        if (!ncall.isStatic) method.AddCallConv(CallConv.Instance);
        code.MethInst(MethodOp.callvirt, method);
//      }
    }

    private void CallNonVirtual()
    {
      // invokespecial in Java is really queer - it can only
      // be used for calls in the declaring class (basically
      // for private methods or super call)
//      String sig = pod.jcall(u2(), CallNonVirtual);
//      int method = emit.method(sig);
//      code.op2(INVOKESPECIAL, method);

      FPod.NMethod ncall = pod.Ncall(U2(), FConst.CallNonVirtual);
      Method method = emitter.FindMethod(ncall.parentType, ncall.methodName,
        ncall.paramTypes, ncall.returnType);
      code.MethInst(MethodOp.call, method);
    }

    private void CallMixinStatic()
    {
//      String sig = pod.jcall(u2(), CallMixinStatic);
//      int method = emit.method(sig);
//      code.op2(INVOKESTATIC, method);
    }

    private void CallMixinVirtual()
    {
//      int index = u2();
//      int[] m = pod.methodRef(index).val;
//      int nargs = m.length-3;
//
//      String sig = pod.jcall(index, CallMixinVirtual);
//      int method = emit.interfaceRef(sig);
//      code.op2(INVOKEINTERFACE, method);
//      code.info.u1(nargs+1);
//      code.info.u1(0);
    }

  //////////////////////////////////////////////////////////////////////////
  // Jump
  //////////////////////////////////////////////////////////////////////////

    private void JumpTrue()
    {
      Field field = emitter.FindField("Fan.Sys.Bool", "val", "System.Boolean");
      code.FieldInst(FieldOp.ldfld, field);
      code.Branch(BranchOp.brtrue, jumps.Add(U2()));
    }

    private void JumpFalse()
    {
      Field field = emitter.FindField("Fan.Sys.Bool", "val", "System.Boolean");
      code.FieldInst(FieldOp.ldfld, field);
      code.Branch(BranchOp.brfalse, jumps.Add(U2()));
    }

    private void Jump()
    {
      code.Branch(BranchOp.br, jumps.Add(U2()));
    }

  //////////////////////////////////////////////////////////////////////////
  // Compare
  //////////////////////////////////////////////////////////////////////////

    private void CompareEQ()
    {
      int peek = PeekOp();
      switch (peek)
      {
        case FConst.JumpFalse:
          if (parent.CompareEQz == null)
            parent.CompareEQz = emitter.FindMethod("Fanx.Util.OpUtil", "CompareEQz",
              new string[] { "Fan.Sys.Obj", "Fan.Sys.Obj" }, "System.Boolean");
          code.MethInst(MethodOp.call, parent.CompareEQz);
          ConsumeOp();
          code.Branch(BranchOp.brfalse, jumps.Add(U2()));
          break;
        case FConst.JumpTrue:
          if (parent.CompareEQz == null)
            parent.CompareEQz = emitter.FindMethod("Fanx.Util.OpUtil", "CompareEQz",
              new string[] { "Fan.Sys.Obj", "Fan.Sys.Obj" }, "System.Boolean");
          code.MethInst(MethodOp.call, parent.CompareEQz);
          ConsumeOp();
          code.Branch(BranchOp.brtrue, jumps.Add(U2()));
          break;
        default:
          if (parent.CompareEQ == null)
            parent.CompareEQ = emitter.FindMethod("Fanx.Util.OpUtil", "CompareEQ",
              new string[] { "Fan.Sys.Obj", "Fan.Sys.Obj" }, "Fan.Sys.Bool");
          code.MethInst(MethodOp.call, parent.CompareEQ);
          break;
      }
    }

    private void CompareNE()
    {
      int peek = PeekOp();
      switch (peek)
      {
        case FConst.JumpFalse:
          if (parent.CompareNEz == null)
            parent.CompareNEz = emitter.FindMethod("Fanx.Util.OpUtil", "CompareNEz",
              new string[] { "Fan.Sys.Obj", "Fan.Sys.Obj" }, "System.Boolean");
          code.MethInst(MethodOp.call, parent.CompareNEz);
          ConsumeOp();
          code.Branch(BranchOp.brfalse, jumps.Add(U2()));
          break;
        case FConst.JumpTrue:
          if (parent.CompareNEz == null)
            parent.CompareNEz = emitter.FindMethod("Fanx.Util.OpUtil", "CompareNEz",
              new string[] { "Fan.Sys.Obj", "Fan.Sys.Obj" }, "System.Boolean");
          code.MethInst(MethodOp.call, parent.CompareNEz);
          ConsumeOp();
          code.Branch(BranchOp.brtrue, jumps.Add(U2()));
          break;
        default:
          if (parent.CompareNE == null)
            parent.CompareNE = emitter.FindMethod("Fanx.Util.OpUtil", "CompareNE",
              new string[] { "Fan.Sys.Obj", "Fan.Sys.Obj" }, "Fan.Sys.Bool");
          code.MethInst(MethodOp.call, parent.CompareNE);
          break;
      }
    }

    private void Compare()
    {
      if (parent.Compare == null)
        parent.Compare = emitter.FindMethod("Fanx.Util.OpUtil", "Compare",
          new string[] { "Fan.Sys.Obj", "Fan.Sys.Obj" }, "Fan.Sys.Int");
      code.MethInst(MethodOp.call, parent.Compare);
    }

    private void CompareLT()
    {
      int peek = PeekOp();
      switch (peek)
      {
       case FConst.JumpFalse:
          if (parent.CompareLTz == null)
            parent.CompareLTz = emitter.FindMethod("Fanx.Util.OpUtil", "CompareLTz",
              new string[] { "Fan.Sys.Obj", "Fan.Sys.Obj" }, "System.Boolean");
          code.MethInst(MethodOp.call, parent.CompareLTz);
          ConsumeOp();
          code.Branch(BranchOp.brfalse, jumps.Add(U2()));
          break;
        case FConst.JumpTrue:
          if (parent.CompareLTz == null)
            parent.CompareLTz = emitter.FindMethod("Fanx.Util.OpUtil", "CompareLTz",
              new string[] { "Fan.Sys.Obj", "Fan.Sys.Obj" }, "System.Boolean");
          code.MethInst(MethodOp.call, parent.CompareLTz);
          ConsumeOp();
          code.Branch(BranchOp.brtrue, jumps.Add(U2()));
          break;
       default:
          if (parent.CompareLT == null)
            parent.CompareLT = emitter.FindMethod("Fanx.Util.OpUtil", "CompareLT",
              new string[] { "Fan.Sys.Obj", "Fan.Sys.Obj" }, "Fan.Sys.Bool");
          code.MethInst(MethodOp.call, parent.CompareLT);
          break;
      }
    }

    private void CompareLE()
    {
      int peek = PeekOp();
      switch (peek)
      {
        case FConst.JumpFalse:
          if (parent.CompareLEz == null)
            parent.CompareLEz = emitter.FindMethod("Fanx.Util.OpUtil", "CompareLEz",
              new string[] { "Fan.Sys.Obj", "Fan.Sys.Obj" }, "System.Boolean");
          code.MethInst(MethodOp.call, parent.CompareLEz);
          ConsumeOp();
          code.Branch(BranchOp.brfalse, jumps.Add(U2()));
          break;
        case FConst.JumpTrue:
          if (parent.CompareLEz == null)
            parent.CompareLEz = emitter.FindMethod("Fanx.Util.OpUtil", "CompareLEz",
              new string[] { "Fan.Sys.Obj", "Fan.Sys.Obj" }, "System.Boolean");
          code.MethInst(MethodOp.call, parent.CompareLEz);
          ConsumeOp();
          code.Branch(BranchOp.brtrue, jumps.Add(U2()));
          break;
        default:
          if (parent.CompareLE == null)
            parent.CompareLE = emitter.FindMethod("Fanx.Util.OpUtil", "CompareLE",
              new string[] { "Fan.Sys.Obj", "Fan.Sys.Obj" }, "Fan.Sys.Bool");
          code.MethInst(MethodOp.call, parent.CompareLE);
          break;
      }
    }

    private void CompareGE()
    {
      int peek = PeekOp();
      switch (peek)
      {
        case FConst.JumpFalse:
          if (parent.CompareGEz == null)
            parent.CompareGEz = emitter.FindMethod("Fanx.Util.OpUtil", "CompareGEz",
              new string[] { "Fan.Sys.Obj", "Fan.Sys.Obj" }, "System.Boolean");
          code.MethInst(MethodOp.call, parent.CompareGEz);
          ConsumeOp();
          code.Branch(BranchOp.brfalse, jumps.Add(U2()));
          break;
        case FConst.JumpTrue:
          if (parent.CompareGEz == null)
            parent.CompareGEz = emitter.FindMethod("Fanx.Util.OpUtil", "CompareGEz",
              new string[] { "Fan.Sys.Obj", "Fan.Sys.Obj" }, "System.Boolean");
          code.MethInst(MethodOp.call, parent.CompareGEz);
          ConsumeOp();
          code.Branch(BranchOp.brtrue, jumps.Add(U2()));
          break;
        default:
          if (parent.CompareGE == null)
            parent.CompareGE = emitter.FindMethod("Fanx.Util.OpUtil", "CompareGE",
              new string[] { "Fan.Sys.Obj", "Fan.Sys.Obj" }, "Fan.Sys.Bool");
          code.MethInst(MethodOp.call, parent.CompareGE);
          break;
      }
    }

    private void CompareGT()
    {
      int peek = PeekOp();
      switch (peek)
      {
        case FConst.JumpFalse:
          if (parent.CompareGTz == null)
            parent.CompareGTz = emitter.FindMethod("Fanx.Util.OpUtil", "CompareGTz",
              new string[] { "Fan.Sys.Obj", "Fan.Sys.Obj" }, "System.Boolean");
          code.MethInst(MethodOp.call, parent.CompareGTz);
          ConsumeOp();
          code.Branch(BranchOp.brfalse, jumps.Add(U2()));
          break;
        case FConst.JumpTrue:
          if (parent.CompareGTz == null)
            parent.CompareGTz = emitter.FindMethod("Fanx.Util.OpUtil", "CompareGTz",
              new string[] { "Fan.Sys.Obj", "Fan.Sys.Obj" }, "System.Boolean");
          code.MethInst(MethodOp.call, parent.CompareGTz);
          ConsumeOp();
          code.Branch(BranchOp.brtrue, jumps.Add(U2()));
          break;
        default:
          if (parent.CompareGT == null)
            parent.CompareGT = emitter.FindMethod("Fanx.Util.OpUtil", "CompareGT",
              new string[] { "Fan.Sys.Obj", "Fan.Sys.Obj" }, "Fan.Sys.Bool");
          code.MethInst(MethodOp.call, parent.CompareGT);
          break;
      }
    }

    private void CompareSame()
    {
      int peek = PeekOp();
      switch (peek)
      {
        case FConst.JumpFalse:
          ConsumeOp();
          code.Branch(BranchOp.bne_un, jumps.Add(U2()));
          break;
        case FConst.JumpTrue:
          ConsumeOp();
          code.Branch(BranchOp.bne_un, jumps.Add(U2()));
          break;
        default:
          if (parent.CompareSame == null)
            parent.CompareSame = emitter.FindMethod("Fanx.Util.OpUtil", "CompareSame",
              new string[] { "System.Object", "System.Object" }, "Fan.Sys.Bool");
          code.MethInst(MethodOp.call, parent.CompareSame);
          break;
      }
    }

    /*
    private void CompareNotSame()
    {
      int peek = peekOp();
      switch (peek)
      {
        case JumpFalse:
          consumeOp();
          code.op(IF_ACMPEQ);
          branch();
          break;
        case JumpTrue:
          consumeOp();
          code.op(IF_ACMPNE);
          branch();
          break;
        default:
          if (parent.CompareNotSame == 0) parent.CompareNotSame = emit.method("fanx/util/OpUtil.compareNotSame(Ljava/lang/Object;Ljava/lang/Object;)Lfan/sys/Bool;");
          code.op2(INVOKESTATIC, parent.CompareNotSame);
      }
    }
    */

    private void CompareNull()
    {
      int peek = PeekOp();
      switch (peek)
      {
        case FConst.JumpFalse:
          ConsumeOp();
          code.Branch(BranchOp.brtrue, jumps.Add(U2()));
          break;
        case FConst.JumpTrue:
          ConsumeOp();
          code.Branch(BranchOp.brfalse, jumps.Add(U2()));
          break;
        default:
          if (parent.CompareNull == null)
            parent.CompareNull = emitter.FindMethod("Fanx.Util.OpUtil", "CompareNull",
              new string[] { "System.Object" }, "Fan.Sys.Bool");
          code.MethInst(MethodOp.call, parent.CompareNull);
          break;
      }
    }

    private void CompareNotNull()
    {
      int peek = PeekOp();
      switch (peek)
      {
        case FConst.JumpFalse:
          ConsumeOp();
          code.Branch(BranchOp.brfalse, jumps.Add(U2()));
          break;
        case FConst.JumpTrue:
          ConsumeOp();
          code.Branch(BranchOp.brtrue, jumps.Add(U2()));
          break;
        default:
          if (parent.CompareNotNull == null)
            parent.CompareNotNull = emitter.FindMethod("Fanx.Util.OpUtil", "CompareNotNull",
              new string[] { "System.Object" }, "Fan.Sys.Bool");
          code.MethInst(MethodOp.call, parent.CompareNotNull);
         break;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Is/As
  //////////////////////////////////////////////////////////////////////////

    private void Is()
    {
      int typeRef = U2();
      PERWAPI.Type type = emitter.FindType(pod.Nname(typeRef));

      // if a generic instance, we have to use a method call
      // because Fan types don't map to Java classes exactly;
      // otherwise we can use straight bytecode
      //if (pod.typeRef(typeRef).isGenericInstance())
      //{
      //  if (parent.IsViaType == 0) parent.IsViaType = emit.method("fanx/util/OpUtil.is(Lfan/sys/Obj;Lfan/sys/Type;)Lfan/sys/Bool;");
      //  loadType(typeRef);
      //  code.op2(INVOKESTATIC, parent.IsViaType);
      //}
      //else
      //{
        code.TypeInst(TypeOp.isinst, type);
        code.Inst(Op.ldnull);
        code.Inst(Op.cgt_un);
        BoolMake();
      //}
    }

    private void As()
    {
      int typeRef = U2();
      PERWAPI.Type type = emitter.FindType(pod.Nname(typeRef));

      // if a generic instance, we have to use a method call
      // because Fan types don't map to Java classes exactly;
      // otherwise we can use straight bytecode
      //if (pod.typeRef(typeRef).isGenericInstance())
      //{
      //  if (parent.AsViaType == 0) parent.AsViaType = emit.method("fanx/util/OpUtil.as(Lfan/sys/Obj;Lfan/sys/Type;)Lfan/sys/Obj;");
      //  loadType(typeRef);
      //  code.op2(INVOKESTATIC, parent.AsViaType);
      //  code.op2(CHECKCAST, cls);
      //}
      //else
      //{
        //code.op(DUP);
        //code.op2(INSTANCEOF, cls);
        //int is = code.branch(IFNE);
        //code.op(POP);
        //code.op(ACONST_NULL);
        //int end = code.branch(GOTO);
        //code.mark(is);
        //code.op2(CHECKCAST, cls);
        //code.mark(end);
        code.TypeInst(TypeOp.isinst, type);
      //}
    }

  //////////////////////////////////////////////////////////////////////////
  // Switch
  //////////////////////////////////////////////////////////////////////////

    /*
    private void Tableswitch()
    {
      int count = u2();

      loadIntVal();
      code.op(L2I);
      int start = code.pos();
      code.op(TABLESWITCH);
      int pad = code.padAlign4();
      code.info.u4(1+pad+12+count*4); // default is always fall thru
      code.info.u4(0);
      code.info.u4(count-1);

      // at this point we don't know how fcode locations (abs) will
      // map to Java bytecode locations (rel offsets), so we just
      // keep track of locations to backpatch in a linked list
      for (int i=0; i<count; ++i)
      {
        JumpNode j = new JumpNode();
        j.fcodeLoc = u2();
        j.size     = 4;
        j.javaFrom = start;
        j.javaMark = code.info.len-CodeEmit.Header;
        j.next = jumps;
        jumps = j;
        code.info.u4(-1);  // place holder for backpatch
      }
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Misc
  //////////////////////////////////////////////////////////////////////////

    private void Cast()
    {
      code.TypeInst(TypeOp.castclass, emitter.FindType(pod.Nname(U2())));
    }

    private void DoThrow()
    {
      if (parent.ErrVal == null)
        parent.ErrVal = emitter.FindField("Fan.Sys.Err", "val", "Fan.Sys.Err/Val");
      code.FieldInst(FieldOp.ldfld, parent.ErrVal);
      code.Inst(Op.throwOp);
    }

    private void DoLeave()
    {
      code.Branch(BranchOp.leave, jumps.Add(U2()));
    }

    private void CatchAllStart()
    {
      exType = "System.Object";
      errBlocks.Push(getLastTryBlock());
      code.StartBlock();
      code.Inst(Op.pop);
    }

    private void CatchErrStart()
    {
      exType = "System.Exception";  // TODO - should this be Fan.Sys.Err/Val?
      for (int i=0; i<tryJump.Length; i++)
        if (startPos == tryJump[i])
        {
          exType = pod.Nname(tryErr[i]) + "/Val";
          break;
        }

      errBlocks.Push(getLastTryBlock());
      code.StartBlock();

      // there is already a System.Exception on the stack, but
      // we need to map into a sys::Err type
      if (parent.ErrMake == null)
        parent.ErrMake = emitter.FindMethod("Fan.Sys.Err", "Make",
          new string[] { "System.Exception" }, "Fan.Sys.Err");
      code.MethInst(MethodOp.call, parent.ErrMake);
      Cast();
    }

    private void CatchEnd()
    {
      PERWAPI.TryBlock lastTry = (PERWAPI.TryBlock)errBlocks.Pop();
      code.EndCatchBlock(emitter.FindType(exType) as PERWAPI.Class, lastTry);
    }

    private void FinallyStart()
    {
      errBlocks.Push(code.EndTryBlock());
      code.StartBlock();
    }

    private void FinallyEnd()
    {
      code.Inst(Op.endfinally);
      code.EndFinallyBlock((PERWAPI.TryBlock)errBlocks.Pop());
    }

    private PERWAPI.TryBlock getLastTryBlock()
    {
      for (int i=0; i<tryJump.Length; i++)
        if (startPos == tryJump[i])
        {
          int start  = tryStart[i];
          int end    = tryEnd[i];
          string key = start + "/" + end;

          PERWAPI.TryBlock block = (PERWAPI.TryBlock)tryBlocks[key];
          if (block == null)
          {
            block = code.EndTryBlock();
            tryBlocks[key] = block;
          }
          return block;
        }

      throw new System.Exception("This is not good");
    }

  //////////////////////////////////////////////////////////////////////////
  // Utils
  //////////////////////////////////////////////////////////////////////////

    /*
    private void LoadBoolVal()
    {
      if (parent.BoolVal == null)
        parent.BoolVal = emitter.FindField("Fan.Sys.Bool", "val", "System.Boolean");
      code.FieldInst(FieldOp.ldsfld, parent.BoolVal);
    }

    private void LoadIntVal()
    {
      if (parent.IntVal == 0) parent.IntVal = emit.field("fan/sys/Int.val:J");
      code.op2(GETFIELD, parent.IntVal);
    }
    */

    private void BoolMake()
    {
      if (parent.BoolMake == null)
        parent.BoolMake = emitter.FindMethod("Fan.Sys.Bool", "Make",
          new string[] { "System.Boolean" }, "Fan.Sys.Bool");
      code.MethInst(MethodOp.call, parent.BoolMake);
    }

  //////////////////////////////////////////////////////////////////////////
  // Buf
  //////////////////////////////////////////////////////////////////////////

    private int ConsumeOp()
    {
      // add labels
      code.CodeLabel(jumps.Add(pos));

      // check for try blocks
      int peek = PeekOp();
      int endPos = -1;
      for (int i=0; i<tryStart.Length; i++)
        //if (pos == tryStart[i])
        if (pos == tryStart[i] && peek != FConst.CatchErrStart && peek != FConst.CatchAllStart)
        {
          int jump = Peek(tryJump[i]);
          if (jump == FConst.CatchErrStart || jump == FConst.CatchAllStart)
          {
            // target is a catch, make sure we don't already
            // have a try block emitted for this region
            if (tryEnd[i] != endPos)
            {
              endPos = tryEnd[i];
              code.StartBlock();
            }
          }
          else
          {
            // target is a finally
            code.StartBlock();
          }
        }

      // get next opcode
      return U1();
    }

    private int PeekOp()
    {
      if (pos < len) return buf[pos];
      return -1;
    }

    private int Peek(int index)
    {
      if (index < len) return buf[index];
      return -1;
    }

    private int U1() { return buf[pos++]; }
    private int U2() { return (buf[pos++] & 0xFF) << 8 | (buf[pos++] & 0xFF); }
    private int U4() { return (buf[pos++] & 0xFF) << 24 | (buf[pos++] & 0xFF) << 16 | (buf[pos++] & 0xFF) << 8 | (buf[pos++] & 0xFF); }

  //////////////////////////////////////////////////////////////////////////
  // Jumps
  //////////////////////////////////////////////////////////////////////////

    internal class Jumps
    {
      public Jumps(CILInstructions code)
      {
        this.code = code;
      }

      public PERWAPI.CILLabel Add(int loc)
      {
        if (map[loc] == null)
          map[loc] = code.NewLabel();
        return (PERWAPI.CILLabel)map[loc];
      }

      public PERWAPI.CILLabel Get(int loc)
      {
        return (PERWAPI.CILLabel)map[loc];
      }

      Hashtable map = new Hashtable();
      CILInstructions code;
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    internal FPod pod;
    internal Emitter emitter;
    internal FTypeEmit parent;
    internal FMethod fmethod;        // maybe null
    internal byte[] buf;
    internal int len;
    internal int pos;
    internal int startPos;          // start pos when processing a new opcode
    internal CILInstructions code;
    internal String podClass;
    internal Jumps jumps;                    // map of jump locs to CodeLabels
    internal int[] tryStart = new int[0];    // start offsets of try blocks
    internal int[] tryEnd   = new int[0];    // end offsets of try blocks
    internal int[] tryJump  = new int[0];    // jump offset of try blocks
    internal int[] tryErr   = new int[0];    // err types for catch blocks
    internal string exType;                  // the exception type for the next catch block
    internal Hashtable tryBlocks = new Hashtable(); // hash of opened try blocks
    internal Stack errBlocks = new Stack();  // stack of try-catch-finally block offsets
  }
}