//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   11 Sep 06  Andy Frank  Creation
//

using System;
using System.Text;
using Fan.Sys;
using Uri = Fan.Sys.Uri;
using Fanx.Util;

namespace Fanx.Fcode
{
  ///
  /// FPod is the read/write fcode representation of sys::Pod.  It's main job in
  /// life is to manage all the pod-wide constant tables for names, literals,
  /// type/slot references and type/slot definitions.
  ///
  public sealed class FPod
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    public FPod(string podName, FStore store)
    {
      this.podName    = podName;
      this.store      = store;
      this.names      = new FTable.Names(this);
      this.typeRefs   = new FTable.TypeRefs(this);
      this.fieldRefs  = new FTable.FieldRefs(this);
      this.methodRefs = new FTable.MethodRefs(this);
      this.ints       = new FTable.Ints(this);
      this.floats     = new FTable.Floats(this);
      this.strs       = new FTable.Strs(this);
      this.durations  = new FTable.Durations(this);
      this.uris       = new FTable.Uris(this);
    }


  //////////////////////////////////////////////////////////////////////////
  // Lookup
  //////////////////////////////////////////////////////////////////////////

    public FType Type(string name)
    {
      for (int i=0; i<types.Length; i++)
        if (Name(TypeRef(types[i].selft).typeName) == name)
          return types[i];
      throw UnknownTypeErr.Make(name).val;
    }

  //////////////////////////////////////////////////////////////////////////
  // Convenience
  //////////////////////////////////////////////////////////////////////////

    public string Name(int index)       { return (string)names.Get(index); }
    public FTypeRef TypeRef(int index)  { return (FTypeRef)typeRefs.Get(index); }
    public FTuple FieldRef(int index)   { return (FTuple)fieldRefs.Get(index); }
    public FTuple MethodRef(int index)  { return (FTuple)methodRefs.Get(index); }
    public Int Integer(int index)       { return (Int)ints.Get(index); }
    public Float Float(int index)       { return (Float)floats.Get(index); }
    public Str Str(int index)           { return (Str)strs.Get(index); }
    public Duration Duration(int index) { return (Duration)durations.Get(index); }
    public Uri Uri(int index)           { return (Uri)uris.Get(index); }

  //////////////////////////////////////////////////////////////////////////
  // .NET Emit Utils
  //////////////////////////////////////////////////////////////////////////

    public class NMethod
    {
      public string parentType;
      public string methodName;
      public string returnType;
      public string[] paramTypes;
      public bool isStatic = false;
    }

    /// <summary>
    /// Map a fcode method signature to a .NET method emit signature.
    /// </summary>
    public NMethod Ncall(int index, int opcode)
    {
      if (ncalls == null) ncalls = new NMethod[methodRefs.Size()];
      NMethod ncall = ncalls[index];
      if (ncall == null)
      {
        int[] m = MethodRef(index).val;
        string parent = Nname(m[0]);
        string name = NameUtil.Upper(Name(m[1]));
        bool onObj = parent == "Fan.Sys.Obj";
        string[] pars = new string[m.Length-3];
        for (int i=0; i<pars.Length; i++) pars[i] = Nname(m[i+3]);

        // static methods on sys::Obj are really FanObj
        if (onObj && (opcode == FConst.CallStatic || opcode == FConst.CallNonVirtual))
        {
          parent = "Fan.Sys.FanObj";
        }

        string ret = Nname(m[2]);
        if (opcode == FConst.CallNew) ret = parent; // factory

        ncall = new NMethod();
        ncall.parentType = parent;
        ncall.methodName = name;
        ncall.returnType = ret;
        ncall.paramTypes = pars;
        if (!onObj) ncalls[index] = ncall;
      }
      return ncall;
    }

    public class NField
    {
      public string parentType;
      public string fieldName;
      public string fieldType;
    }

    /// <summary>
    /// Map a fcode field signature to a .NET field emit signature.
    /// </summary>
    public NField Nfield(int index, bool mixin)
    {
      if (nfields == null) nfields = new NField[fieldRefs.Size()];
      NField nfield = nfields[index];
      if (nfield == null)
      {
        int[] v = FieldRef(index).val;
        nfield = new NField();
        nfield.parentType = Nname(v[0]);
        nfield.fieldName  = Name(v[1]);
        nfield.fieldType  = Nname(v[2]);
      }
      return nfield;
    }

    /// <summary>
    /// Map a .NET type name (:: replaced wit .) via typeRefs table.
    /// <summary>
    public string Nname(int index)
    {
      if (nnames == null) nnames = new string[typeRefs.Size()];
      string n = nnames[index];
      if (n == null)
      {
        FTypeRef refer = TypeRef(index);
        n = nnames[index] = "Fan." + NameUtil.Upper(Name(refer.podName)) + '.' + Name(refer.typeName);
      }
      return n;
    }

  //////////////////////////////////////////////////////////////////////////
  // Read
  //////////////////////////////////////////////////////////////////////////

    ///
    /// Read from a FStore which provides random access.
    ///
    public void Read()
    {
      names.Read(store.Read("names.def", true));
      typeRefs.Read(store.Read("typeRefs.def"));
      fieldRefs.Read(store.Read("fieldRefs.def"));
      methodRefs.Read(store.Read("methodRefs.def"));
      ints.Read(store.Read("ints.def"));
      floats.Read(store.Read("floats.def"));
      strs.Read(store.Read("strs.def"));
      durations.Read(store.Read("durations.def"));
      uris.Read(store.Read("uris.def"));

      // pod meta
      ReadPodMeta(store.Read("pod.def", true));

      // type meta
      ReadTypeMeta(store.Read("types.def", true));

      // full fcode always lazy loaded in Type.reflect()
    }

    private void ReadPodMeta(FStore.Input input)
    {
      int first2 = input.U2();
      if (first2 != 0x0FC0)
      {
        System.Console.WriteLine("FPod.readPodMeta: old version of " + podName);
        version = "0";
        depends = new Depend[0];
        facets  = new FFacet[0];
        return;
      }
      if (input.U2() != 0xDE05)
        throw new System.IO.IOException("Invalid magic");

      int ver = input.U4();
      if (ver != FConst.FCodeVersion)
        throw new System.IO.IOException("Invalid version 0x" + ver.ToString("x"));

      podName = input.Utf();
      version = input.Utf();
      depends = new Depend[input.U1()];
      for (int i=0; i<depends.Length; i++)
        depends[i] = Depend.Parse(Fan.Sys.Str.Make(input.Utf()));

      facets = new FFacet[input.U2()];
      for (int i=0; i<facets.Length; i++)
        facets[i] = new FFacet().Read(input);

      input.Close();
    }

    private void ReadTypeMeta(FStore.Input input)
    {
      types = new FType[input.U2()];
      for (int i=0; i<types.Length; i++)
      {
        types[i] = new FType(this).ReadMeta(input);
        types[i].hollow = true;
      }
      input.Close();
    }

    /*
    private void ReadType(String name, FStore.Input input)
    {
      if (types == null || types.length == 0)
        throw new IOException("types.def must be defined first");

      String typeName = name.substring(0, name.length()-".fcode".length());
      for (int i=0; i<types.length; ++i)
      {
        String n = name(typeRef(types[i].self).typeName);
        if (n.equals(typeName )) { types[i].read(input); return; }
      }

      throw new IOException("Unexpected fcode file: " + name);
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public string podName;     // pod's unique name
    public string version;     // pod version
    public Depend[] depends;   // pod dependencies
    public FFacet[] facets;    // pod facets
    public FStore store;       // store we using to read/write
    public FType[] types;      // pod's declared types
    public FTable names;       // identifier names: foo
    public FTable typeRefs;    // types refs:   [pod,type,variances*]
    public FTable fieldRefs;   // fields refs:  [parent,name,type]
    public FTable methodRefs;  // methods refs: [parent,name,ret,params*]
    public FTable ints;        // Int literals
    public FTable floats;      // Float literals
    public FTable strs;        // Str literals
    public FTable durations;   // Duration literals
    public FTable uris;        // Uri literals
    private string[] nnames;   // cached fan typeRef   -> .net name
    private NMethod[] ncalls;  // cached fan methodRef -> .net method signatures
    private NField[] nfields;  // cached fan fieldRef  -> .net field signatures
  }
}