//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   12 Sep 06  Andy Frank  Creation
//

using System;
using System.IO;
using Fanx.Util;

namespace Fanx.Fcode
{
  ///
  /// FPod is the read/write fcode representation of sys::Pod.
  ///
  public class FPrinter : StreamWriter
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    public FPrinter(FPod pod) : this(pod, Console.OpenStandardOutput())
    {
    }

    public FPrinter(FPod pod, Stream stream) : base(stream)
    {
      this.pod = pod;
    }

  //////////////////////////////////////////////////////////////////////////
  // Dump
  //////////////////////////////////////////////////////////////////////////

    public void All()
    {
      Tables();
      Types();
      Flush();
    }

  //////////////////////////////////////////////////////////////////////////
  // Const Tables
  //////////////////////////////////////////////////////////////////////////

    public void Tables()
    {
      WriteLine("##### Tables #####");
      WriteLine("--- names ---");      pod.names.Dump(pod, this);
      WriteLine("--- typeRefs ---");   pod.typeRefs.Dump(pod, this);
      WriteLine("--- fieldRefs ---");  pod.fieldRefs.Dump(pod, this);
      WriteLine("--- methodRefs ---"); pod.methodRefs.Dump(pod, this);
      WriteLine("--- ints ---");       pod.ints.Dump(pod, this);
      WriteLine("--- floats ---");     pod.floats.Dump(pod, this);
      WriteLine("--- strs ---");       pod.strs.Dump(pod, this);
      WriteLine("--- durations ---");  pod.durations.Dump(pod, this);
      WriteLine("--- uris ---");       pod.uris.Dump(pod, this);
      Flush();
    }

  //////////////////////////////////////////////////////////////////////////
  // Types
  //////////////////////////////////////////////////////////////////////////

    public void Types()
    {
      WriteLine("##### Types #####");
      for (int i=0; i<pod.types.Length; i++)
        Type(pod.types[i]);
      Flush();
    }

    public void Type(FType type)
    {
      WriteLine("--" + TypeRef(type.selft) + " extends " + TypeRef(type.baset) + "--");
      if (type.mixins.Length > 0)
      {
        Write("  mixin ");
        for (int i=0; i<type.mixins.Length; i++)
        {
          if (i > 0) Write(", ");
          Write(TypeRef(type.mixins[i]));
        }
        WriteLine();
      }
      //facets(type.facets);
      WriteLine();
      for (int i=0; i<type.fields.Length; i++)  Field(type.fields[i]);
      for (int i=0; i<type.methods.Length; i++) Method(type.methods[i]);
      Flush();
    }
    public void Slot(FSlot s)
    {
      if (s is FField)
        Field((FField)s);
      else
        Method((FMethod)s);
      Flush();
    }

    public void Field(FField f)
    {
      WriteLine("  " + TypeRef(f.type) + " " + Name(f.name) + " [" + StrUtil.FlagsToString(f.flags).Trim() + "]");
      //facets(f.facets);
      WriteLine();
    }

    public void Method(FMethod m)
    {
      Write("  " + TypeRef(m.ret) + " " + Name(m.name) + "(");
      FMethodVar[] pars= m.pars();
      for (int i=0; i<pars.Length; ++i)
      {
        FMethodVar p = pars[i];
        if (i > 0) Write(", ");
        Write(TypeRef(p.type) + " " + Name(p.name));
      }
      WriteLine(") [" + StrUtil.FlagsToString(m.flags).Trim() + "]");
      for (int i=0; i<m.vars.Length; i++)
      {
        FMethodVar v = m.vars[i];
        string role = v.IsParam() ?  "Param" : "Local";
        int reg = i + ((m.flags & FConst.Static) != 0 ? 0 : 1);
        WriteLine("    [" + role + " " + reg + "] " + Name(v.name) + ": " + TypeRef(v.type));
        if (v.def != null) Code(v.def);
      }
      if (m.code != null)
      {
        WriteLine("    [Code]");
        Code(m.code);
      }
      //facets(m.facets);
      WriteLine();
    }

    public void Code(FBuf code)
    {
      if (!showCode) return;
      Flush();
      //new FCodePrinter(pod, out).code(code);
      WriteLine("      CODE - TODO");
    }

  //////////////////////////////////////////////////////////////////////////
  // Dump Utils
  //////////////////////////////////////////////////////////////////////////

    private string TypeRef(int index)
    {
      if (index == 65535) return "null";
      return pod.typeRefs.ToString(index) + ShowIndex(index);
    }

    private string Name(int index)
    {
      return pod.Name(index) + ShowIndex(index);
    }

    private string ShowIndex(int index)
    {
      if (showIndex) return "[" + index + "]";
      return "";
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public readonly FPod pod;
    public bool showIndex = false;
    public bool showCode  = true;
    public bool showLines = false;
  }
}