//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   19 Sep 06  Andy Frank  Creation
//

using System;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using Fan.Sys;
using Fanx.Emit;
using Fanx.Fcode;
using Fanx.Util;

namespace Fanx.Test
{
  /// <summary>
  /// EmitTest.
  /// </summary>
  public class EmitTest : Test
  {
    public override void Run()
    {
      VerifyBasic();
      VerifySysFindType();
      VerifyTypeParser();
      VerifySysReflect();
    }

    public void VerifyBasic()
    {
      Emitter e = new Emitter("foo");
      e.EmitClass("System.Object", "Foo.Bar", PERWAPI.TypeAttr.Public);

      e.EmitField("FieldA", "System.Boolean", PERWAPI.FieldAttr.Public);
      e.EmitField("FieldB", "System.UInt32",  PERWAPI.FieldAttr.Public);
      e.EmitField("FieldC", "System.String",  PERWAPI.FieldAttr.Public);

      // TODO - methods require 'ret' - should auto handle that in API
      e.EmitMethod("MethodA", "System.Void", s(), s(),
        PERWAPI.MethAttr.PublicStatic, new string[0], new string[0]).Inst(PERWAPI.Op.ret);
      e.EmitMethod("MethodB", "System.Void", s("pa"), s("System.UInt32"),
        PERWAPI.MethAttr.PublicStatic, new string[0], new string[0]).Inst(PERWAPI.Op.ret);
      e.EmitMethod("MethodC", "System.Void", s("pa", "pb"), s("System.UInt32", "System.Double"),
        PERWAPI.MethAttr.PublicStatic, new string[0], new string[0]).Inst(PERWAPI.Op.ret);

      byte[] buf = e.Commit();

      // dump to file
      //BinaryWriter writer = new BinaryWriter(
      //  new FileInfo(@"c:\dev\fan\test.dll").OpenWrite());
      //writer.Write(buf, 0, buf.Length);
      //writer.Flush();
      //writer.Close();

      Assembly assembly = Assembly.Load(buf);
      System.Type type = assembly.GetType("Foo.Bar");
      Verify(type.ToString(), "Foo.Bar");

      Verify(type.GetField("FieldA").ToString(), "Boolean FieldA");
      Verify(type.GetField("FieldB").ToString(), "UInt32 FieldB");
      Verify(type.GetField("FieldC").ToString(), "System.String FieldC");

      Verify(type.GetMethod("MethodA").ToString(), "Void MethodA()");
      Verify(type.GetMethod("MethodB").ToString(), "Void MethodB(UInt32)");
      Verify(type.GetMethod("MethodC").ToString(), "Void MethodC(UInt32, Double)");
    }

    public void VerifySysFindType()
    {
      Pod pod = Sys.FindPod("sys", true, null);
      Verify(pod.FindType("Bool", true) == Sys.BoolType);
      Verify(pod.FindType("Int", true)  == Sys.IntType);
      Verify(pod.FindType("Str", true)  == Sys.StrType);

      Verify(Sys.FindType(Str.Make("sys::Bool")) == Sys.BoolType);
      Verify(Sys.FindType(Str.Make("sys::Bool"), Bool.True) == Sys.BoolType);
      Verify(Sys.FindType("sys::Bool") == Sys.BoolType);
      Verify(Sys.FindType("sys::Bool", true) == Sys.BoolType);
      Verify(Sys.FindType("sys", "Bool", true) == Sys.BoolType);
    }

    public void VerifyTypeParser()
    {
      VerifyTypeParserErr("");
      VerifyTypeParserErr("x");
      VerifyTypeParserErr("xy");
      VerifyTypeParserErr("xyz");
      VerifyTypeParserErr("x:z");
      VerifyTypeParserErr("xz[");
      VerifyTypeParserErr("xz[]");
      VerifyTypeParserErr("[]");
    }

    public void VerifyTypeParserErr(string sig)
    {
      try
      {
        Sys.FindType(sig, false);
        Console.WriteLine("didn't fail: " + sig);
        Fail();
      }
      catch(Err.Val err)
      {
        //Console.WriteLine(" -- " + err);
        Verify(err.Err().Message().val.StartsWith("Invalid type signature '" + sig + "'"));
      }
    }

    public void VerifySysReflect()
    {
      VerifySysImpl("sys::Bool", "Fan.Sys.Bool");
      VerifySysImpl("sys::Int",  "Fan.Sys.Int");
      VerifySysImpl("sys::Str",  "Fan.Sys.Str");
    }

    public void VerifySysImpl(string fname, string nname)
    {
      Fan.Sys.Type ftype = Sys.FindType(fname, true);
      System.Type ntype  = System.Type.GetType(nname);

      Verify(ftype.Emit(), ntype);

      /*
      for (int i=0; i<ftype.Fields().Sz(); i++)
      {
        Field ffield = (Field)ftype.Fields().Get(i);
        string name = NameUtil.Upper(ffield.Name().val);
        FieldInfo nfield = ntype.GetField(name);
        Verify(name, nfield.Name);
      }
      */

      /*
      for (int i=0; i<ftype.Methods().Sz(); i++)
      {
        Method fmeth = (Method)ftype.Methods().Get(i);
        string name = NameUtil.Upper(fmeth.Name().val);
        MethodInfo nmeth = ntype.GetMethod(name);
        Verify(name, nmeth.Name);
      }
      */
    }

    /*
    public void VerifyFooBar()
    {
      string code = "class Foo {\n" +
        "  Bool b\n" +
        "  Int i\n" +
        "  Str s\n" +
        "}";
      stub(code);
      compile();

      Pod pod = Sys.FindPod("nsystest", true, null);
      Fan.Sys.Type type = pod.FindType("Foo", true);
    }

    private void stub(String code)
    {
      string path = @"c:\dev\fan\src\nsystest\fan";
      Directory.CreateDirectory(path);
      StreamWriter w = new StreamWriter(new FileInfo(path + @"\Foo.fan").OpenWrite());
      w.WriteLine(code);
      w.Flush();
      w.Close();
    }

    private void compile()
    {
      Process proc = new Process();
      proc.StartInfo.WorkingDirectory = @"c:\dev\fan\bin";
      proc.StartInfo.FileName = @"c:\dev\fan\bin\ntestc.bat";
      proc.StartInfo.CreateNoWindow = true;
      proc.Start();
      proc.WaitForExit();
    }
    */

    private string[] s() { return new string[0]; }
    private string[] s(string a) { return new string[] { a }; }
    private string[] s(string a, string b) { return new string[] { a, b }; }
  }
}