//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   26 Mar 06  Brian Frank  Creation
//

**
** File is used to represent a Uri path to a file or directory.
**
abstract class File
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  **
  ** Make a File for the Uri which represents a file on the local
  ** file system.  If creating a Uri to a directory, then the Uri
  ** must end in a trailing "/" slash - Uri.isDir() must return true.
  ** If the file exists and it doesn't match Uri.isDir() then IOErr
  ** is thrown.  If the file doesn't exist, then it is assumed to
  ** to a directory based on Uri.isDir().  If the Uri has a relative
  ** path, then it is assumed to be relative to the current working
  ** directory.
  **
  static File make(Uri uri)

  **
  ** Make a File for the specified operating system specific path
  ** on the local file system.
  **
  static File os(Str osPath)

  **
  ** Create a temporary file which is guaranteed to be a new, empty
  ** file with a unique name.  The file name will be generated using
  ** the specified prefix and suffix.  If dir is non-null then it is used
  ** as the file's parent directory, otherwise the system's default
  ** temporary file is used.  If dir is specified it must be a directory
  ** on the local file system.  See deleteOnExit() if you wish to have
  ** the file automatically deleted on exit.  Throw IOErr on error.
  **
  ** Examples:
  **   File.createTemp("x", ".txt") -> `/tmp/x67392.txt`
  **   File.createTemp.deleteOnExit -> `/tmp/fan5284.tmp`
  **
  static File createTemp(Str prefix := "fan", Str suffix := ".tmp", File dir := null)

  **
  ** Private constructor
  **
  private new init(Uri uri)

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

  **
  ** File equality is based on the un-normalized Uri used to create the File.
  **
  override Bool equals(Obj that)

  **
  ** Return uri.hash.
  **
  override Int hash()

  **
  ** Return uri.toStr.
  **
  override Str toStr()

//////////////////////////////////////////////////////////////////////////
// Naming
//////////////////////////////////////////////////////////////////////////

  **
  ** Return the Uri path used to create this File.  This Uri
  ** may be absolute or relative.
  **
  Uri uri()

  **
  ** Convenience for uri.isDir
  **
  Bool isDir()

  **
  ** Convenience for uri.path.
  **
  Str[] path()

  **
  ** Convenience for uri.pathStr.
  **
  Str pathStr()

  **
  ** Convenience for uri.name.
  **
  Str name()

  **
  ** Convenience for uri.basename.
  **
  Str basename()

  **
  ** Convenience for uri.ext.
  **
  Str ext()

//////////////////////////////////////////////////////////////////////////
// Access
//////////////////////////////////////////////////////////////////////////

  **
  ** Return if this file exists in the file system.
  **
  abstract Bool exists()

  **
  ** Return the size of the file in bytes otherwise null if a
  ** directory or unknown.
  **
  abstract Int size()

  **
  ** Get or set the time the file was last modified.
  ** Modify time is null if unknown.
  **
  Time modified /* TODO { abstract get, abstract set }*/

  **
  ** Get this File as an operating system specific path on
  ** the local system.  If this File doesn't represent a
  ** path on the local file system then return null.
  **
  abstract Str osPath()

  **
  ** Get the parent directory of this file or null.
  ** Also see Uri.parent().
  **
  abstract File parent()

  **
  ** List the files contained by this directory.  If the directory is
  ** empty or this file doesn't represent a directory, then return
  ** an empty list.
  **
  abstract File[] list()

  **
  ** Normalize this file path to its canonical representation.
  **
  abstract File normalize()

  **
  ** Make a new File instance by joining this file's Uri
  ** together with the specified path.
  **
  ** Example:
  **  File.make(`a/b/`) + `c` -> File.make(`a/b/c`)
  **  File.make(`a/b`) + `c` -> File.make(`a/c`)
  **
  abstract File plus(Uri path)

//////////////////////////////////////////////////////////////////////////
// Management
//////////////////////////////////////////////////////////////////////////

  **
  ** Create a file or directory represented by this Uri.  If isDir() is
  ** false then create an empty file, or if the file already exists
  ** overwrite it to empty.  If isDir() is true then create a directory,
  ** or if the directory already exists do nothing.  This method will
  ** automatically create any parent directories.  Throw IOErr on error.
  ** Return this.
  **
  abstract File create()

  **
  ** Delete this file.  If this file represents a directory, then
  ** recursively delete it.  If the file does not exist, then no
  ** action is taken.  Throw IOErr on error.
  **
  abstract Void delete()

  **
  ** Request that the file or directory represented by this File
  ** be deleted when the virtual machine exits.  Long running applications
  ** should use this method will care since each file marked to delete will
  ** consume resources.  Throw IOErr on error.  Return this.
  **
  abstract File deleteOnExit()

//////////////////////////////////////////////////////////////////////////
// IO
//////////////////////////////////////////////////////////////////////////

  **
  ** Open a new buffered InStream used to read from this file.  A
  ** bufferSize of zero will return an unbuffered input stream.
  ** Throw IOErr on error.
  **
  abstract InStream in(Int bufferSize := 4096)

  **
  ** Open a new buffered OutStream used to write to this file.  If append is
  ** true, then we open the file to append to the end, otherwise it is
  ** opened as an empty file.  A bufferSize of zero will return an unbuffered
  ** input stream.  Throw IOErr on error.
  **
  abstract OutStream out(Bool append := false, Int bufferSize := 4096)

  **
  ** Convenience for in.readAllBuf().
  ** The input stream is guaranteed to be closed.
  **
  Buf readAllBuf()

  **
  ** Convenience for in.readAllLines().
  ** The input stream is guaranteed to be closed.
  **
  Str[] readAllLines()

  **
  ** Convenience for in.readAllStr(normalizeNewlines).
  ** The input stream is guaranteed to be closed.
  **
  Str readAllStr(Bool normalizeNewlines := true)

  **
  ** Convenience for in.readProps().
  ** The input stream is guaranteed to be closed.
  **
  Map readProps()

  **
  ** Convenience for out.writeProps().
  ** The output stream is guaranteed to be closed.
  **
  Void writeProps(Str:Str props)

  // TODO:
  //  - random access as Buf
  //  - copy, move, rename, etc

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  **
  ** Return the platform's file path separator - typically a slash.
  ** TODO: make field?
  **
  static Str sep()

}

**************************************************************************
** LocalFile
**************************************************************************

internal class LocalFile : File
{
  private new init()
  override Bool exists()
  override Int size()
  override Str osPath()
  override File parent()
  override File[] list()
  override File normalize()
  override File plus(Uri uri)
  override File create()
  override Void delete()
  override File deleteOnExit()
  override InStream in(Int bufferSize := 4096)
  override OutStream out(Bool append := false, Int bufferSize := 4096)
}

**************************************************************************
** ZipEntryFile
**************************************************************************

internal class ZipEntryFile : File
{
  private new init()
  override Bool exists()
  override Int size()
  override Str osPath()
  override File parent()
  override File[] list()
  override File normalize()
  override File plus(Uri uri)
  override File create()
  override Void delete()
  override File deleteOnExit()
  override InStream in(Int bufferSize := 4096)
  override OutStream out(Bool append := false, Int bufferSize := 4096)
}