//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   16 Mar 06  Brian Frank  Creation
//

**
** OutStream is used to write data to an output stream.  OutStreams are
** always buffered for performance.
**
mixin OutStream
{

//////////////////////////////////////////////////////////////////////////
// Factories
//////////////////////////////////////////////////////////////////////////

  **
  ** Make an output stream designed to write characters into the specified
  ** StrBuf.  The output stream is designed to write character data, attempts
  ** to do binary writes will throw UnsupportedErr.
  **
  static OutStream makeForStrBuf(StrBuf strBuf)

//////////////////////////////////////////////////////////////////////////
// Abstract
//////////////////////////////////////////////////////////////////////////

  **
  ** Write a byte to the output stream.  Throw IOErr on error.
  ** Return this.
  **
  abstract OutStream write(Int byte)

  **
  ** Write n bytes from the specified Buf at it's current position to
  ** this output stream.  If n is defaulted to buf.remaining(), then
  ** everything left in the buffer is drained to this output stream.
  ** The buf's position is advanced n bytes upon return.  Throw
  ** IOErr on error.  Return this.
  **
  abstract OutStream writeBuf(Buf buf, Int n := buf.remaining)

//////////////////////////////////////////////////////////////////////////
// Binary Data
//////////////////////////////////////////////////////////////////////////

  **
  ** Write two bytes as a 16-bit number in network byte order.  This method
  ** may be paired with InStream.readU2 or InStream.readS2.  Throw IOErr
  ** on error.  Return this.
  **
  virtual OutStream writeI2(Int n)

  **
  ** Write four bytes as a 32-bit number in network byte order.  This
  ** method may be paired with InStream.readU4 or InStream.readS4.  Throw
  ** IOErr on error.  Return this.
  **
  virtual OutStream writeI4(Int n)

  **
  ** Write eight bytes as a 64-bit number in network byte order.  This
  ** is paired with InStream.readS8.  Throw IOErr on error.  Return this.
  **
  virtual OutStream writeI8(Int n)

  **
  ** Write four bytes as a 32-bit floating point number in network byte
  ** order according to Float.bits32.  This is paired with InStream.readF4.
  ** Throw IOErr on error.  Return this.
  **
  virtual OutStream writeF4(Float r)

  **
  ** Write eight bytes as a 64-bit floating point number in network byte
  ** order according to Float.bits.  This is paired with InStream.readF8.
  ** Throw IOErr on error.  Return this.
  **
  virtual OutStream writeF8(Float r)

  **
  ** Write one byte, one if true or zero if false.  This method is paired
  ** with InStream.readBool.  Throw IOErr on error.  Return this.
  **
  virtual OutStream writeBool(Bool b)

  **
  ** Write a Str in modified UTF-8 format according the java.io.DataOutput
  ** specification.  This method is paired with InStream.readUtf.  Throw
  ** IOErr on error.  Return this.
  **
  virtual OutStream writeUtf(Str s)

//////////////////////////////////////////////////////////////////////////
// Text Data
//////////////////////////////////////////////////////////////////////////

  **
  ** The current charset used to encode Unicode characters into
  ** bytes.  The default charset should always be UTF-8.
  **
  abstract Charset charset

  **
  ** Write one or more bytes to the stream for the specified Unicode
  ** character based on the current charset encoding.
  **
  abstract OutStream writeChar(Int char)

  **
  ** Write the Unicode characters in the specified string to the
  ** stream using the current charset encoding.  Off specifies
  ** the index offset to start writing characters and len the
  ** number of characters in str to write.
  ** - TODO match writeBuf (no offset, use pos instead)
  abstract OutStream writeChars(Str str, Int off := 0, Int len := str.size-off)

  **
  ** Convenience for writeChars(obj.toStr).  If obj is null,
  ** then print the string "null".
  **
  virtual OutStream print(Obj s)

  **
  ** Convenience for writeChars(obj.toStr + "\n").  If obj
  ** is null then print the string "null\n".
  **
  virtual OutStream printLine(Obj obj := "")

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  **
  ** Write the given map of Str name/value pairs to the output stream
  ** according to the Fan props file format (see InStream.readProps for
  ** full specification).  The props are written using UTF-8 regardless
  ** of this stream's current charset.  If close argument is true, then
  ** automatically close the stream.  Return this.
  **
  virtual OutStream writeProps(Str:Str props, Bool close := true)

  **
  ** Flush the stream so any buffered bytes are written out.  Default
  ** implementation does nothing.  Throw IOErr on error.
  **
  virtual Void flush()

  **
  ** Close the input stream.  This method is guaranteed to never
  ** throw an IOErr.  Return true if the stream was closed successfully
  ** or false if the stream was closed abnormally.  Default implementation
  ** does nothing and returns true.
  **
  virtual Bool close()

}

**************************************************************************
** SysOutStream
**************************************************************************

internal class SysOutStream : OutStream
{
  override OutStream write(Int byte)
  override OutStream writeBuf(Buf buf, Int n)
  override OutStream writeChar(Int char)
  override OutStream writeChars(Str s, Int off, Int n)
  override Charset charset
}