//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   4 Jul 06 - Fireworks!  Brian Frank  Creation
//

**
** FacetsTests
**
class FacetsTest : Test
{
  Str aField;
  Void aMethod() {}

//////////////////////////////////////////////////////////////////////////
// Map
//////////////////////////////////////////////////////////////////////////

  Void testReflection()
  {
    verifyEq(type.facets["LineNumber"], "12")
    verifyEq(type.facets["SourceFile"], "FacetsTest.fan")
    verifyEq(type.lineNumber, 12)
    verifyEq(type.sourceFile, "FacetsTest.fan")
    verifyEq(type.facets.isRO, true)
    verifyErr(ReadonlyErr.type) |,| { FacetsTest.type.facets.set("x", "y") }

    field := type.field("aField");
    verifyEq(field.facets.isRO, true)
    verifyEq(field.facets["LineNumber"], "14")
    verifyEq(field.lineNumber, 14)
    verifyErr(ReadonlyErr.type) |,| { field.facets.set("x", "y") }

    method := type.method("aMethod");
    verifyEq(method.facets.isRO, true)
    verifyEq(method.facets["LineNumber"], "15")
    verifyEq(method.lineNumber, 15)
    verifyErr(ReadonlyErr.type) |,| { method.facets.set("x", "y") }
  }

//////////////////////////////////////////////////////////////////////////
// Map
//////////////////////////////////////////////////////////////////////////

  Void testMap()
  {
    f := Facets.make
    verifyEq(f.size, 0)
    verifyEq(f.keys, Str[,])

    f.set("strA", "")
    f.set("strB", "hello world")
    f.setBool("boolF", false)
    f.setBool("boolT", true)
    f.setInt("intA", 0)
    f.setInt("intB", 0xFFFF)
    f.setFloat("floatA", 7.3)
    f.setFloat("floatB", Float.nan)

    verifyEq(f.size, 8)
    verifyEq(f.keys.sort, ["strA", "strB", "boolF", "boolT", "floatA", "floatB", "intA", "intB"].sort)

    verifyEq(f.get("foo"), null)
    verifyEq(f.get("foo", "roo"), "roo")
    verifyEq(f["foo"], null)
    verifyEq(f.get("strA"), "")
    verifyEq(f.get("strB", "roo"), "hello world")
    verifyEq(f["strB"], "hello world")

    verifyEq(f.getBool("foo"), null)
    verifyEq(f.getBool("foo", true), true)
    verifyEq(f.getBool("boolT"), true)
    verifyEq(f.getBool("boolF", true), false)

    verifyEq(f.getInt("foo"), null)
    verifyEq(f.getInt("foo", -1), -1)
    verifyEq(f.getInt("intA"), 0)
    verifyEq(f.getInt("intB", -1), 0xFFFF)

    verifyEq(f.getFloat("foo"), null)
    verifyEq(f.getFloat("foo", -1f), -1f)
    verifyEq(f.getFloat("floatA"), 7.3)
    verifyEq(f.getFloat("floatB", 9.0), Float.nan)

    verifyEq(f.remove("blah"), null)
    verifyEq(f.remove("floatB"), Float.nan.toStr)
    verifyEq(f.remove("strB"), "hello world")
    verifyEq(f.size, 6)
    verifyEq(f.keys.sort, ["strA", "boolF", "boolT", "intA", "intB", "floatA"].sort)

    f.clear
    verifyEq(f.isEmpty, true)
    verifyEq(f.size, 0)
    verifyEq(f.keys, Str[,])
  }

//////////////////////////////////////////////////////////////////////////
// Readonly
//////////////////////////////////////////////////////////////////////////

  Void testReadonly()
  {
    // create facets
    x := Facets.make.set("foo", "bar")
    verifyEq(x.isRW, true)
    verifyEq(x.isRO, false)
    verifySame(x.rw, x)

    // get ro
    r := x.ro
    verifyEq(x.isRW, true)
    verifyEq(x.isRO, false)
    verifySame(x.rw, x)
    verifyEq(r.isRW, false)
    verifyEq(r.isRO, true)
    verifyEq(r, x)

    // verify all idempotent methods work
    verifyEq(r.type, Facets.type)
    verifyEq(r.isEmpty, false)
    verifyEq(r.size, 1)
    verifyEq(r["foo"], "bar")
    verifyEq(r.get("foo"), "bar")
    verifyEq(r.getBool("x"), null)
    verifyEq(r.getInt("x"),  null)
    verifyEq(r.getFloat("x"), null)
    verifyEq(r.keys, ["foo"])
    verifyEq(r.toStr, Facets.make.set("foo", "bar").toStr)

    // verify all modification methods throw ReadonlyErr
    verifyErr(ReadonlyErr.type) |,| { r["x"] = "y" }
    verifyErr(ReadonlyErr.type) |,| { r.set("x", "y") }
    verifyErr(ReadonlyErr.type) |,| { r.setBool("x", true) }
    verifyErr(ReadonlyErr.type) |,| { r.setInt("x", 99)}
    verifyErr(ReadonlyErr.type) |,| { r.setFloat("x", 9.9)}
    verifyErr(ReadonlyErr.type) |,| { r.remove("x")}
    verifyErr(ReadonlyErr.type) |,| { r.clear()}

    // verify rw detaches ro
    x["x"] = "y"
    r2 := x.ro
    verifyNotSame(r2, r)
    verifyNotSame(x.ro, r)
    verifyEq(r.isRO, true)
    verifyEq(r.size, 1)
    verifyEq(r, Facets.make.set("foo", "bar"))
    verifyNotEq(r, x)
    verifyNotEq(r, r2)

    // verify ro to rw
    y := r.rw
    verifyEq(y.isRW, true)
    verifyEq(y.isRO, false)
    verifySame(y.rw, y)
    verifyEq(y, r)
    verifyEq(r.isRO, true)
    verifyEq(r.size, 1)
    verifyEq(r, Facets.make.set("foo", "bar"))
    verifyEq(y, Facets.make.set("foo", "bar"))
  }

}