//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   28 May 06  Andy Frank  Creation
//

**
** WebOutStream provides methods for generating XML and XHTML content.
**
mixin WebOutStream : OutStream
{

//////////////////////////////////////////////////////////////////////////
// Factory
//////////////////////////////////////////////////////////////////////////

  **
  ** Return a WebOutStream that wraps the given OutStream.
  **
  static WebOutStream wrap(OutStream out)
  {
    return WebOutStreamWrapper.make(out)
  }

//////////////////////////////////////////////////////////////////////////
// General Methods
//////////////////////////////////////////////////////////////////////////

  **
  ** Convenience for writeChars(s).
  **
  WebOutStream w(Str s)
  {
    writeChars(s)
    return this
  }

  **
  ** Convenience for writeChars(Str.spaces(numSpaces)).
  **
  WebOutStream tab(Int numSpaces := 2)
  {
    writeChars(Str.spaces(numSpaces))
    return this
  }

  **
  ** Convenience for writeChar('\n').
  **
  WebOutStream nl()
  {
    writeChar('\n')
    return this
  }

//////////////////////////////////////////////////////////////////////////
// Xml Methods
//////////////////////////////////////////////////////////////////////////

  **
  ** Write a start tag. Use attrs to fully specify the attributes
  ** manually. Use empty to optionally close this element without
  ** using an end tag.
  **
  WebOutStream tag(Str elemName, Str attrs := null, Bool empty := false)
  {
    writeChar('<')
    writeChars(elemName)
    if (attrs != null) writeChar(' ').writeChars(attrs)
    if (empty) writeChars(" /")
    writeChar('>')
    return this
  }

  **
  ** Write an end tag.
  **
  WebOutStream tagEnd(Str elemName)
  {
    writeChars("</").writeChars(elemName).writeChar('>')
    return this
  }

//////////////////////////////////////////////////////////////////////////
// DOCTYPE
//////////////////////////////////////////////////////////////////////////

  **
  ** Write the XHTML Transitional DOCTYPE.
  **
  WebOutStream docType()
  {
    writeChars("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\"\n")
    writeChars(" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n")
    return this
  }

//////////////////////////////////////////////////////////////////////////
// html
//////////////////////////////////////////////////////////////////////////

  **
  ** Start a <html> tag.
  **
  WebOutStream html()
  {
    return tag("html").nl
  }

  **
  ** End a <html> tag.
  **
  WebOutStream htmlEnd()
  {
    return tagEnd("html").nl
  }

//////////////////////////////////////////////////////////////////////////
// head
//////////////////////////////////////////////////////////////////////////

  **
  ** Start a <head> tag.
  **
  WebOutStream head()
  {
    return tag("head").nl
  }

  **
  ** End a <head> tag.
  **
  WebOutStream headEnd()
  {
    return tagEnd("head").nl
  }

  **
  ** Write a complete <title> tag.
  **
  WebOutStream title(Str title)
  {
    return tag("title").w(title).tagEnd("title").nl
  }

  **
  ** Write a complete <link> tag for an external CSS stylesheet.
  **
  WebOutStream css(Uri href)
  {
    attrs := "rel='stylesheet' type='text/css' href='$href'"
    return tag("link", attrs, true).nl
  }

  **
  ** Write a complete <script> tag for an external JavaScript file.
  **
  WebOutStream js(Uri href)
  {
    tag("script", "type='text/javascript' src='$href'")
    return tagEnd("script").nl
  }

//////////////////////////////////////////////////////////////////////////
// body
//////////////////////////////////////////////////////////////////////////

  **
  ** Start a <body> tag.
  **
  WebOutStream body(Str attrs := null)
  {
    return tag("body", attrs).nl
  }

  **
  ** End a <body> tag.
  **
  WebOutStream bodyEnd()
  {
    return tagEnd("body").nl
  }

//////////////////////////////////////////////////////////////////////////
// h1, h2, h3, h4, h5, h6
//////////////////////////////////////////////////////////////////////////

  **
  ** Write a complete <h1> tag.
  **
  WebOutStream h1(Str content, Str attrs := null)
  {
    return tag("h1", attrs).w(content).tagEnd("h1").nl
  }

  **
  ** Write a complete <h2> tag.
  **
  WebOutStream h2(Str content, Str attrs := null)
  {
    return tag("h2", attrs).w(content).tagEnd("h2").nl
  }

  **
  ** Write a complete <h3> tag.
  **
  WebOutStream h3(Str content, Str attrs := null)
  {
    return tag("h3", attrs).w(content).tagEnd("h3").nl
  }

  **
  ** Write a complete <h4> tag.
  **
  WebOutStream h4(Str content, Str attrs := null)
  {
    return tag("h4", attrs).w(content).tagEnd("h4").nl
  }

  **
  ** Write a complete <h5> tag.
  **
  WebOutStream h5(Str content, Str attrs := null)
  {
    return tag("h5", attrs).w(content).tagEnd("h5").nl
  }

  **
  ** Write a complete <h6> tag.
  **
  WebOutStream h6(Str content, Str attrs := null)
  {
    return tag("h6", attrs).w(content).tagEnd("h6").nl
  }

//////////////////////////////////////////////////////////////////////////
// div
//////////////////////////////////////////////////////////////////////////

  **
  ** Start a <div> tag.
  **
  WebOutStream div(Str attrs := null)
  {
    return tag("div", attrs).nl
  }

  **
  ** End a <div> tag.
  **
  WebOutStream divEnd()
  {
    return tagEnd("div").nl
  }

//////////////////////////////////////////////////////////////////////////
// span
//////////////////////////////////////////////////////////////////////////

  **
  ** Start a <span> tag.
  **
  WebOutStream span(Str attrs := null)
  {
    return tag("span", attrs)
  }

  **
  ** End a <span> tag.
  **
  WebOutStream spanEnd()
  {
    return tagEnd("span")
  }

//////////////////////////////////////////////////////////////////////////
// p
//////////////////////////////////////////////////////////////////////////

  **
  ** Start a <p> tag.
  **
  WebOutStream p(Str attrs := null)
  {
    return tag("p", attrs).nl
  }

  **
  ** End a <p> tag.
  **
  WebOutStream pEnd()
  {
    return tagEnd("p").nl
  }

//////////////////////////////////////////////////////////////////////////
// pre
//////////////////////////////////////////////////////////////////////////

  **
  ** Start a <pre> tag.
  **
  WebOutStream pre(Str attrs := null)
  {
    return tag("pre", attrs)
  }

  **
  ** End a <pre> tag.
  **
  WebOutStream preEnd()
  {
    return tagEnd("pre").nl
  }

//////////////////////////////////////////////////////////////////////////
// code
//////////////////////////////////////////////////////////////////////////

  **
  ** Start a <code> tag.
  **
  WebOutStream code(Str attrs := null)
  {
    return tag("code", attrs)
  }

  **
  ** End a <code> tag.
  **
  WebOutStream codeEnd()
  {
    return tagEnd("code")
  }

//////////////////////////////////////////////////////////////////////////
// hr
//////////////////////////////////////////////////////////////////////////

  **
  ** Write out a complete <hr/> tag.
  **
  WebOutStream hr()
  {
    return tag("hr", null, true).nl
  }

//////////////////////////////////////////////////////////////////////////
// br
//////////////////////////////////////////////////////////////////////////

  **
  ** Write out a complete <br/> tag.
  **
  WebOutStream br()
  {
    return tag("br", null, true)
  }

//////////////////////////////////////////////////////////////////////////
// a
//////////////////////////////////////////////////////////////////////////

  **
  ** Start a <a> tag.
  **
  WebOutStream a(Uri href, Str attrs := null)
  {
    return tag("a href='$href'", attrs)
  }

  **
  ** End a <a> tag.
  **
  WebOutStream aEnd()
  {
    return tagEnd("a")
  }

//////////////////////////////////////////////////////////////////////////
// img
//////////////////////////////////////////////////////////////////////////

  **
  ** Write a complete <img> tag.
  **
  WebOutStream img(Str src, Str alt := "", Str attrs := null)
  {
    return tag("img src='$src' alt='$alt'", attrs, true)
  }

//////////////////////////////////////////////////////////////////////////
// table
//////////////////////////////////////////////////////////////////////////

  **
  ** Start a <table> tag.
  **
  WebOutStream table(Str attrs := null)
  {
    return tag("table", attrs).nl
  }

  **
  ** End a <table> tag.
  **
  WebOutStream tableEnd()
  {
    return tagEnd("table").nl
  }

//////////////////////////////////////////////////////////////////////////
// tr
//////////////////////////////////////////////////////////////////////////

  **
  ** Start a <tr> tag.
  **
  WebOutStream tr(Str attrs := null)
  {
    return tag("tr", attrs).nl
  }

  **
  ** End a <tr> tag.
  **
  WebOutStream trEnd()
  {
    return tagEnd("tr").nl
  }

//////////////////////////////////////////////////////////////////////////
// td
//////////////////////////////////////////////////////////////////////////

  **
  ** Start a <td> tag.
  **
  WebOutStream td(Str attrs := null)
  {
    return tag("td", attrs)
  }

  **
  ** End a <td> tag.
  **
  WebOutStream tdEnd()
  {
    return tagEnd("td").nl
  }

//////////////////////////////////////////////////////////////////////////
// ul/ol/li
//////////////////////////////////////////////////////////////////////////

  **
  ** Start a <ul> tag.
  **
  WebOutStream ul(Str attrs := null)
  {
    return tag("ul", attrs).nl
  }

  **
  ** End a <ul> tag.
  **
  WebOutStream ulEnd()
  {
    return tagEnd("ul").nl
  }

  **
  ** Start a <ol> tag.
  **
  WebOutStream ol(Str attrs := null)
  {
    return tag("ol", attrs).nl
  }

  **
  ** End a <ol> tag.
  **
  WebOutStream olEnd()
  {
    return tagEnd("ol").nl
  }

  **
  ** Start a <li> tag.
  **
  WebOutStream li(Str attrs := null)
  {
    return tag("li", attrs).nl
  }

  **
  ** End a <li> tag.
  **
  WebOutStream liEnd()
  {
    return tagEnd("li").nl
  }

//////////////////////////////////////////////////////////////////////////
// dl/dd/dt
//////////////////////////////////////////////////////////////////////////

  **
  ** Start a <dl> tag.
  **
  WebOutStream dl(Str attrs := null)
  {
    return tag("dl", attrs).nl
  }

  **
  ** End a <dl> tag.
  **
  WebOutStream dlEnd()
  {
    return tagEnd("dl").nl
  }

  **
  ** Start a <dt> tag.
  **
  WebOutStream dt(Str attrs := null)
  {
    return tag("dt", attrs).nl
  }

  **
  ** End a <dt> tag.
  **
  WebOutStream dtEnd()
  {
    return tagEnd("dt").nl
  }

  **
  ** Start a <dd> tag.
  **
  WebOutStream dd(Str attrs := null)
  {
    return tag("dd", attrs).nl
  }

  **
  ** End a <dd> tag.
  **
  WebOutStream ddEnd()
  {
    return tagEnd("dd").nl
  }

//////////////////////////////////////////////////////////////////////////
// form
//////////////////////////////////////////////////////////////////////////

  **
  ** Start a <form> tag.
  **
  WebOutStream form(Str attrs := null)
  {
    return tag("form", attrs).nl
  }

  **
  ** End a <form> tag.
  **
  WebOutStream formEnd()
  {
    return tagEnd("form").nl
  }

//////////////////////////////////////////////////////////////////////////
// input
//////////////////////////////////////////////////////////////////////////

  **
  ** Write a complete <input> tag.
  **
  WebOutStream input(Str attrs := null)
  {
    return tag("input", attrs, true)
  }

  **
  ** Convenience for input("type='text'" + attrs).
  **
  WebOutStream textField(Str attrs := null)
  {
    return tag("input type='text'", attrs, true)
  }

  **
  ** Convenience for input("type='button'" + attrs).
  **
  WebOutStream button(Str attrs := null)
  {
    return tag("input type='button'", attrs, true)
  }

  **
  ** Convenience for input("type='submit'" + attrs).
  **
  WebOutStream submit(Str attrs := null)
  {
    return tag("input type='submit'", attrs, true)
  }

//////////////////////////////////////////////////////////////////////////
// select
//////////////////////////////////////////////////////////////////////////

  **
  ** Start a <select> tag.
  **
  WebOutStream select(Str attrs := null)
  {
    return tag("select", attrs).nl
  }

  **
  ** End a <select> tag.
  **
  WebOutStream selectEnd()
  {
    return tagEnd("select").nl
  }

  **
  ** Start a <option> tag.
  **
  WebOutStream option(Str attrs := null)
  {
    return tag("option", attrs)
  }

  **
  ** End a <option> tag.
  **
  WebOutStream optionEnd()
  {
    return tagEnd("option").nl
  }

}

//////////////////////////////////////////////////////////////////////////
// WebOutStreamWrapper
//////////////////////////////////////////////////////////////////////////

internal class WebOutStreamWrapper : WebOutStream
{
  new make(OutStream out) { this.out = out }

  override OutStream write(Int byte)
  {
    out.write(byte)
    return this
  }

  override OutStream writeBuf(Buf buf, Int n)
  {
    out.writeBuf(buf, n)
    return this
  }

  override OutStream writeChar(Int char)
  {
    out.writeChar(char)
    return this
  }

  override OutStream writeChars(Str s, Int off := 0, Int n := s.size-off)
  {
    out.writeChars(s, off, n)
    return this
  }

  override Charset charset := Charset.utf8
  OutStream out := null
}
